unit zbar_bch15_5;
(*Copyright (C) 2008-2009  Timothy B. Terriberry (tterribe@xiph.org)
  You can redistribute this library and/or modify it under the terms of the
   GNU Lesser General Public License as published by the Free Software
   Foundation; either version 2.1 of the License, or (at your option) any later
   version.*)

interface

function bch15_5_correct(var _y: LongWord): integer;

implementation

const
  (*A cycle in GF(2**4) generated by alpha=(x**4+x+1).
    It is extended an extra 16 entries to avoid some expensive mod operations.*)
  gf16_exp: array [0..30] of byte = (1,2,4,8,3,6,12,11,5,10,7,14,15,13,9,1,2,4,8,3,6,12,11,5,10,7,14,15,13,9,1);

  (*The location of each integer 1...16 in the cycle.*)
  gf16_log: array [0..15] of shortint = (-1,0,1,4,2,8,5,10,3,14,9,7,6,13,11,12);


(*Multiplication in GF(2**4) using logarithms.*)
function gf16_mul(_a, _b: LongWord): LongWord;
begin
  if (_a=0) or (_b=0) then
    Result := 0
  else
    Result := gf16_exp[gf16_log[_a]+gf16_log[_b]];
end;

(*Division in GF(2**4) using logarithms.
  The result when dividing by zero is undefined.*)
function gf16_div(_a, _b: LongWord): LongWord;
begin
  if _a = 0 then
    Result := 0
  else
    Result := gf16_exp[gf16_log[_a]+15-gf16_log[_b]];
end;

(*Multiplication in GF(2**4) when the second argument is known to be non-zero
   (proven by representing it by its logarithm).*)
function gf16_hmul(_a, _logb: LongWord): LongWord;
begin
  if _a = 0 then
    Result := 0
  else
    Result := gf16_exp[gf16_log[_a]+_logb];
end;

(*The syndrome normally has five values, S_1 ... S_5.
  We only calculate and store the odd ones in _s, since S_2=S_1**2 and
   S_4=S_2**2.
  Returns zero if all the syndrome values are zero.*)
function bch15_5_calc_syndrome(_s: PLongWord;  _y: LongWord): boolean;
var
  p: LongWord;
  i, j: integer;
begin
  p := 0;
  for i:=0 to Pred(15) do
    if (_y and (1 shl i)) <> 0 then
      p := p xor gf16_exp[i];
  _s[0] := p;

  p := 0;
  for i:=0 to Pred(3) do
    for j:=0 to Pred(5) do
      if (_y and (1 shl (5 * i + j))) <> 0 then
        p := p xor gf16_exp[j*3];
  _s[1] := p;

  p := 0;
  for i:=0 to Pred(5) do
    for j:=0 to Pred(3) do
      if (_y and (1 shl (3 * i + j))) <> 0 then
        p := p xor gf16_exp[j*5];
  _s[2] := p;

  Result := (_s[0]<>0) or (_s[1]<>0) or (_s[2]<>0);
end;

(*Compute the coefficients of the error-locator polynomial.
  Returns the number of errors (the degree of the polynomial).*)
function bch15_5_calc_omega(_o, _s: PLongWord): integer;
var
  s02, tt, dd: LongWord;
begin
  _o[0] := _s[0];
  s02 := gf16_mul(_s[0], _s[0]);
  dd := _s[1] xor gf16_mul(_s[0], s02);
  tt := _s[2] xor gf16_mul(s02, _s[1]);
  if dd = 0 then
    _o[1] := 0
  else
    _o[1] := gf16_div(tt,dd);
  _o[2] := dd xor gf16_mul(_s[0],_o[1]);
  Result := 3;
  while (Result>0) and (_o[Result-1]=0) do Dec(Result);
end;

(*Find the roots of the error polynomial.
  Returns the number of roots found, or a negative value if the polynomial did
   not have enough roots, indicating a decoding error.*)
function bch15_5_calc_epos(_epos, _s: PLongWord): integer;
var
  o: array [0..2] of LongWord;
  d, i, i2: integer;
begin
  d := bch15_5_calc_omega(o,_s);
  Result := 0;
  if d=1 then begin
    _epos[Result] := gf16_log[o[0]];
    Inc(Result);
  end
  else if d>0 then begin
    for i:=0 to Pred(15) do begin
      i2 := gf16_log[gf16_exp[i shl 1]];
      if (gf16_exp[i+i2] xor gf16_hmul(o[0],i2) xor gf16_hmul(o[1],i) xor o[2]) = 0 then begin
        _epos[Result] := i;
        Inc(Result);
      end;
    end;
    if Result<d then Result:=-1;
  end;
end;

function bch15_5_encode(_x: LongWord): LongWord;
begin
  Result := (-(_x and 1) and $0537) xor (-(_x shr 1 and 1) and $0A6E) xor (-(_x shr 2 and 1) and $11EB) xor (-(_x shr 3 and 1) and $23D6) xor (-(_x shr 4 and 1) and $429B);
end;

function bch15_5_correct(var _y: LongWord): integer;
var
  s, epos: array [0..Pred(3)] of LongWord;
  y: LongWord;
  i: integer;
begin
  y := _y;
  if not bch15_5_calc_syndrome(s, y) then
    Exit(0);
  Result := bch15_5_calc_epos(epos, s);
  if Result > 0 then begin
    (*If we had a non-zero syndrome value, we should always find at least one
      error location, or we've got a decoding error.*)
    for i:=0 to Pred(Result) do
      y := y xor (1 shl epos[i]);
    (*If there were too many errors, we may not find enough roots to reduce the
      syndrome to zero.
      We could recompute it to check, but it's much faster just to check that
      we have a valid codeword.*)
    if bch15_5_encode(y shr 10) = y then begin
      (*Decoding succeeded.*)
      _y := y;
      Exit;
    end;
  end;
  (*Decoding failed due to too many bit errors.*)
  Result := -1;
end;

end.
