unit zbar_binarize;
(*Copyright (C) 2008-2009  Timothy B. Terriberry (tterribe@xiph.org)
  You can redistribute this library and/or modify it under the terms of the
   GNU Lesser General Public License as published by the Free Software
   Foundation; either version 2.1 of the License, or (at your option) any later
   version.*)

interface

function qr_binarize(_img: pbyte;  _width, _height: integer): pbyte; 

implementation

type
  unsigned = LongWord;
  punsigned = ^unsigned;

(*The above algorithms are computationally expensive, and do not work as well
   as the simple algorithm below.
  Sauvola by itself does an excellent job of classifying regions outside the
   QR code as background, which greatly reduces the chance of false alarms.
  However, it also tends to over-shrink isolated black dots inside the code,
   making them easy to miss with even slight mis-alignment.
  Since the Gatos method uses Sauvola as input to its background interpolation
   method, it cannot possibly mark any pixels as foreground which Sauvola
   classified as background, and thus suffers from the same problem.
  The following simple adaptive threshold method does not have this problem,
   though it produces essentially random noise outside the QR code region.
  QR codes are structured well enough that this does not seem to lead to any
   actual false alarms in practice, and it allows many more codes to be
   detected and decoded successfully than the Sauvola or Gatos binarization
   methods.*)

(*A simplified adaptive thresholder.
  This compares the current pixel value to the mean value of a (large) window
   surrounding it.*)

//#define QR_MAXI(_a,_b)      ((_a)-((_a)-(_b)&-((_b)>(_a))))
function QR_MAXI(_a, _b: integer): integer; inline;
begin
 if _a > _b then
   Result := _a
 else
   Result := _b;
end;

//#define QR_MINI(_a,_b)      ((_a)+((_b)-(_a)&-((_b)<(_a))))
function QR_MINI(_a, _b: integer): integer; inline;
begin
 if _a < _b then
   Result := _a
 else
   Result := _b;
end;

function qr_binarize(_img: PByte;  _width, _height: integer): PByte; 
var
  col_sums: punsigned;
  logwindw, logwindh: integer; 
  windw, windh: integer; 
  y0offs, y1offs: integer; 
  g,m: unsigned;
  x,y,x0,x1: integer;
begin
  Result := GetMem(_width*_height*sizeof(Byte));
  (*We keep the window size fairly large to ensure it doesn't fit completely
    inside the center of a finder pattern of a version 1 QR code at full
    resolution.*) 
  if (_width>0) and (_height>0) then begin 
    logwindw:=4;
    while (logwindw<8) and ((1 shl logwindw) < ((_width+7) shr 3)) do Inc(logwindw);
    logwindh:=4;
    while (logwindh<8) and ((1 shl logwindh) < ((_height+7) shr 3)) do Inc(logwindh);

    windw := 1 shl logwindw; 
    windh := 1 shl logwindh; 

    col_sums := GetMem(_width*sizeof(unsigned));
    (*Initialize sums down each column.*)
    for x:=0 to Pred(_width) do begin 
      g := _img[x]; 
      col_sums[x] := (g shl (logwindh-1)) + g; 
    end;

    for y:=1 to Pred((windh shr 1)) do begin 
      y1offs := QR_MINI(y,_height-1)*_width; 
      for x:=0 to Pred(_width) do begin 
        g := _img[y1offs+x]; 
        Inc(col_sums[x], g); 
      end;
    end;

    for y:=0 to Pred(_height) do begin 
      (*Initialize the sum over the window.*)
      m := (col_sums[0] shl (logwindw-1)) + col_sums[0]; 
      for x:=1 to Pred((windw shr 1)) do begin 
        x1 := QR_MINI(x,_width-1); 
        Inc(m, col_sums[x1]); 
      end;
      for x:=0 to Pred(_width) do begin 
        (*Perform the test against the threshold T = (m/n)-D,
                   where n=windw*windh and D=3.*)
        g := _img[y*_width+x];
        // black -> 255, white -> 0
        Result[y*_width+x] := (-Ord((g+3) shl (logwindw+logwindh) < m)) and $FF;
        (*Update the window sum.*)
        if x+1<_width then begin 
          x0 := QR_MAXI(0, x-(windw shr 1)); 
          x1 := QR_MINI(x+(windw shr 1), _width-1); 
          Inc(m, col_sums[x1]-col_sums[x0]); 
        end;
      end;

      (*Update the column sums.*)
      if y+1<_height then begin 
        y0offs := QR_MAXI(0, y-(windh shr 1))*_width;
        y1offs := QR_MINI(y+(windh shr 1), _height-1)*_width;
        for x:=0 to Pred(_width) do begin 
          Dec(col_sums[x], _img[y0offs+x]); 
          Inc(col_sums[x], _img[y1offs+x]); 
        end;
      end;
    end;

    FreeMem(col_sums);
  end;
end;

end.
