(*Written by Timothy B. Terriberry (tterribe@xiph.org) 1999-2009 public domain.
  Based on the public domain implementation by Robert J. Jenkins Jr.*)
unit zbar_isaac;

{$SMARTLINK ON}
{$POINTERMATH ON}
{$OVERFLOWCHECKS OFF}
{$RANGECHECKS OFF}

interface

const
  ISAAC_SZ_LOG = 8;
  ISAAC_SZ = 1 shl ISAAC_SZ_LOG;
  ISAAC_SEED_SZ_MAX = ISAAC_SZ shl 2;

(*ISAAC is the most advanced of a series of Pseudo-Random Number Generators
   designed by Robert J. Jenkins Jr. in 1996.
  http://www.burtleburtle.net/bob/rand/isaac.html
  To quote:
    No efficient method is known for deducing their internal states.
    ISAAC requires an amortized 18.75 instructions to produce a 32-bit value.
    There are no cycles in ISAAC shorter than 2**40 values.
    The expected cycle length is 2**8295 values.*)

type
  unsigned = LongWord;
  punsigned = ^unsigned;

  isaac_ctx = record
    n: unsigned;
    r: array[0..ISAAC_SZ-1] of unsigned;
    m: array[0..ISAAC_SZ-1] of unsigned;
    a, b, c: unsigned;
  end;
  pisaac_ctx = ^isaac_ctx;

procedure isaac_init(var _ctx: isaac_ctx;  _seed: pointer;  _nseed: integer);
function isaac_next_uint(var _ctx: isaac_ctx;  _n: unsigned): unsigned;

implementation

const
  ISAAC_MASK = $FFFFFFFF;


procedure isaac_update(var _ctx: isaac_ctx);
var
  m, r: punsigned;
  a,b, x,y: unsigned;
  i: integer;
begin
  m := @_ctx.m[0];
  r := @_ctx.r[0];
  a := _ctx.a;
  inc(_ctx.c);
  b := (_ctx.b + {++}_ctx.c) and ISAAC_MASK;

  i := 0;
  while i < ISAAC_SZ div 2 do begin
    x := m[i];
    a := ((a xor (a shl 13)) + m[i+ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[(y shr (ISAAC_SZ_LOG+2)) and (ISAAC_SZ-1)] + x) and ISAAC_MASK;
    r[i] := b;
    inc(i);
    x := m[{++}i];
    a := ((a xor (a shr 6)) + m[i+ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[y shr (ISAAC_SZ_LOG+2) and (ISAAC_SZ-1)] + x) and ISAAC_MASK;
    r[i] := b;
    inc(i);
    x := m[{++}i];
    a := ((a xor (a shl 2)) + m[i+ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[(y shr (ISAAC_SZ_LOG+2)) and (ISAAC_SZ-1)]+x) and ISAAC_MASK;
    r[i] := b;
    inc(i);
    x := m[{++}i];
    a := ((a xor (a shr 16)) + m[i+ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[(y shr (ISAAC_SZ_LOG+2)) and (ISAAC_SZ-1)]+x) and ISAAC_MASK;
    r[i] := b;
    Inc(i);
  end;

  i := ISAAC_SZ div 2;
  while i < ISAAC_SZ do begin
    x := m[i];
    a := ((a xor (a shl 13)) + m[i-ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[(y shr (ISAAC_SZ_LOG+2)) and (ISAAC_SZ-1)]+x) and ISAAC_MASK;
    r[i] := b;
    inc(i);
    x := m[{++}i];
    a := ((a xor (a shr 6)) + m[i-ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[(y shr (ISAAC_SZ_LOG+2)) and (ISAAC_SZ-1)]+x) and ISAAC_MASK;
    r[i] := b;
    inc(i);
    x := m[{++}i];
    a := ((a xor (a shl 2)) + m[i-ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[(y shr (ISAAC_SZ_LOG+2)) and (ISAAC_SZ-1)]+x) and ISAAC_MASK;
    r[i] := b;
    inc(i);
    x := m[{++}i];
    a := ((a xor (a shr 16)) + m[i-ISAAC_SZ div 2]) and ISAAC_MASK;
    y := (m[(x and ((ISAAC_SZ-1) shl 2)) shr 2]+a+b) and ISAAC_MASK;
    m[i] := y;
    b := (m[(y shr (ISAAC_SZ_LOG+2)) and (ISAAC_SZ-1)]+x) and ISAAC_MASK;
    r[i]:= b;
    Inc(i);
  end;
  _ctx.b := b;
  _ctx.a := a;
  _ctx.n := ISAAC_SZ;
end;

procedure isaac_mix(_x: punsigned);
const
  SHIFT: array [0..7] of byte = (11,2,8,16,10,4,8,9);
var
  i: integer;
begin
  i := 0;
  while i < 8 do begin
    _x[i] := _x[i] xor (_x[(i+1) and 7] shl SHIFT[i]);
    Inc(_x[(i+3) and 7], _x[i]);
    Inc(_x[(i+1) and 7], _x[(i+2) and 7]);
    Inc(i);
    _x[i] := _x[i] xor (_x[(i+1) and 7] shr SHIFT[i]);
    Inc(_x[(i+3) and 7], _x[i]);
    Inc(_x[(i+1) and 7], _x[(i+2) and 7]);
    Inc(i);
  end;
end;

procedure isaac_init(var _ctx: isaac_ctx;  _seed: pointer;  _nseed: integer);
var
  seed: pbyte;
  x: array [0..7] of unsigned;
  i,j: integer;
begin
  _ctx.a := 0;
  _ctx.b := 0;
  _ctx.c := 0;
  for i:=0 to High(x) do x[i]:=$9E3779B9;
  for i:=0 to Pred(4) do isaac_mix(@x[0]);
  if _nseed>ISAAC_SEED_SZ_MAX then _nseed:= ISAAC_SEED_SZ_MAX;
  seed := PByte(_seed);

  i := 0;
  while i < _nseed shr 2 do begin
    _ctx.r[i] := seed[(i shl 2) or 3] shl 24 or seed[(i shl 2) or 2] shl 16 or seed[(i shl 2) or 1] shl 8 or seed[i shl 2];
    Inc(i);
  end;
  if (_nseed and 3) <> 0 then begin
    _ctx.r[i] := seed[i shl 2];
    for j:=1 to Pred((_nseed and 3)) do
      Inc(_ctx.r[i], seed[(i shl 2) or j] shl (j shl 3));
    Inc(i);
  end;

  FillByte(_ctx.r[i], (ISAAC_SZ-i)*SizeOf(_ctx.r[0]), 0);
  i := 0;
  while i<ISAAC_SZ do begin
    for j:=0 to Pred(8) do
      Inc(x[j], _ctx.r[i+j]);
    isaac_mix(@x[0]);
    Move(x[0], _ctx.m[i], SizeOf(x));
    Inc(i, 8);
  end;

  i:= 0;
  while i<ISAAC_SZ do begin
    for j:=0 to Pred(8) do
      Inc(x[j], _ctx.m[i+j]);
    isaac_mix(@x[0]);
    Move(x[0], _ctx.m[i], SizeOf(x));
    Inc(i, 8);
  end;

  isaac_update(_ctx);
end;

function isaac_next_uint32(var _ctx: isaac_ctx): unsigned;
begin
  if _ctx.n = 0 then isaac_update(_ctx);
  dec(_ctx.n);
  Result := _ctx.r[{--}_ctx.n];
end;

(*Returns a uniform random integer less than the given maximum value.
  _n: The upper bound on the range of numbers returned (not inclusive).
      This must be strictly less than 2**32.
  Return: An integer uniformly distributed between 0 (inclusive) and _n
           (exclusive).*)
function isaac_next_uint(var _ctx: isaac_ctx;  _n: unsigned): unsigned;
var
  r,v,d: unsigned;
begin
  repeat
    r := isaac_next_uint32(_ctx);
    v := r mod _n;
    d := r-v;
  until not (((d+_n-1) and ISAAC_MASK)<d);
  Result := v;
end;

end.
