unit zbar_rs;
(*Copyright (C) 1991-1995  Henry Minsky (hqm@ua.com, hqm@ai.mit.edu)
  Copyright (C) 2008-2009  Timothy B. Terriberry (tterribe@xiph.org)
  You can redistribute this library and/or modify it under the terms of the
   GNU Lesser General Public License as published by the Free Software
   Foundation; either version 2.1 of the License, or (at your option) any later
   version.*)

interface

(*Reed-Solomon encoder and decoder.
  Original implementation (C) Henry Minsky (hqm@ua.com, hqm@ai.mit.edu),
   Universal Access 1991-1995.
  Updates by Timothy B. Terriberry (C) 2008-2009:
   - Properly reject codes when error-locator polynomial has repeated roots or
      non-trivial irreducible factors.
   - Removed the hard-coded parity size and performed general API cleanup.
   - Allow multiple representations of GF(2**8), since different standards use
      different irreducible polynomials.
   - Allow different starting indices for the generator polynomial, since
      different standards use different values.
   - Greatly reduced the computation by eliminating unnecessary operations.
   - Explicitly solve for the roots of low-degree polynomials instead of using
      an exhaustive search.
     This is another major speed boost when there are few errors.*)

const
  (*This is one of 16 irreducible primitive polynomials of degree 8:
      x**8+x**4+x**3+x**2+1.
    Under such a polynomial, x (i.e., 0x02) is a generator of GF(2**8).
    The high order 1 bit is implicit.
    From~\cite{MD88}, Ch. 5, p. 275 by Patel.
    @BOOK{MD88,
      author="C. Dennis Mee and Eric D. Daniel",
      title="Video, Audio, and Instrumentation Recording",
      series="Magnetic Recording",
      volume=3,
      publisher="McGraw-Hill Education",
      address="Columbus, OH",
      month=Jun,
      year=1988
    }*)
  QR_PPOLY = $1D;

  (*The index to start the generator polynomial from (0...254).*)
  QR_M0 = 0;

type
  rs_gf256 = record
    (* A logarithm table in GF(2**8). *)
    log: array [0..255] of Byte;

    (* An exponential table in GF(2**8): exp[i] contains x^i reduced modulo the
       irreducible primitive polynomial used to define the field.
      The extra 256 entries are used to do arithmetic mod 255, since some extra
       table lookups are generally faster than doing the modulus.*)
    exp: array[0..510] of Byte;
  end;
  prs_gf256 = ^rs_gf256;

(*Initialize discrete logarithm tables for GF(2**8) using a given primitive
     irreducible polynomial.*)
procedure rs_gf256_init(var _gf: rs_gf256;  _ppoly: LongWord);

(*Corrects a codeword with _ndata<256 bytes, of which the last _npar are parity
   bytes.
  Known locations of errors can be passed in the _erasures array.
  Twice as many (up to _npar) errors with a known location can be corrected
   compared to errors with an unknown location.
  Returns the number of errors corrected if successful, or a negative number if
   the message could not be corrected because too many errors were detected.*)
function rs_correct(_gf: prs_gf256;  _m0: integer;  _data: pbyte;  _ndata: integer;  _npar: integer;  _erasures: pbyte;  _nerasures: integer): integer;

(*Create an _npar-coefficient generator polynomial for a Reed-Solomon code with
   _npar<256 parity bytes.*)
//procedure rs_compute_genpoly(_gf: prs_gf256;  _m0: integer;  _genpoly: pbyte;  _npar: integer);

(*Adds _npar<=_ndata parity bytes to an _ndata-_npar byte message.
  _data must contain room for _ndata<256 bytes.*)
//procedure rs_encode(_gf: prs_gf256;  _data: pbyte;  _ndata: integer;  _genpoly: pbyte;  _npar: integer);


implementation

(* Galois Field arithmetic in GF(2**8). *)
procedure rs_gf256_init(var _gf: rs_gf256;  _ppoly: LongWord);
var
  p: LongWord;
  i: integer;
begin
  (*Initialize the table of powers of a primtive root, alpha=0x02.*)  
  p := 1;
  for i:=0 to 255 do begin
    _gf.exp[i] := p;
    _gf.exp[i+255] := p;
    p := ((p shl 1) xor (-(p shr 7) and _ppoly)) and $FF;
  end;
  (*Invert the table to recover the logs.*)
  for i:=0 to 254 do
    _gf.log[_gf.exp[i]] := i;
  (*Note that we rely on the fact that _gf->log[0]=0 below.*)
  _gf.log[0] := 0;
end;

(*Multiplication in GF(2**8) using logarithms.*)
function rs_gmul(_gf: prs_gf256;  _a, _b: LongWord): LongWord;
begin
  if (_a=0) or (_b=0) then
    Result := 0
  else
    Result := _gf^.exp[_gf^.log[_a]+_gf^.log[_b]];
end;

(*Division in GF(2**8) using logarithms.
  The result of division by zero is undefined.*)
function rs_gdiv(_gf: prs_gf256;  _a, _b: LongWord): LongWord;
begin
  if _a = 0 then
    Result := 0
  else
    Result := _gf^.exp[_gf^.log[_a]+255-_gf^.log[_b]];
end;

(*Multiplication in GF(2**8) when one of the numbers is known to be non-zero
   (proven by representing it by its logarithm).*)
function rs_hgmul(_gf: prs_gf256;  _a, _logb: LongWord): LongWord;
begin
  if _a = 0 then
    Result := 0
  else
    Result := _gf^.exp[_gf^.log[_a]+_logb];
end;

(*Square root in GF(2**8) using logarithms.*)
function rs_gsqrt(_gf: prs_gf256;  _a: LongWord): LongWord;
var
  loga: LongWord;
begin
  if _a = 0 then
    Result := 0
  else begin
    loga := _gf^.log[_a];
    Result := _gf^.exp[(loga + (255 and -(loga and 1))) shr 1];
  end;
end;

(*Polynomial root finding in GF(2**8).
  Each routine returns a list of the distinct roots (i.e., with duplicates
   removed).*)
(*Solve a quadratic equation x**2 + _b*x + _c in GF(2**8) using the method
   of~\cite{Wal99}.
  Returns the number of distinct roots.
  ARTICLE{Wal99,
    author="C. Wayne Walker",
    title="New Formulas for Solving Quadratic Equations over Certain Finite
     Fields",
    journal="{IEEE} Transactions on Information Theory",
    volume=45,
    number=1,
    pages="283--284",
    month=Jan,
    year=1999
  }*)
function rs_quadratic_solve(_gf: prs_gf256;  _b: LongWord;  _c: LongWord;  _x: PByte{unsigned char _x[2]}): integer;
var
  b, logb, logb2, logb4, logb8, logb12, logb14, logc, logc2, logc4, c8, g3, z3, l3, c0, g2, l2, z2: LongWord;
  inc: integer;
begin
  (*If _b is zero, all we need is a square root.*)  
  if _b = 0 then begin
    _x[0] := rs_gsqrt(_gf,_c);
    Result := 1;
    Exit;
  end;
  (*If _c is zero, 0 and _b are the roots.*)
  if _c = 0 then begin
    _x[0] := 0;
    _x[1] := _b;
    Result := 2;
    Exit;
  end;
  logb := _gf^.log[_b];
  logc := _gf^.log[_c];
  (*If _b lies in GF(2**4), scale x to move it out.*)
  if logb mod (255 div 15)=0 then
    inc := 1
  else
    inc := 0;
  if inc<>0 then begin
    b := _gf^.exp[logb+254];
    logb := _gf^.log[b];
    _c := _gf^.exp[logc+253];
    logc := _gf^.log[_c];
  end
  else
    b := _b;
  logb2 := _gf^.log[_gf^.exp[logb shl 1]];
  logb4 := _gf^.log[_gf^.exp[logb2 shl 1]];
  logb8 := _gf^.log[_gf^.exp[logb4 shl 1]];
  logb12:= _gf^.log[_gf^.exp[logb4+logb8]]; 
  logb14:= _gf^.log[_gf^.exp[logb2+logb12]]; 
  logc2 := _gf^.log[_gf^.exp[logc shl 1]];
  logc4 := _gf^.log[_gf^.exp[logc2 shl 1]];
  c8 := _gf^.exp[logc4 shl 1];
  g3 := rs_hgmul(_gf, _gf^.exp[logb14+logc] xor _gf^.exp[logb12+logc2] xor _gf^.exp[logb8+logc4] xor c8, logb);
  (*If g3 doesn't lie in GF(2**4), then our roots lie in an extension field.
      Note that we rely on the fact that _gf->log[0]==0 here.*)
  if _gf^.log[g3] mod (255 div 15)<>0 then begin
    Result := 0;
    Exit;
  end;
  (*Construct the corresponding quadratic in GF(2**4):
      x**2 + x/alpha**(255/15) + l3/alpha**(2*(255/15))*)
  z3 := rs_gdiv(_gf, g3, _gf^.exp[logb8 shl 1] xor b);
  l3 := rs_hgmul(_gf, rs_gmul(_gf,z3,z3) xor rs_hgmul(_gf,z3,logb) xor _c,255-logb2);
  c0 := rs_hgmul(_gf, l3, 255-2*(255 div 15));
  (*Construct the corresponding quadratic in GF(2**2):
      x**2 + x/alpha**(255/3) + l2/alpha**(2*(255/3))*)
  g2 := rs_hgmul(_gf, rs_hgmul(_gf,c0,255-2*(255 div 15)) xor rs_gmul(_gf,c0,c0),255-255 div 15);
  z2 := rs_gdiv(_gf, g2, _gf^.exp[255-(255 div 15)*4] xor _gf^.exp[255-(255 div 15)]);
  l2 := rs_hgmul(_gf, rs_gmul(_gf,z2,z2) xor rs_hgmul(_gf,z2,255-(255 div 15)) xor c0,2*(255 div 15));
  (*Back substitute to the solution in the original field.*)
  _x[0]:= _gf^.exp[_gf^.log[z3 xor rs_hgmul(_gf,rs_hgmul(_gf,l2,255 div 3) xor rs_hgmul(_gf,z2,255 div 15),logb)]+inc];
  _x[1]:= _x[0] xor _b;
  Result := 2;
end;

(*Solve a cubic equation x**3 + _a*x**2 + _b*x + _c in GF(2**8).
  Returns the number of distinct roots.*)
function rs_cubic_solve(_gf: prs_gf256;  _a: LongWord;  _b: LongWord;  _c: LongWord;  _x: PByte{unsigned char _x[3]}): integer;
var
  k, logd, d2, logd2, logw: LongWord;
  nroots, logx: integer;
begin
  (*If _c is zero, factor out the 0 root.*)  
  if _c = 0 then begin
    nroots := rs_quadratic_solve(_gf,_a,_b,_x);
    if _b<>0 then
      _x[nroots{++}]:= 0;
    Inc(nroots);
    Exit(nroots);
  end;
  (*Substitute x=_a+y*sqrt(_a**2+_b) to get y**3 + y + k == 0,
         k = (_a*_b+c)/(_a**2+b)**(3/2).*)
  k := rs_gmul(_gf,_a,_b) xor _c;
  d2 := rs_gmul(_gf,_a,_a) xor _b;
  if d2 = 0 then begin
    if k = 0 then begin
      (*We have a triple root.*)
      _x[0] := _a;
      Exit(1);
    end;
    logx := _gf^.log[k];
    if logx mod 3<>0 then
      Exit(0);
    logx := logx div 3;
    _x[0] := _a xor _gf^.exp[logx];
    _x[1] := _a xor _gf^.exp[logx+255 div 3];
    _x[2] := _a xor _x[0] xor _x[1];
    Exit(3);
  end;
  logd2 := _gf^.log[d2];
  logd := (logd2 + (255 and -(logd2 and 1))) shr 1;
  k := rs_gdiv(_gf,k,_gf^.exp[logd+logd2]);
  (*Substitute y=w+1/w and z=w**3 to get z**2 + k*z + 1 == 0.*)
  nroots := rs_quadratic_solve(_gf,k,1,_x);
  if nroots<1 then begin
    (*The Reed-Solomon code is only valid if we can find 3 distinct roots in
           GF(2**8), so if we know there's only one, we don't actually need to find
           it.
          Note that we're also called by the quartic solver, but if we contain a
           non-trivial irreducible factor, than so does the original
           quartic~\cite{LW72}, and failing to return a root here actually saves us
           some work there, also.*)
    Exit(0);
  end;
  (*Recover w from z.*)
  logw := _gf^.log[_x[0]];
  if logw<>0 then begin
    if logw mod 3<>0 then
      Exit(0);
    logw := logw div 3;
    (*Recover x from w.*)
    _x[0] := _gf^.exp[_gf^.log[_gf^.exp[logw] xor _gf^.exp[255-logw]]+logd] xor _a;
    logw := logw + (255 div 3);
    _x[1] := _gf^.exp[_gf^.log[_gf^.exp[logw] xor _gf^.exp[255-logw]]+logd] xor _a;
    _x[2] := _x[0] xor _x[1] xor _a;
    Result := 3;
  end
  else begin
    _x[0]:= _a;
    (*In this case _x[1] is a double root, so we know the Reed-Solomon code is
           invalid.
          Note that we still have to return at least one root, because if we're
           being called by the quartic solver, the quartic might still have 4
           distinct roots.
          But we don't need more than one root, so we can avoid computing the
           expensive one.*)
    (*_x[1]=_gf->exp[_gf->log[_gf->exp[255/3]^_gf->exp[2*(255/3)]]+logd]^_a;*)
    Result := 1;
  end;
end;

(*Solve a quartic equation x**4 + _a*x**3 + _b*x**2 + _c*x + _d in GF(2**8) by
   decomposing it into the cases given by~\cite{LW72}.
  Returns the number of distinct roots.
  @ARTICLE{LW72,
    author="Philip A. Leonard and Kenneth S. Williams",
    title="Quartics over $GF(2^n)$",
    journal="Proceedings of the American Mathematical Society",
    volume=36,
    number=2,
    pages="347--450",
    month=Dec,
    year=1972
  }*)
function rs_quartic_solve(_gf: prs_gf256;  _a: LongWord;  _b: LongWord;  _c: LongWord;  _d: LongWord;  _x: PByte{unsigned char _x[3]}): integer;
var
  r, s, t, b: LongWord;
  nroots, i: integer;
  loga, logti: LongWord;
begin
  (*If _d is zero, factor out the 0 root.*)  
  if _d=0 then begin
    nroots := rs_cubic_solve(_gf,_a,_b,_c,_x);
    if _c<>0 then
      _x[nroots{++}] := 0;
    Inc(nroots);
    Exit(nroots);
  end;

  if _a<>0 then begin
    (*Substitute x=(1/y) + sqrt(_c/_a) to eliminate the cubic term.*)
    loga := _gf^.log[_a];
    r := rs_hgmul(_gf,_c,255-loga);
    s := rs_gsqrt(_gf,r);
    t := _d xor rs_gmul(_gf,_b,r) xor rs_gmul(_gf,r,r);
    if t<>0 then begin
      logti := 255-_gf^.log[t];
      nroots := rs_quartic_solve(_gf,0,rs_hgmul(_gf,_b xor rs_hgmul(_gf,s,loga),logti), _gf^.exp[loga+logti], _gf^.exp[logti],_x); (*The result is still quartic, but with no cubic term.*)
      for i:=0 to Pred(nroots) do
      _x[i] := _gf^.exp[255-_gf^.log[_x[i]]] xor s;
    end
    else begin
      (*s must be a root~\cite{LW72}, and is in fact a double-root~\cite{CCO69}.
              Thus we're left with only a quadratic to solve.
              @ARTICLE{CCO69,
                author="Robert T. Chien and B. D. Cunningham and I. B. Oldham",
                title="Hybrid Methods for Finding Roots of a Polynomial---With
                 Applications to {BCH} Decoding",
                journal="{IEEE} Transactions on Information Theory",
                volume=15,
                number=2,
                pages="329--335",
                month=Mar,
                year=1969
              }*)
      nroots := rs_quadratic_solve(_gf,_a,_b xor r,_x);
      (*s may be a triple root if s=_b/_a, but not quadruple, since _a!=0.*)
      if (nroots<>2) or (_x[0]<>s) and (_x[1]<>s) then begin
        _x[nroots{++}] := s;
        Inc(nroots);
      end;
    end;
    Exit(nroots);
  end;
  (*If there are no odd powers, it's really just a quadratic in disguise.*)
  if _c = 0 then begin
    Exit( rs_quadratic_solve(_gf,rs_gsqrt(_gf,_b),rs_gsqrt(_gf,_d),_x) );
  end;
  (*Factor into (x**2 + r*x + s)*(x**2 + r*x + t) by solving for r, which can
       be shown to satisfy r**3 + _b*r + _c == 0.*)
  nroots:= rs_cubic_solve(_gf,0,_b,_c,_x);
  if nroots<1 then begin
    (*The Reed-Solomon code is only valid if we can find 4 distinct roots in
           GF(2**8).
          If the cubic does not factor into 3 (possibly duplicate) roots, then we
           know that the quartic must have a non-trivial irreducible factor.*)
    Exit(0);
  end;
  r := _x[0];
  (*Now solve for s and t.*)
  b := rs_gdiv(_gf,_c,r);
  nroots := rs_quadratic_solve(_gf,b,_d,_x);
  if nroots<2 then Exit(0);
  s := _x[0];
  t := _x[1];
  (*_c=r*(s^t) was non-zero, so s and t must be distinct.
      But if z is a root of z**2 ^ r*z ^ s, then so is (z^r), and s = z*(z^r).
      Hence if z is also a root of z**2 ^ r*z ^ t, then t = s, a contradiction.
      Thus all four roots are distinct, if they exist.*)
  nroots := rs_quadratic_solve(_gf,r,s,_x);
  Result := nroots+rs_quadratic_solve(_gf,r,t,_x+nroots);
end;


(*Polynomial arithmetic with coefficients in GF(2**8).*)

procedure rs_poly_zero(_p: pbyte;  _dp1: integer);
begin
  FillByte(_p^, _dp1*sizeof(_p^), 0);
end;

procedure rs_poly_copy(_p, _q: pbyte;  _dp1: integer);
begin
  Move(_q^, _p^, _dp1*sizeof(_p^));
end;

(*Multiply the polynomial by the free variable, x (shift the coefficients).
  The number of coefficients, _dp1, must be non-zero.*)
procedure rs_poly_mul_x(_p, _q: pbyte;  _dp1: integer);
begin
  Move(_q^, _p[1], (_dp1-1)*sizeof(_p^));
  _p[0] := 0;
end;

(*Divide the polynomial by the free variable, x (shift the coefficients).
  The number of coefficients, _dp1, must be non-zero.*)
procedure rs_poly_div_x(_p, _q: pbyte;  _dp1: integer);
begin
  Move(_q[1], _p^, (_dp1-1)*sizeof(_p^));
  _p[_dp1-1] := 0;
end;

(*Compute the first (d+1) coefficients of the product of a degree e and a
   degree f polynomial.*)
procedure rs_poly_mult(_gf: prs_gf256;  _p: pbyte;  _dp1: integer;  _q: pbyte;  _ep1: integer;  _r: pbyte;  _fp1: integer);
var
  m,i, n,j: integer;
  logqi: LongWord;
begin
  rs_poly_zero(_p, _dp1);
  if _ep1<_dp1 then
    m := _ep1
  else
    m :=_dp1;

  for i:=0 to Pred(m) do
    if _q[i]<>0 then begin
      if _dp1-i<_fp1 then
        n := _dp1-i
      else
        n := _fp1;
      logqi := _gf^.log[_q[i]];

      for j:=0 to Pred(n) do
        _p[i+j] := _p[i+j] xor rs_hgmul(_gf,_r[j], logqi);
    end;
end;

(*Decoding.*)

(*Computes the syndrome of a codeword.*)
procedure rs_calc_syndrome(_gf: prs_gf256;  _m0: integer;  _s: pbyte;  _npar: integer;  _data: pbyte;  _ndata: integer);
var
  i, j: integer;
  alphaj, sj: LongWord;
begin
  for j:=0 to Pred(_npar) do begin
    sj := 0;
    alphaj := _gf^.log[_gf^.exp[j+_m0]];
    for i:=0 to Pred(_ndata) do
      sj := _data[i] xor rs_hgmul(_gf, sj, alphaj);
    _s[j] := sj;
  end;
end;

(*Berlekamp-Peterson and Berlekamp-Massey Algorithms for error-location,
   modified to handle known erasures, from \cite{CC81}, p. 205.
  This finds the coefficients of the error locator polynomial.
  The roots are then found by looking for the values of alpha**n where
   evaluating the polynomial yields zero.
  Error correction is done using the error-evaluator equation on p. 207.
  @BOOK{CC81,
    author="George C. Clark, Jr and J. Bibb Cain",
    title="Error-Correction Coding for Digitial Communications",
    series="Applications of Communications Theory",
    publisher="Springer",
    address="New York, NY",
    month=Jun,
    year=1981
  }*)
(*Initialize lambda to the product of (1-x*alpha**e[i]) for erasure locations
   e[i].
  Note that the user passes in array indices counting from the beginning of the
   data, while our polynomial indexes starting from the end, so
   e[i]=(_ndata-1)-_erasures[i].*)
procedure rs_init_lambda(_gf: prs_gf256;  _lambda: pbyte;  _npar: integer;  _erasures: pbyte;  _nerasures: integer;  _ndata: integer);
var
  n, i, j: integer;
begin
  if _npar < 4 then
    n := 4
  else
    n := _npar;
  rs_poly_zero(_lambda, n+1);
  _lambda[0] := 1;
  for i:=0 to Pred(_nerasures) do
    for j:=i+1 downto Succ(0) do
      _lambda[j] := _lambda[j] xor rs_hgmul(_gf, _lambda[j-1], _ndata-1-_erasures[i]);
end;

(*From \cite{CC81}, p. 216.
  Returns the number of errors detected (degree of _lambda).*)
function rs_modified_berlekamp_massey(_gf: prs_gf256;  _lambda: pbyte;  _s: pbyte;  _omega: pbyte;  _npar: integer;  _erasures: pbyte;  _nerasures: integer;  _ndata: integer): integer;
var
  tt: array [0..255] of Byte;
  n, l, k, i, t: integer;
  d, logd,tti: LongWord;
begin
  (*Initialize _lambda, the error locator-polynomial, with the location of
       known erasures.*)
  rs_init_lambda(_gf, _lambda, _npar, _erasures, _nerasures, _ndata);
  rs_poly_copy(tt, _lambda, _npar+1);
  l := _nerasures;
  k := 0;
  for n:=_nerasures+1 to _npar do begin
    rs_poly_mul_x(tt,tt,n-k+1);
    d:= 0; 
    for i:=0 to l do
      d:= d xor (rs_gmul(_gf,_lambda[i],_s[n-1-i]));
    if d<>0 then begin
      logd := _gf^.log[d];
      if l<n-k then begin
        for i:=0 to n-k do begin
          tti := tt[i];
          tt[i] := rs_hgmul(_gf,_lambda[i],255-logd);
          _lambda[i] := _lambda[i] xor rs_hgmul(_gf,tti,logd);
        end;
        t := n-k;
        k := n-l;
        l := t;
      end
      else
        for i:=0 to l do
          _lambda[i] := _lambda[i] xor rs_hgmul(_gf,tt[i],logd);
    end;
  end;
  rs_poly_mult(_gf, _omega, _npar, _lambda, l+1, _s, _npar);
  Result := l;
end;

(*Finds all the roots of an error-locator polynomial _lambda by evaluating it
   at successive values of alpha, and returns the positions of the associated
   errors in _epos.
  Returns the number of valid roots identified.*)
function rs_find_roots(_gf: prs_gf256;  _epos: pbyte;  _lambda: pbyte;  _nerrors: integer;  _ndata: integer): integer;
var
  alpha, alphai, sum: LongWord;
  nroots, i: integer;
begin
  nroots := 0;
  if _nerrors<=4 then begin
    (*Explicit solutions for higher degrees are possible.
          Chien uses large lookup tables to solve qLongWordics, and Truong et al. give
           special algorithms for degree up through 11, though they use exhaustive
           search (with reduced complexity) for some portions.
          Quartics are good enough for reading CDs, and represent a reasonable code
           complexity trade-off without requiring any extra tables.
          Note that _lambda[0] is always 1.*)
    _nerrors := rs_quartic_solve(_gf,_lambda[1],_lambda[2],_lambda[3],_lambda[4],_epos);
    for i:=0 to Pred(_nerrors) do
      if _epos[i]<>0 then begin
        alpha := _gf^.log[_epos[i]];
        if integer(alpha) < _ndata then begin
          _epos[nroots] := alpha;
          Inc(nroots);
        end;
    end;
  end
  else begin
    for alpha:=0 to Pred(_ndata) do begin // for(alpha=0;(int)alpha<_ndata;alpha++){
      sum := 0;
      alphai := 0;
      for i:=0 to _nerrors do begin
        sum := sum xor rs_hgmul(_gf, _lambda[_nerrors-i], alphai);
        alphai := _gf^.log[_gf^.exp[alphai+alpha]];
      end;
      if sum = 0 then begin
        _epos[nroots{++}] := alpha;
        Inc(nroots);
      end;
    end;
  end;
  Result := nroots;
end;

(*Corrects a codeword with _ndata<256 bytes, of which the last _npar are parity
   bytes.
  Known locations of errors can be passed in the _erasures array.
  Twice as many (up to _npar) errors with a known location can be corrected
   compared to errors with an unknown location.
  Returns the number of errors corrected if successful, or a negative number if
   the message could not be corrected because too many errors were detected.*)
function rs_correct(_gf: prs_gf256;  _m0: integer;  _data: pbyte;  _ndata: integer;  _npar: integer;  _erasures: pbyte;  _nerasures: integer): integer;
var
  (*lambda must have storage for at least five entries to avoid special cases
         in the low-degree polynomial solver.*)
  lambda, omega, epos, s: array [0..255] of Byte;
  i, j, nerrors: integer;
  a, b, alpha, alphan1, alphan2, alphanj: LongWord;
begin
  (*If we already have too many erasures, we can't possibly succeed.*)
  if _nerasures>_npar then Exit(-1);

  (*Compute the syndrome values.*)
  rs_calc_syndrome(_gf,_m0,s,_npar,_data,_ndata);

  (*Check for a non-zero value.*)
  i := 0;
  while i < _npar do begin
    if s[i]<>0 then begin
      (*Construct the error locator polynomial.*)
      nerrors := rs_modified_berlekamp_massey(_gf,lambda,s,omega,_npar,_erasures,_nerasures,_ndata);
      (*If we can't locate any errors, we can't force the syndrome values to
          zero, and must have a decoding error.
        Conversely, if we have too many errors, there's no reason to even attempt
          the root search.*)
      if (nerrors<=0) or (nerrors-_nerasures > (_npar-_nerasures) shr 1) then Exit(-1);
      (*Compute the locations of the errors.
         If they are not all distinct, or some of them were outside the valid
         range for our block size, we have a decoding error.*)
      if rs_find_roots(_gf,epos,lambda,nerrors,_ndata) < nerrors then Exit(-1);
      (*Now compute the error magnitudes.*)

      for i:=0 to Pred(nerrors) do begin
        alpha := epos[i];
        (*Evaluate omega at alpha**-1.*)
        a := 0;
        alphan1 := 255-alpha;
        alphanj := 0;
        for j:=0 to Pred(_npar) do begin
          a := a xor (rs_hgmul(_gf, omega[j], alphanj));
          alphanj := _gf^.log[_gf^.exp[alphanj+alphan1]];
        end;
        (*Evaluate the derivative of lambda at alpha**-1
                All the odd powers vanish.*)
        b := 0;
        alphan2 := _gf^.log[_gf^.exp[alphan1 shl 1]];
        alphanj := alphan1+_m0*alpha mod 255;
        j := 1;
        while j<=_npar do begin
          b := b xor (rs_hgmul(_gf,lambda[j],alphanj));
          alphanj := _gf^.log[_gf^.exp[alphanj+alphan2]];
          Inc(j, 2);
        end;
        (*Apply the correction.*)
        _data[_ndata-1-alpha] := _data[_ndata-1-alpha] xor rs_gdiv(_gf,a,b);
      end;
      Exit(nerrors);
    end;
    Inc(i);
  end;
  Result := 0;
end;


(*Encoding.*)

(*Create an _npar-coefficient generator polynomial for a Reed-Solomon code
   with _npar<256 parity bytes.*)
procedure rs_compute_genpoly(_gf: prs_gf256;  _m0: integer;  _genpoly: pbyte;  _npar: integer);
var
  i,j,n: integer;
  alphai: LongWord;
begin
  if _npar<=0 then Exit;
  rs_poly_zero(_genpoly, _npar);
  _genpoly[0]:= 1;
  (*Multiply by (x+alpha^i) for i = 1 ... _ndata.*)
  for i:=0 to Pred(_npar) do begin
    if i+1 < _npar-1 then
      n := i+1
    else
      n := _npar-1;
    alphai := _gf^.log[_gf^.exp[_m0+i]];
    for j:=n downto 1 do
      _genpoly[j] := _genpoly[j-1] xor rs_hgmul(_gf, _genpoly[j], alphai);
    _genpoly[0] := rs_hgmul(_gf, _genpoly[0], alphai);
  end;
end;

(*Adds _npar<=_ndata parity bytes to an _ndata-_npar byte message.
  _data must contain room for _ndata<256 bytes.*)
procedure rs_encode(_gf: prs_gf256;  _data: pbyte;  _ndata: integer;  _genpoly: pbyte;  _npar: integer);
var
  lfsr: pbyte;
  i,j: integer;
  d,logd: LongWord;
begin
  if _npar<=0 then Exit;
  lfsr := _data+_ndata-_npar;
  rs_poly_zero(lfsr, _npar);
  for i:=0 to Pred(_ndata-_npar) do begin
    d := _data[i] xor lfsr[0];
    if d<>0 then begin
      logd := _gf^.log[d];
      for j:=0 to Pred(_npar-1) do
        lfsr[j] := lfsr[j+1] xor rs_hgmul(_gf, _genpoly[_npar-1-j], logd);
      lfsr[_npar-1] := rs_hgmul(_gf, _genpoly[0], logd);
    end
    else
      rs_poly_div_x(lfsr, lfsr, _npar);
  end;
end;

end.
