unit zbar_qrdec;
(*Copyright (C) 2008-2009  Timothy B. Terriberry (tterribe@xiph.org)
  You can redistribute this library and/or modify it under the terms of the
   GNU Lesser General Public License as published by the Free Software
   Foundation; either version 2.1 of the License, or (at your option) any later
   version.*)

(* C to Pascal translation by Ladislav Karrach in 2018 (with help of c2pas32)
 * Call chain in zbar library:
 * ---------------------------
 * zbar_processor_create
 *   zbar_image_scanner_create
 *     _zbar_qr_create
 *       qr_reader_init -> isaac_init a rs_gf256_init (reader.gf)
 *
 * zbar_process_image -> _zbar_process_image
 *   zbar_scan_image
 *     _zbar_qr_reset
 *     zbar_scanner_new_scan -> zbar_decoder_new_scan
 *     zbar_scan_y
 *     _zbar_qr_decode
 *       qr_finder_centers_locate
 *       qr_binarize
 *       qr_reader_match_centers
 *         qr_reader_try_configuration
 *           (qr_finder_estimate_module_size_and_version)
 *           (qr_hom_fit)
 *             (qr_alignment_pattern_search)
 *           qr_finder_fmt_info_decode
 *           (qr_finder_version_decode)
 *       --> qr_code_decode
 *             qr_sampling_grid_init
 *               qr_sampling_grid_fp_mask_rect
 *               (qr_alignment_pattern_search)
 *             qr_sampling_grid_sample
 *             qr_samples_unpack
 *             rs_correct
 *             qr_code_data_parse
 *       qr_code_data_list_extract_text
 *
 * Usage:
 * ------
 * qimg := GetMem(w*w);                  // binary matrix for QR code of size "w" in black/white modules
 * qr_reader_init(qreader);
 * qdata.bbox[0][0]:=2;                  // top-left
 * qdata.bbox[0][1]:=2;
 * qdata.bbox[1][0]:=2+(w-1)*4;       // top-right
 * qdata.bbox[1][1]:=2;
 * qdata.bbox[2][0]:=2;                  // bottom-left
 * qdata.bbox[2][1]:=2+(w-1)*4;
 * qdata.bbox[3][0]:=2+(w-1)*4;       // bottom-right
 * qdata.bbox[3][1]:=2+(w-1)*4;
 * qr_hom_init(qhom, qdata.bbox[0][0],qdata.bbox[0][1], qdata.bbox[1][0],qdata.bbox[1][1], qdata.bbox[2][0],qdata.bbox[2][1], qdata.bbox[3][0],qdata.bbox[3][1], QR_HOM_BITS);
 * qr_finder_init(qfp1, @qhom, 3*4+2, 3*4+2);       // top-left finder pattern
 * qr_finder_init(qfp2, @qhom, (w-3-1)*4+2, 3*4+2); // top-right
 * qr_finder_init(qfp3, @qhom, 3*4+2, (w-3-1)*4+2); // bottom-left
 * for y:=0 to w - 1 do
 *   for x:=0 to w - 1 do
 *     qimg[x+y*w] := Ord(Image.Pixels[x,y] < gcGray); // 0-white, 1-black
 * qfmt_info := qr_finder_fmt_info_decode(@qfp1, @qfp2, @qfp3, @qhom, qimg, w, w);
 * if qfmt_info > 0 then begin
 *   if qr_code_decode(qdata, @qreader.gf, qfp1.c^.pos, qfp2.c^.pos, qfp3.c^.pos, (w-17) shr 2, qfmt_info, qimg, w, w) = 0 then
 *     ...                         // data are in qdata.entries
 *     qr_code_data_clear(qdata);  // finally free
 *)

{ $DEFINE HAVE_zbar_qr_decode}
{ $DEFINE DEBUG_ZBAR}

interface

uses
  zbar_qrcode, zbar_rs, zbar_bch15_5, zbar_isaac;

type
  uint32_t = DWord;
  unsigned = LongWord;
  punsigned = ^unsigned;

const
  CHAR_BIT = 8;
  QR_INT_BITS = SizeOf(integer)*CHAR_BIT; // 32
  QR_INT_LOGBITS = 5; //(QR_ILOG(QR_INT_BITS));

  (*A 14 bit resolution for a homography ensures that the ideal module size for a
   version 40 code differs from that of a version 39 code by at least 2.*)
  QR_HOM_BITS = 14;

  (*The number of bits of sub-module precision to use when searching for
     alignment patterns.
    Two bits allows an alignment pattern to be found even if the modules have
     been eroded by up to 50% (due to blurring, etc.).
    This must be at least one, since it affects the dynamic range of the
     transforms, and we sample at half-module resolution to compute a bounding
     quadrilateral for the code.*)
  QR_ALIGN_SUBPREC = 2;

type
  qr_mode = (
    QR_MODE_TERM=0,
    (*Numeric digits ('0'...'9').*)
    QR_MODE_NUM=1,
    (*Alphanumeric characters ('0'...'9', 'A'...'Z', plus the punctuation
     ' ', '$', '%', '*', '+', '-', '.', '/', ':').*)
    QR_MODE_ALNUM,
    (*Structured-append header.*)
    QR_MODE_STRUCT,
    (*Raw 8-bit bytes.*)
    QR_MODE_BYTE,
    (*FNC1 marker (for more info, see http://www.mecsw.com/specs/uccean128.html).
    In the "first position" data is formatted in accordance with GS1 General
     Specifications.*)
    QR_MODE_FNC1_1ST,
    (*Mode 6 reserved?*)
    (*Extended Channel Interpretation code.*)
    QR_MODE_ECI=7,
    (*SJIS kanji characters.*)
    QR_MODE_KANJI,
    (*FNC1 marker (for more info, see http://www.mecsw.com/specs/uccean128.html).
    In the "second position" data is formatted in accordance with an industry
     application as specified by AIM Inc.*)
    QR_MODE_FNC1_2ND,
    QR_MODE_INVALID
  );

  (*A single unit of parsed QR code data.*)
  qr_code_data_entry = record
    (*The mode of this data block.*)
    mode: qr_mode;
    payload: record
      (*Data buffer for modes that have one.*)
      data: record
        buf: PByte;
        len: integer;
      end;
      (*Decoded "Extended Channel Interpretation" data.*)
      eci: LongWord;
      (*Decoded "Application Indicator" for FNC1 in 2nd position.*)
      ai: integer;
      (*Structured-append header data.*)
      sa: record
        sa_index, sa_size, sa_parity: unsigned;
      end;
    end;
  end;
  pqr_code_data_entry = ^qr_code_data_entry;

  (*Low-level QR code data.*)
  qr_code_data = record
    (*The decoded data entries.*)
    entries: pqr_code_data_entry;
    nentries: integer;
    (*The code version (1...40).*)
    version: byte;
    (*The ECC level (0...3, corresponding to 'L', 'M', 'Q', and 'H').*)
    ecc_level: byte;
    (*Structured-append information.*)
    (*The index of this code in the structured-append group.
      If sa_size is zero, this is undefined.*)
    sa_index: byte;
    (*The size of the structured-append group, or 0 if there was no S-A header.*)
    sa_size: byte;
    (*The parity of the entire structured-append group.
      If sa_size is zero, this is undefined.*)
    sa_parity: byte;
    (*The parity of this code.
      If sa_size is zero, this is undefined.*)
    self_parity: byte;
    (*An approximate bounding box for the code.
      Points appear in the order up-left, up-right, down-left, down-right,
      relative to the orientation of the QR code.*)
    bbox: array[0..3] of qr_point;
  end;
  pqr_code_data = ^qr_code_data;

  qr_code_data_list = record
    qrdata: pqr_code_data;
    nqrdata: integer;
    cqrdata: integer;
  end;
  pqr_code_data_list = ^qr_code_data_list;

  qr_finder_lines = record
    lines: pqr_finder_line;
    nlines: integer;
    clines: integer;
  end;
  pqr_finder_lines = ^qr_finder_lines;

  (*A cluster of lines crossing a finder pattern (all in the same direction).*)
  qr_finder_cluster = record
    (*Pointers to the lines crossing the pattern.*)
    lines: ^pqr_finder_line;
    (*The number of lines in the cluster.*)
    nlines: integer;
  end;
  pqr_finder_cluster = ^qr_finder_cluster;
  ppqr_finder_cluster = ^pqr_finder_cluster;

  (*A point on the edge of a finder pattern.
    These are obtained from the endpoints of the lines crossing this particular
     pattern.*)
  qr_finder_edge_pt = record
    (*The location of the edge point.*)
    pos: qr_point;
    (*A label classifying which edge this belongs to:
      0: negative u edge (left)
      1: positive u edge (right)
      2: negative v edge (top)
      3: positive v edge (bottom)*)
    edge: integer;
    (*The (signed) perpendicular distance of the edge point from a line parallel
       to the edge passing through the finder center, in (u,v) coordinates.
      This is also re-used by RANSAC to store inlier flags.*)
    extent: integer;
  end;
  pqr_finder_edge_pt = ^qr_finder_edge_pt;

  (*The center of a finder pattern obtained from the crossing of one or more
    clusters of horizontal finder lines with one or more clusters of vertical
    finder lines.*)
  qr_finder_center = record
    (*The estimated location of the finder center.*)
    pos: qr_point;
    (*The list of edge points from the crossing lines.*)
    edge_pts: pqr_finder_edge_pt;
    (*The number of edge points from the crossing lines.*)
    nedge_pts: integer;
  end;
  pqr_finder_center = ^qr_finder_center;
  ppqr_finder_center = ^pqr_finder_center;

  (*All the information we've collected about a finder pattern in the current
    configuration.*)
  qr_finder = record
    (*The module size along each axis (in the square domain).*)
    size: array [0..1] of integer;
    (*The version estimated from the module size along each axis.*)
    eversion: array[0..1] of integer;
    (*The list of classified edge points for each edge.*)
    edge_pts: array[0..3] of pqr_finder_edge_pt;
    (*The number of edge points classified as belonging to each edge.*)
    nedge_pts: array [0..3] of integer;
    (*The number of inliers found after running RANSAC on each edge.*)
    ninliers: array[0..3] of integer;
    (*The center of the finder pattern (in the square domain).*)
    o: qr_point;
    (*The finder center information from the original image.*)
    c: pqr_finder_center;
  end;
  pqr_finder = ^qr_finder;

  (*A full homography.
    Like the affine homography, this maps from the image (at subpel resolution)
     to a square domain with power-of-two sides (of res bits) and back.*)
  qr_hom = record
    fwd, inv: array[0..2, 0..1] of integer;
    fwd22, inv22: integer;
    x0, y0: integer;
    res: integer;
  end;
  pqr_hom = ^qr_hom;

  qr_reader = record
    (*The GF(256) representation used in Reed-Solomon decoding.*)
    gf: rs_gf256;
    (*The random number generator used by RANSAC.*)
    isaac: isaac_ctx;
    (* current finder state, horizontal and vertical lines *)
    finder_lines: array [0..1] of qr_finder_lines;
  end;
  pqr_reader = ^qr_reader;


procedure qr_reader_init(out reader: qr_reader);
procedure qr_hom_init(out _hom: qr_hom; _x0,_y0, _x1,_y1, _x2,_y2, _x3,_y3, _res: integer);
procedure qr_finder_init(out fp: qr_finder; _hom: pqr_hom; _x,_y: integer);
procedure qr_finder_clear(var fp: qr_finder);

// 18 bit version information (version>=7 only); 6 bit binary string that encodes the QR version, followed by a string of 12 error correction bits
function qr_finder_version_decode(_f: pqr_finder;  _hom: pqr_hom;  _img: pbyte;  _width, _height: integer;  _dir: integer): integer;
// 15 bit format information; 2 bits error correction level and 3 bits the mask pattern
function qr_finder_fmt_info_decode(_ul, _ur, _dl: pqr_finder;  _hom: pqr_hom;  _img: pbyte;  _width, _height: integer): integer;

function qr_code_decode(var _qrdata: qr_code_data;  _gf: prs_gf256;  const _ul_pos, _ur_pos, _dl_pos: qr_point;  _version, _fmt_info: integer;  _img: pbyte;  _width, _height: integer): integer;
procedure qr_code_data_clear(var _qrdata: qr_code_data);

{$IFDEF HAVE_zbar_qr_decode}
type
  refcnt_t = integer;

  (**********
   * zbar.h *
   **********)
  (* "color" of element: bar or space. *)
  zbar_color_t = (
      ZBAR_SPACE = 0,    (*< light area or space between bars *)
      ZBAR_BAR = 1       (*< dark area or colored bar segment *)
  );

  (* decoded symbol type. *)
  zbar_symbol_type_t = (
     ZBAR_NONE        =      0,  (**< no symbol decoded *)
     ZBAR_PARTIAL     =      1,  (**< intermediate status *)
     ZBAR_EAN2        =      2,  (**< GS1 2-digit add-on *)
     ZBAR_EAN5        =      5,  (**< GS1 5-digit add-on *)
     ZBAR_EAN8        =      8,  (**< EAN-8 *)
     ZBAR_UPCE        =      9,  (**< UPC-E *)
     ZBAR_ISBN10      =     10,  (**< ISBN-10 (from EAN-13). @since 0.4 *)
     ZBAR_UPCA        =     12,  (**< UPC-A *)
     ZBAR_EAN13       =     13,  (**< EAN-13 *)
     ZBAR_ISBN13      =     14,  (**< ISBN-13 (from EAN-13). @since 0.4 *)
     ZBAR_COMPOSITE   =     15,  (**< EAN/UPC composite *)
     ZBAR_I25         =     25,  (**< Interleaved 2 of 5. @since 0.4 *)
     ZBAR_DATABAR     =     34,  (**< GS1 DataBar (RSS). @since 0.11 *)
     ZBAR_DATABAR_EXP =     35,  (**< GS1 DataBar Expanded. @since 0.11 *)
     ZBAR_CODABAR     =     38,  (**< Codabar. @since 0.11 *)
     ZBAR_CODE39      =     39,  (**< Code 39. @since 0.4 *)
     ZBAR_PDF417      =     57,  (**< PDF417. @since 0.6 *)
     ZBAR_QRCODE_     =     64,  (**< QR Code. @since 0.10 *)
     ZBAR_CODE93      =     93,  (**< Code 93. @since 0.11 *)
     ZBAR_CODE128     =    128,  (**< Code 128 *)

     (** mask for base symbol type.
      * @deprecated in 0.11, remove this from existing code
      *)
     ZBAR_SYMBOL      = $00ff,
     (** 2-digit add-on flag.
      * @deprecated in 0.11, a ::ZBAR_EAN2 component is used for
      * 2-digit GS1 add-ons
      *)
     ZBAR_ADDON2      = $0200,
     (** 5-digit add-on flag.
      * @deprecated in 0.11, a ::ZBAR_EAN5 component is used for
      * 5-digit GS1 add-ons
      *)
     ZBAR_ADDON5      = $0500,
     (** add-on flag mask.
      * @deprecated in 0.11, GS1 add-ons are represented using composite
      * symbols of type ::ZBAR_COMPOSITE; add-on components use ::ZBAR_EAN2
      * or ::ZBAR_EAN5
      *)
     ZBAR_ADDON       = $0700
  );

  (* decoded symbol coarse orientation. @since 0.11 *)
  zbar_orientation_t = (
     ZBAR_ORIENT_UNKNOWN = -1,   (**< unable to determine orientation *)
     ZBAR_ORIENT_UP,             (**< upright, read left to right *)
     ZBAR_ORIENT_RIGHT,          (**< sideways, read top to bottom *)
     ZBAR_ORIENT_DOWN,           (**< upside-down, read right to left *)
     ZBAR_ORIENT_LEFT            (**< sideways, read bottom to top *)
  );

  (* error codes. *)
  zbar_error_t = (
     ZBAR_OK = 0,                (**< no error *)
     ZBAR_ERR_NOMEM,             (**< out of memory *)
     ZBAR_ERR_INTERNAL,          (**< internal library error *)
     ZBAR_ERR_UNSUPPORTED,       (**< unsupported request *)
     ZBAR_ERR_INVALID,           (**< invalid request *)
     ZBAR_ERR_SYSTEM,            (**< system error *)
     ZBAR_ERR_LOCKING,           (**< locking error *)
     ZBAR_ERR_BUSY,              (**< all resources busy *)
     ZBAR_ERR_XDISPLAY,          (**< X11 display error *)
     ZBAR_ERR_XPROTO,            (**< X11 protocol error *)
     ZBAR_ERR_CLOSED,            (**< output window is closed *)
     ZBAR_ERR_WINAPI,            (**< windows system error *)
     ZBAR_ERR_NUM                (**< number of error codes *)
  );

  (* decoder configuration options. @since 0.4 *)
  zbar_config_t = (
     ZBAR_CFG_ENABLE = 0,        (**< enable symbology/feature *)
     ZBAR_CFG_ADD_CHECK,         (**< enable check digit when optional *)
     ZBAR_CFG_EMIT_CHECK,        (**< return check digit when present *)
     ZBAR_CFG_ASCII,             (**< enable full ASCII character set *)
     ZBAR_CFG_NUM,               (**< number of boolean decoder configs *)

     ZBAR_CFG_MIN_LEN = $20,     (**< minimum data length for valid decode *)
     ZBAR_CFG_MAX_LEN,           (**< maximum data length for valid decode *)

     ZBAR_CFG_UNCERTAINTY = $40, (**< required video consistency frames *)

     ZBAR_CFG_POSITION = $80,    (**< enable scanner to collect position data *)

     ZBAR_CFG_X_DENSITY = $100,  (**< image scanner vertical scan density *)
     ZBAR_CFG_Y_DENSITY          (**< image scanner horizontal scan density *)
  );

  (* decoder symbology modifier flags. @since 0.11 *)
  zbar_modifier_t = (
     (** barcode tagged as GS1 (EAN.UCC) reserved
      * (eg, FNC1 before first data character).
      * data may be parsed as a sequence of GS1 AIs
      *)
     ZBAR_MOD_GS1 = 0,

     (** barcode tagged as AIM reserved
      * (eg, FNC1 after first character or digit pair)
      *)
     ZBAR_MOD_AIM,

     (** number of modifiers *)
     ZBAR_MOD_NUM
  );

const
  RECYCLE_BUCKETS = 5;
  NUM_SCN_CFGS = Ord(ZBAR_CFG_Y_DENSITY) - Ord(ZBAR_CFG_X_DENSITY) + 1;
  NUM_SYMS = 20;
  (* size of bar width history (implementation assumes power of two) *)
  DECODE_WINDOW = 16;

type
 (************
  * symbol.h *
  ************)
  point_t = record
    x, y: integer;
  end;

  zbar_symbol_set_t = record
    refcnt: refcnt_t;
    nsyms: integer;              (* number of filtered symbols *)
    head: ^zbar_symbol_t;        (* first of decoded symbol results *)
    tail: ^zbar_symbol_t;        (* last of unfiltered symbol results *)
  end;
  pzbar_symbol_set_t = ^zbar_symbol_set_t;

  pzbar_symbol_t = ^zbar_symbol_t;
  zbar_symbol_t = record
      &type: zbar_symbol_type_t;  (* symbol type *)
      configs: unsigned;       (* symbology boolean config bitmask *)
      modifiers: unsigned;     (* symbology modifier bitmask *)
      data_alloc: unsigned;    (* allocation size of data *)
      datalen: unsigned;       (* length of binary symbol data *)
      data: PByte;             (* symbol data *)

      pts_alloc: unsigned;     (* allocation size of pts *)
      npts: unsigned;          (* number of points in location polygon *)
      pts: ^point_t;           (* list of points in location polygon *)
      orient: zbar_orientation_t;  (* coarse orientation *)

      refcnt: refcnt_t;            (* reference count *)
      next: pzbar_symbol_t;        (* linked list of results (or siblings) *)
      syms: pzbar_symbol_set_t;    (* components of composite result *)
      time: unsigned;              (* relative symbol capture time *)
      cache_count: integer;        (* cache state *)
      quality: integer;            (* relative symbol reliability metric *)
  end;

 (***********
  * image.h *
  ***********)
  pzbar_image_t = ^zbar_image_t;
  pzbar_video_t = pointer; // ToDo
  (** cleanup handler callback function.
  * called to free sample data when an image is destroyed.
  *)
  zbar_image_cleanup_handler_t = procedure(image: pzbar_image_t);
  zbar_image_t = record
     format: uint32_t;            (* fourcc image format code *)
     width, height: unsigned;     (* image size *)
     data: pointer;               (* image sample data *)
     datalen: unsigned;           (* allocated/mapped size of data *)
     crop_x, crop_y: unsigned;    (* crop rectangle *)
     crop_w, crop_h: unsigned;
     userdata: pointer;           (* user specified data associated w/image *)

     (* cleanup handler *)
     cleanup: zbar_image_cleanup_handler_t;
     refcnt: refcnt_t;            (* reference count *)
     src: pzbar_video_t;          (* originator *)
     srcidx: integer;             (* index used by originator *)
     next: ^zbar_image_t;         (* internal image lists *)

     seq: unsigned;               (* page/frame sequence number *)
     syms: ^zbar_symbol_set_t;    (* decoded result set *)
  end;

 (***********
  * error.h *
  ***********)
  errmodule_t = (
     ZBAR_MOD_PROCESSOR,
     ZBAR_MOD_VIDEO,
     ZBAR_MOD_WINDOW,
     ZBAR_MOD_IMAGE_SCANNER,
     ZBAR_MOD_UNKNOWN
  );

  errinfo_t = record
     magic: uint32_t;            (* just in case *)
     module: errmodule_t;        (* reporting module *)
     buf: PChar;                 (* formatted and passed to application *)
     errnum: integer;            (* errno for system errors *)

     //errsev_t sev;
     //zbar_error_t type;
     //const char *func;           /* reporting function */
     //const char *detail;         /* description */
     arg_str: PAnsiChar;           (* single string argument *)
     //int arg_int;                /* single integer argument */
  end;

  (***************
   * qr_finder.h *
   ***************)
  (* QR Code symbol finder state *)
  qr_finder_t = record
      s5: unsigned;                (* finder pattern width *)
      line: qr_finder_line ;       (* position info needed by decoder *)

      config: unsigned;
  end;
  pqr_finder_t = ^qr_finder_t;

  (*************
   * decoder.h *
   *************)
  pzbar_decoder_t = ^zbar_decoder_t;
  (** decoder data handler callback function.
    * called by decoder when new data has just been decoded *)
  zbar_decoder_handler_t = procedure(decoder: pzbar_decoder_t);
  (* symbology independent decoder state *)
  zbar_decoder_t = record
    idx: byte;                          (* current width index *)
    w: array[0..DECODE_WINDOW-1] of unsigned; (* window of last N bar widths *)
    &type: zbar_symbol_type_t;          (* type of last decoded data *)
    lock: zbar_symbol_type_t;           (* buffer lock *)
    modifiers: unsigned;                (* symbology modifier *)
    direction: integer;                 (* direction of last decoded data *)
    s6: unsigned;                       (* 6-element character width *)

    (* everything above here is automatically reset *)
    buf_alloc: unsigned;                (* dynamic buffer allocation *)
    buflen: unsigned;                   (* binary data length *)
    buf: PByte;                         (* decoded characters *)
    userdata: pointer;                  (* application data *)
    handler: zbar_decoder_handler_t;    (* application callback *)

    (* symbology specific state *)
    qrf: qr_finder_t;                    (* QR Code finder state *)
  end;

  (*************
   * scanner.c *
   *************)
  (* scanner state *)
  zbar_scanner_t = record
      decoder: pzbar_decoder_t;   (* associated bar width decoder *)
      y1_min_thresh: unsigned;    (* minimum threshold *)

      x: unsigned;                (* relative scan position of next sample *)
      y0: array[0..3] of integer; (* short circular buffer of average intensities *)

      y1_sign: integer;           (* slope at last crossing *)
      y1_thresh: unsigned;        (* current slope threshold *)

      cur_edge: unsigned;         (* interpolated position of tracking edge *)
      last_edge: unsigned;        (* interpolated position of last located edge *)
      width: unsigned;            (* last element width *)
  end;
  pzbar_scanner_t = ^zbar_scanner_t;

 (*****************
  * img_scanner.c *
  *****************)
  recycle_bucket_t = record
    nsyms: integer;
    head: ^zbar_symbol_t;
  end;
  precycle_bucket_t = ^recycle_bucket_t;

 (** data handler callback function.
   * called when decoded symbol results are available for an image *)
  zbar_image_data_handler_t = procedure(image: pzbar_image_t; userdata: pointer);

  (* image scanner state *)
  zbar_image_scanner_t = record
    scn: pzbar_scanner_t;         (* associated linear intensity scanner *)
    dcode: pzbar_decoder_t;       (* associated symbol decoder *)
    qr: ^qr_reader;               (* QR Code 2D reader *)

    userdata: pointer;            (* application data *)
    (* user result callback *)
    handler: zbar_image_data_handler_t;

    time: unsigned;               (* scan start time *)
    img: ^zbar_image_t;           (* currently scanning image *root* *)
    dx, dy, du, umin, v: integer; (* current scan direction *)
    syms: ^zbar_symbol_set_t;     (* previous decode results *)
    (* recycled symbols in 4^n size buckets *)
    recycle: array[0..RECYCLE_BUCKETS-1] of recycle_bucket_t;

    enable_cache: integer;        (* current result cache state *)
    cache: ^zbar_symbol_t;        (* inter-image result cache entries *)

    (* configuration settings *)
    config: unsigned;             (* config flags *)
    ean_config: unsigned;
    configs: array[0..NUM_SCN_CFGS-1] of integer;  (* int valued configurations *)
    sym_configs: array[0..0,0..NUM_SYMS-1] of integer; (* per-symbology configurations *)

  {$ifndef NO_STATS}
    stat_syms_new: integer;
    stat_iscn_syms_inuse, stat_iscn_syms_recycle: integer;
    stat_img_syms_inuse, stat_img_syms_recycle: integer;
    stat_sym_new: integer;
    stat_sym_recycle: array[0..RECYCLE_BUCKETS-1] of integer;
  {$endif}
  end;
  pzbar_image_scanner_t = ^zbar_image_scanner_t;

 (***************
  * processor.c *
  ***************)
  zbar_processor_t = record
   err: errinfo_t;                      (* error reporting *)
   userdata: pointer;                   (* application data *)

   //video: ^zbar_video_t;                (* input video device abstraction *)
   //window: ^zbar_window_t;              (* output window abstraction *)
   scanner: ^zbar_image_scanner_t;      (* barcode scanner *)

   handler: zbar_image_data_handler_t;  (* application data handler *)

   req_width, req_height: unsigned;     (* application requested video size *)
   req_intf, req_iomode: integer;       (* application requested interface *)
   force_input: uint32_t;               (* force input format (debug) *)
   force_output: uint32_t;              (* force format conversion (debug) *)

   input: integer;                      (* user input status *)

   (* state flags *)
   threaded: integer;
   visible: integer;                    (* output window mapped to display *)
   streaming: integer;                  (* video enabled *)
   dumping: integer;                    (* debug image dump *)

   display: pointer;                    (* X display connection *)
   xwin: unsigned;                      (* toplevel window *)

   //input_thread: zbar_thread_t;         (* video input handler *)
   //video_thread: zbar_thread_t;         (* window event handler *)

   syms: ^zbar_symbol_set_t;            (* previous decode results *)

   //mutex: zbar_mutex_t;                 (* shared data mutex *)

   (* API serialization lock *)
   //lock_level: int;
   //lock_owner: zbar_thread_id_t;
   //wait_head, wait_tail, wait_next: ^proc_waiter_t;
   //free_waiter: ^proc_waiter_t;

   //state: ^processor_state_t;
  end;
  pzbar_processor_t = ^zbar_processor_t;


function zbar_version(var major, minor: cardinal): integer;
procedure zbar_set_verbosity(level: integer);

function zbar_processor_create(threaded: integer): pzbar_processor_t;
procedure zbar_processor_destroy(proc: pzbar_processor_t);
function zbar_processor_init(processor: pzbar_processor_t; video_device: PAnsiChar; enable_display: Integer): Integer;
function zbar_processor_set_visible(processor: pzbar_processor_t; visible: Integer): Integer;

function zbar_image_create(): pzbar_image_t;
procedure zbar_image_destroy(img: pzbar_image_t);
procedure zbar_image_set_format(img: pzbar_image_t; fmt: PAnsiChar);
procedure zbar_image_set_size(img: pzbar_image_t; w,h: unsigned);
procedure zbar_image_set_data(img: pzbar_image_t; data: Pointer; len: unsigned; cleanup: zbar_image_cleanup_handler_t);

function zbar_process_image(proc: pzbar_processor_t; img: pzbar_image_t): integer;
function zbar_image_first_symbol(img: pzbar_image_t): pzbar_symbol_t;
function zbar_symbol_get_type(sym: pzbar_symbol_t): zbar_symbol_type_t;
function zbar_symbol_get_loc_x(sym: pzbar_symbol_t; idx: unsigned): integer;
function zbar_symbol_get_loc_y (sym: pzbar_symbol_t; idx: unsigned): integer;
function zbar_symbol_get_data(sym: pzbar_symbol_t): PAnsiChar;
function zbar_symbol_next(sym: pzbar_symbol_t): pzbar_symbol_t;
function zbar_get_symbol_name(sym: zbar_symbol_type_t): PAnsiChar;
{$ENDIF}

implementation

{$IFDEF HAVE_zbar_qr_decode}
uses
  zbar_binarize
  {$IF FPC_FULLVERSION > 30202},SortBase{$ENDIF};
{$ENDIF}

type
  (*A homography from one region of the grid back to the image.
    Unlike a qr_hom, this does not include an inverse transform and maps directly
     from the grid points, not a square with power-of-two sides.*)
  qr_hom_cell = record
    fwd: array[0..2,0..2] of integer;
    x0, y0, u0, v0: integer;
  end;
  pqr_hom_cell = ^qr_hom_cell;

  (*The grid used to sample the image bits.
    The grid is divided into separate cells bounded by finder patterns and/or
     alignment patterns, and a separate map back to the original image is
     constructed for each cell.
    All of these structural elements, as well as the timing patterns, version
     info, and format info, are marked in fpmask so they can easily be skipped
     during decode.*)
  qr_sampling_grid = record
    cells: array[0..5] of pqr_hom_cell;
    fpmask: punsigned; // bitmask
    cell_limits: array[0..5] of integer;
    ncells: integer;
  end;
  pqr_sampling_grid = ^qr_sampling_grid;



//#define QR_ILOG4(_v) (((_v)&0xFFFF0000)?16+QR_ILOG3((_v)>>16):QR_ILOG3(_v))
function QR_ILOG4(_v: unsigned): unsigned;
  //#define QR_ILOG0(_v) (!!((_v)&0x2))
  function QR_ILOG0(_v: byte): byte;
  begin
    Result := Ord((_v and $2)<>0);
  end;
  //#define QR_ILOG1(_v) (((_v)&0xC)?2+QR_ILOG0((_v)>>2):QR_ILOG0(_v))
  function QR_ILOG1(_v: byte): byte;
  begin
    if (_v and $C) <> 0 then
      Result := 2+QR_ILOG0(_v shr 2)
    else
      Result := QR_ILOG0(_v);
  end;
  //#define QR_ILOG2(_v) (((_v)&0xF0)?4+QR_ILOG1((_v)>>4):QR_ILOG1(_v))
  function QR_ILOG2(_v: word): word;
  begin
    if (_v and $F0) <> 0 then
      Result := 4+QR_ILOG1(_v shr 4)
    else
      Result := QR_ILOG1(_v);
  end;
  //#define QR_ILOG3(_v) (((_v)&0xFF00)?8+QR_ILOG2((_v)>>8):QR_ILOG2(_v))
  function QR_ILOG3(_v: unsigned): unsigned;
  begin
    if (_v and $FF00) <> 0 then
      Result := 8+QR_ILOG2(_v shr 8)
    else
      Result := QR_ILOG2(_v);
  end;
begin
  if (_v and $FFFF0000) <> 0 then
    Result := 16 + QR_ILOG3(_v shr 16)
  else
    Result := QR_ILOG3(_v);
end;

//#define QR_MAXI(_a,_b)      ((_a)-((_a)-(_b)&-((_b)>(_a))))
function QR_MAXI(_a, _b: integer): integer; inline;
begin
  if _a > _b then
    Result := _a
  else
    Result := _b;
end;

//#define QR_MINI(_a,_b)      ((_a)+((_b)-(_a)&-((_b)<(_a))))
function QR_MINI(_a, _b: integer): integer; inline;
begin
  if _a < _b then
    Result := _a
  else
    Result := _b;
end;

//#define QR_SIGNMASK(_x)     (-((_x)<0))
function QR_SIGNMASK(_x: integer): integer; inline;
begin
  Result := -Ord(_x<0)
end;
function QR_SIGNMASK(_x: int64): int64; inline;
begin
  Result := -Ord(_x<0)
end;

(*Unlike copysign(), simply inverts the sign of _a if _b is negative.*)
//#define QR_FLIPSIGNI(_a,_b) ((_a)+QR_SIGNMASK(_b)^QR_SIGNMASK(_b))
function QR_FLIPSIGNI(_a,_b: integer): integer; inline;
begin
  //Result := (_a + QR_SIGNMASK(_b)) xor QR_SIGNMASK(_b)
  if _b < 0 then
    Result := -_a
  else
    Result := _a;
end;
function QR_FLIPSIGNI(_a,_b: int64): int64; inline;
begin
  if _b < 0 then
    Result := -_a
  else
    Result := _a;
end;

(*Divides a signed integer by a positive value with exact rounding.*)
//#define QR_DIVROUND(_x,_y)  (((_x)+QR_FLIPSIGNI(_y>>1,_x))/(_y))
function QR_DIVROUND(_x,_y: integer): integer;
begin
  Result := (_x + QR_FLIPSIGNI(_y div 2,_x)) div _y;
end;
function QR_DIVROUND(_x,_y: int64): integer;
var r: int64;
begin
  r := (_x + QR_FLIPSIGNI(_y div 2,_x)) div _y;
  Assert((r>=Low(integer)) and (r<=High(integer)));
  Result := r;
end;

//#define QR_CLAMPI(_a,_b,_c) (QR_MAXI(_a,QR_MINI(_b,_c)))
function QR_CLAMPI(_a,_b,_c: integer): integer; inline;
begin
  if _b < _c then
    Result := _b
  else
    Result := _c;
  if Result < _a then
    Result := _a;
end;

procedure QR_SWAP2I(var _a,_b: integer); inline;
var t: integer;
begin
  t := _a;
  _a := _b;
  _b := t;
end;

(*Swaps two integers _a and _b if _a>_b.*)
procedure QR_SORT2I(var _a,_b: integer);
var t: integer;
begin
  t := QR_MINI(_a,_b) xor _a;
  _a := _a xor t;
  _b := _b xor t;
end;

(*Multiplies 32-bit numbers _a and _b, adds (possibly 64-bit) number _r, and
   takes bits [_s,_s+31] of the result.*)
//#define QR_FIXMUL(_a,_b,_r,_s) ((int)((_a)*(long long)(_b)+(_r)>>(_s)))
function QR_FIXMUL(_a,_b: integer; _r: int64; _s: integer): integer; inline;
var b: int64;
begin
  b := _b;
  Result := integer((_a*b + _r) div (1 shl _s));
end;

(*Multiplies 32-bit numbers _a and _b, adds (possibly 64-bit) number _r, and gives all 64 bits of the result.*)
// #define QR_EXTMUL(_a,_b,_r)    ((_a)*(long long)(_b)+(_r))
function QR_EXTMUL(_a,_b: integer; _r: int64): int64; inline;
var b: int64;
begin
  b := _b;
  Result := _a*b + _r;
end;

function qr_ilog(_v: unsigned): integer;
var m: integer;
begin
  m := {!!}Ord((_v and $FFFF0000)<>0) shl 4;
  _v := _v shr m;
  Result := m;
  m := {!!}Ord((_v and $FF00)<>0) shl 3;
  _v := _v shr m;
  Result := Result or m;
  m := Ord((_v and $F0)<>0) shl 2;
  _v := _v shr m;
  Result := Result or m;
  m := Ord((_v and $0C)<>0) shl 1;
  _v := _v shr m;
  Result := Result or m;
  Result := Result or {!!}Ord((_v and $2)<>0);
  Result := Result + {!!}Ord(_v<>0);
  //DebugLn('QR_ILOG: %d = %d', [__v, Result]);
end;


(*Initializes a client reader handle.*)
procedure qr_reader_init(out reader: qr_reader);
begin
  FillByte(reader, SizeOf(reader), 0);
  isaac_init(reader.isaac, nil, 0);
  rs_gf256_init(reader.gf, QR_PPOLY);
end;


(*Computes the Hamming distance between two bit patterns (the number of bits
   that differ).
  May stop counting after _maxdiff differences.*)
function qr_hamming_dist(_y1,_y2: unsigned;  _maxdiff: integer): integer;
var
  y: unsigned;
begin
  y := _y1 xor _y2;
  Result := 0;
  while (Result < _maxdiff) and (y<>0) do begin
    y := y and (y-1);
    Inc(Result);
  end;
end;


(*Retrieve a bit (guaranteed to be 0 or 1) from the image, given coordinates in
   subpel resolution which have not been bounds checked.*)
function qr_img_get_bit(_img: pbyte;  _width, _height: integer;  _x, _y: integer): unsigned;
begin
  _x := _x div (1 shl QR_FINDER_SUBPREC);
  _y := _y div (1 shl QR_FINDER_SUBPREC);
  Result := Ord(_img[QR_CLAMPI(0,_y,_height-1)*_width + QR_CLAMPI(0,_x,_width-1)] <> 0); // !=0
end;



{$IFDEF HAVE_zbar_qr_decode}
procedure qr_hom_init(out _hom: qr_hom; _x0,_y0, _x1,_y1, _x2,_y2, _x3,_y3, _res: integer);
var
  dx10, dx20, dx30, dx31, dx32, dy10, dy20, dy30, dy31, dy32, a20, a21, a22, b0, b1, b2, s1, s2, r1, r2: integer;
begin
  // stredy rohovych modulov
  // 0(TL)--1(TR)
  // |      |
  // 2(BL)--3(BR)
  // x,y suradnice su vynasobene QR_FINDER_SUBPREC
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_hom_init: (%d,%d)-(%d,%d)-(%d,%d)-(%d,%d); (%d,%d)-(%d,%d)-(%d,%d)-(%d,%d); res=%d', [_x0,_y0, _x1,_y1, _x2,_y2, _x3,_y3, _x0 shr QR_FINDER_SUBPREC,_y0 shr QR_FINDER_SUBPREC, _x1 shr QR_FINDER_SUBPREC,_y1 shr QR_FINDER_SUBPREC, _x2 shr QR_FINDER_SUBPREC,_y2 shr QR_FINDER_SUBPREC, _x3 shr QR_FINDER_SUBPREC,_y3 shr QR_FINDER_SUBPREC, _res]);{$ENDIF}
  dx10:=_x1-_x0;
  dx20:=_x2-_x0;
  dx30:=_x3-_x0;
  dx31:=_x3-_x1;
  dx32:=_x3-_x2;
  dy10:=_y1-_y0;
  dy20:=_y2-_y0;
  dy30:=_y3-_y0;
  dy31:=_y3-_y1;
  dy32:=_y3-_y2;
  a20:=dx32*dy10-dx10*dy32;
  a21:=dx20*dy31-dx31*dy20;
  a22:=dx32*dy31-dx31*dy32;
  (*Figure out if we need to downscale anything.*)
  b0:=qr_ilog(QR_MAXI(abs(dx10),abs(dy10)))+qr_ilog(abs(a20+a22));
  b1:=qr_ilog(QR_MAXI(abs(dx20),abs(dy20)))+qr_ilog(abs(a21+a22));
  b2:=qr_ilog(QR_MAXI(QR_MAXI(abs(a20),abs(a21)),abs(a22)));
  s1:=QR_MAXI(0,_res+QR_MAXI(QR_MAXI(b0,b1),b2)-(QR_INT_BITS-2));
  r1:=(1 shl s1) shr 1;
  (*Compute the final coefficients of the forward transform.
    The 32x32->64 bit multiplies are really needed for accuracy with large
     versions.*)
  _hom.fwd[0][0]:=QR_FIXMUL(dx10,a20+a22,r1,s1);
  _hom.fwd[0][1]:=QR_FIXMUL(dx20,a21+a22,r1,s1);
  _hom.x0:=_x0;
  _hom.fwd[1][0]:=QR_FIXMUL(dy10,a20+a22,r1,s1);
  _hom.fwd[1][1]:=QR_FIXMUL(dy20,a21+a22,r1,s1);
  _hom.y0:=_y0;
  _hom.fwd[2][0]:=(a20+r1) div (1 shl s1);
  _hom.fwd[2][1]:=(a21+r1) div (1 shl s1);
  if s1>_res then
    _hom.fwd22 := (a22+(r1 shr _res)) div (1 shl (s1-_res))
  else
    _hom.fwd22 := a22 shl (_res-s1);
  (*Now compute the inverse transform.*)
  b0:=qr_ilog(QR_MAXI(QR_MAXI(abs(dx10),abs(dx20)),abs(dx30)))+
      qr_ilog(QR_MAXI(abs(_hom.fwd[0][0]),abs(_hom.fwd[1][0])));
  b1:=qr_ilog(QR_MAXI(QR_MAXI(abs(dy10),abs(dy20)),abs(dy30)))+
      qr_ilog(QR_MAXI(abs(_hom.fwd[0][1]),abs(_hom.fwd[1][1])));
  b2:=qr_ilog(abs(a22))-s1;
  s2:=QR_MAXI(0,QR_MAXI(b0,b1)+b2-(QR_INT_BITS-3));
  r2:=(1 shl s2) shr 1;
  Inc(s1, s2);
  r1:= r1 shl s2;
  (*The 32x32->64 bit multiplies are really needed for accuracy with large
     versions.*)
  _hom.inv[0][0]:=QR_FIXMUL(_hom.fwd[1][1],a22,r1,s1);
  _hom.inv[0][1]:=QR_FIXMUL(-_hom.fwd[0][1],a22,r1,s1);
  _hom.inv[1][0]:=QR_FIXMUL(-_hom.fwd[1][0],a22,r1,s1);
  _hom.inv[1][1]:=QR_FIXMUL(_hom.fwd[0][0],a22,r1,s1);
  _hom.inv[2][0]:=QR_FIXMUL(_hom.fwd[1][0],_hom.fwd[2][1],
                 -QR_EXTMUL(_hom.fwd[1][1],_hom.fwd[2][0],r2), s2);
  _hom.inv[2][1]:=QR_FIXMUL(_hom.fwd[0][1],_hom.fwd[2][0],
                 -QR_EXTMUL(_hom.fwd[0][0],_hom.fwd[2][1],r2), s2);
  _hom.inv22:=QR_FIXMUL(_hom.fwd[0][0],_hom.fwd[1][1],
             -QR_EXTMUL(_hom.fwd[0][1],_hom.fwd[1][0],r2), s2);
  _hom.res:=_res;
end;

(*Finish a partial projection, converting from homogeneous coordinates to the
   normal 2-D representation.
  In loops, we can avoid many multiplies by computing the homogeneous _x, _y,
   and _w incrementally, but we cannot avoid the divisions, done here.*)
procedure qr_hom_fproject(out _p: qr_point;  _hom: pqr_hom;  _x, _y, _w: integer);
begin
  if _w=0 then begin
    if _x < 0 then
      _p[0] := Low(integer)
    else
      _p[0] := High(integer);
    if _y < 0 then
      _p[1] := Low(integer)
    else
      _p[1] := High(integer);
  end
  else begin
    if _w<0 then begin
      _x := -_x;
      _y := -_y;
      _w := -_w;
    end;
    _p[0] := QR_DIVROUND(_x,_w)+_hom^.x0; (*Divides a signed integer by a positive value with exact rounding.*)
    _p[1] := QR_DIVROUND(_y,_w)+_hom^.y0;
  end;
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_hom_fproject: (%d,%d)->(%d,%d); (%d,%d)', [_x,_y, _p[0],_p[1], _p[0] shr QR_FINDER_SUBPREC, _p[1] shr QR_FINDER_SUBPREC]);{$ENDIF}
end;

(*Map from the image (at subpel resolution) into the square domain.
  Returns a negative value if the point went to infinity.*)
function qr_hom_unproject(out _q: qr_point; _hom: pqr_hom; _x,_y: integer): integer;
var
  x, y, w: integer;
begin
  // strana stvorca je: 2^_hom^.res
  Assert(_hom^.res > 0);
  Dec(_x, _hom^.x0);
  Dec(_y, _hom^.y0);
  x:= _hom^.inv[0][0]*_x + _hom^.inv[0][1]*_y;
  y:= _hom^.inv[1][0]*_x + _hom^.inv[1][1]*_y;
  w:=(_hom^.inv[2][0]*_x + _hom^.inv[2][1]*_y + _hom^.inv22 + (1 shl (_hom^.res-1))) {>>}div (1 shl _hom^.res);
  if (w=0) then begin
    if x<0 then
      _q[0]:=Low(integer)
    else
      _q[0]:=High(integer);
    if y<0 then
      _q[1]:=Low(integer)
    else
      _q[1]:=High(integer);
    Exit(-1);
  end
  else begin
    if (w<0) then begin
      x:=-x;
      y:=-y;
      w:=-w;
    end;
    _q[0]:=QR_DIVROUND(x,w);
    _q[1]:=QR_DIVROUND(y,w);
  end;
  Exit(0);
end;

procedure qr_hom_cell_init(out _cell: qr_hom_cell;  _u0, _v0, _u1, _v1, _u2, _v2, _u3, _v3, _x0, _y0, _x1, _y1, _x2, _y2, _x3, _y3: integer);
var
  du10,du20,du30,du31,du32,dv10,dv20,dv30,dv31,dv32,dx10,dx20,dx30,dx31,dx32,dy10,dy20,dy30,dy31,dy32,
  a00,a01,a02,a10,a11,a12,a20,a21,a22,i00,i01,i10,i11,i20,i21,i22,b0,b1,b2,shift,round,x,y,w: integer;
begin
  // suradnice stredov Finder Pattern a Alignment Pattern
  // x,y suradnice su vynasobene QR_FINDER_SUBPREC
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_hom_cell_init: (%d,%d), (%d,%d), (%d,%d), (%d,%d) -> (%d,%d), (%d,%d), (%d,%d), (%d,%d); (%d,%d), (%d,%d), (%d,%d), (%d,%d)', [_u0, _v0, _u1, _v1, _u2, _v2, _u3, _v3, _x0, _y0, _x1, _y1, _x2, _y2, _x3, _y3, _x0 shr QR_FINDER_SUBPREC, _y0 shr QR_FINDER_SUBPREC, _x1 shr QR_FINDER_SUBPREC, _y1 shr QR_FINDER_SUBPREC, _x2 shr QR_FINDER_SUBPREC, _y2 shr QR_FINDER_SUBPREC, _x3 shr QR_FINDER_SUBPREC, _y3 shr QR_FINDER_SUBPREC]);{$ENDIF}
  (*First, correct for the arrangement of the source points.
    We take advantage of the fact that we know the source points have a very
     limited dynamic range (so there will never be overflow) and a small amount
     of projective distortion.*)
  du10:=_u1-_u0;
  du20:=_u2-_u0;
  du30:=_u3-_u0;
  du31:=_u3-_u1;
  du32:=_u3-_u2;
  dv10:=_v1-_v0;
  dv20:=_v2-_v0;
  dv30:=_v3-_v0;
  dv31:=_v3-_v1;
  dv32:=_v3-_v2;
  (*Compute the coefficients of the forward transform from the unit square to
     the source point configuration.*)
  a20:=du32*dv10-du10*dv32;
  a21:=du20*dv31-du31*dv20;
  if (a20<>0) or (a21<>0) then
    a22:=du32*dv31-du31*dv32
    (*If the source grid points aren't in a non-affine arrangement, there's no
       reason to scale everything by du32*dv31-du31*dv32.
      Not doing so allows a much larger dynamic range, and is the only way we can
       initialize a base cell that covers the whole grid.*)
  else
    a22:=1;
  a00:=du10*(a20+a22);
  a01:=du20*(a21+a22);
  a10:=dv10*(a20+a22);
  a11:=dv20*(a21+a22);
  (*Now compute the inverse transform.*)
  i00:=a11*a22;
  i01:=-a01*a22;
  i10:=-a10*a22;
  i11:=a00*a22;
  i20:=a10*a21-a11*a20;
  i21:=a01*a20-a00*a21;
  i22:=a00*a11-a01*a10;
  (*Invert the coefficients.
    Since i22 is the largest, we divide it by all the others.
    The quotient is often exact (e.g., when the source points contain no
     projective distortion), and is never zero.
    Hence we can use zero to signal "infinity" when the divisor is zero.*)
  if (i00<>0) then i00:=QR_FLIPSIGNI(QR_DIVROUND(i22,abs(i00)),i00);
  if (i01<>0) then i01:=QR_FLIPSIGNI(QR_DIVROUND(i22,abs(i01)),i01);
  if (i10<>0) then i10:=QR_FLIPSIGNI(QR_DIVROUND(i22,abs(i10)),i10);
  if (i11<>0) then i11:=QR_FLIPSIGNI(QR_DIVROUND(i22,abs(i11)),i11);
  if (i20<>0) then i20:=QR_FLIPSIGNI(QR_DIVROUND(i22,abs(i20)),i20);
  if (i21<>0) then i21:=QR_FLIPSIGNI(QR_DIVROUND(i22,abs(i21)),i21);
  (*Now compute the map from the unit square into the image.*)
  dx10:=_x1-_x0;
  dx20:=_x2-_x0;
  dx30:=_x3-_x0;
  dx31:=_x3-_x1;
  dx32:=_x3-_x2;
  dy10:=_y1-_y0;
  dy20:=_y2-_y0;
  dy30:=_y3-_y0;
  dy31:=_y3-_y1;
  dy32:=_y3-_y2;
  a20:=dx32*dy10-dx10*dy32;
  a21:=dx20*dy31-dx31*dy20;
  a22:=dx32*dy31-dx31*dy32;
  (*Figure out if we need to downscale anything.*)
  b0:=qr_ilog(QR_MAXI(abs(dx10),abs(dy10)))+qr_ilog(abs(a20+a22));
  b1:=qr_ilog(QR_MAXI(abs(dx20),abs(dy20)))+qr_ilog(abs(a21+a22));
  b2:=qr_ilog(QR_MAXI(QR_MAXI(abs(a20),abs(a21)),abs(a22)));
  shift:=QR_MAXI(0, QR_MAXI(QR_MAXI(b0,b1),b2)-(QR_INT_BITS-3-QR_ALIGN_SUBPREC));
  round:=(1 shl shift) shr 1;
  (*Compute the final coefficients of the forward transform.*)
  a00:=QR_FIXMUL(dx10,a20+a22,round,shift);
  a01:=QR_FIXMUL(dx20,a21+a22,round,shift);
  a10:=QR_FIXMUL(dy10,a20+a22,round,shift);
  a11:=QR_FIXMUL(dy20,a21+a22,round,shift);
  (*And compose the two transforms.
    Since we inverted the coefficients above, we divide by them here instead
     of multiplying.
    This lets us take advantage of the full dynamic range.
    Note a zero divisor is really "infinity", and thus the quotient should also
     be zero.*)
  FillByte(_cell, SizeOf(_cell), 0);
  //_cell->fwd[0][0]=(i00?QR_DIVROUND(a00,i00):0)+(i10?QR_DIVROUND(a01,i10):0);
  if i00<>0 then Inc(_cell.fwd[0][0], QR_DIVROUND(a00,i00));
  if i10<>0 then Inc(_cell.fwd[0][0], QR_DIVROUND(a01,i10));
  //_cell->fwd[0][1]=(i01?QR_DIVROUND(a00,i01):0)+(i11?QR_DIVROUND(a01,i11):0);
  if i01<>0 then Inc(_cell.fwd[0][1], QR_DIVROUND(a00,i01));
  if i11<>0 then Inc(_cell.fwd[0][1], QR_DIVROUND(a01,i11));
  //_cell->fwd[1][0]=(i00?QR_DIVROUND(a10,i00):0)+(i10?QR_DIVROUND(a11,i10):0);
  if i00<>0 then Inc(_cell.fwd[1][0], QR_DIVROUND(a10,i00));
  if i10<>0 then Inc(_cell.fwd[1][0], QR_DIVROUND(a11,i10));
  //_cell->fwd[1][1]=(i01?QR_DIVROUND(a10,i01):0)+(i11?QR_DIVROUND(a11,i11):0);
  if i01<>0 then Inc(_cell.fwd[1][1], QR_DIVROUND(a10,i01));
  if i11<>0 then Inc(_cell.fwd[1][1], QR_DIVROUND(a11,i11));
  //_cell->fwd[2][0]=(i00?QR_DIVROUND(a20,i00):0)+(i10?QR_DIVROUND(a21,i10):0)+(i20?QR_DIVROUND(a22,i20):0)+round>>shift;
  if i00<>0 then Inc(_cell.fwd[2][0], QR_DIVROUND(a20,i00));
  if i10<>0 then Inc(_cell.fwd[2][0], QR_DIVROUND(a21,i10));
  if i20<>0 then Inc(_cell.fwd[2][0], QR_DIVROUND(a22,i20));
  Inc(_cell.fwd[2][0], round);
  _cell.fwd[2][0] := _cell.fwd[2][0] div (1 shl shift);
  //_cell->fwd[2][1]=(i01?QR_DIVROUND(a20,i01):0)+(i11?QR_DIVROUND(a21,i11):0)+(i21?QR_DIVROUND(a22,i21):0)+round>>shift;
  if i01<>0 then Inc(_cell.fwd[2][1], QR_DIVROUND(a20,i01));
  if i11<>0 then Inc(_cell.fwd[2][1], QR_DIVROUND(a21,i11));
  if i21<>0 then Inc(_cell.fwd[2][1], QR_DIVROUND(a22,i21));
  Inc(_cell.fwd[2][1], round);
  _cell.fwd[2][1] := _cell.fwd[2][1] div (1 shl shift);
  //_cell->fwd[2][2]=a22+round>>shift;
  _cell.fwd[2][2] := (a22+round) div (1 shl shift);
  (*Mathematically, a02 and a12 are exactly zero.
    However, that concentrates all of the rounding error in the (_u3,_v3)
     corner; we compute offsets which distribute it over the whole range.*)
  x:=_cell.fwd[0][0]*du10+_cell.fwd[0][1]*dv10;
  y:=_cell.fwd[1][0]*du10+_cell.fwd[1][1]*dv10;
  w:=_cell.fwd[2][0]*du10+_cell.fwd[2][1]*dv10+_cell.fwd[2][2];
  a02:=dx10*w-x;
  a12:=dy10*w-y;
  x:=_cell.fwd[0][0]*du20+_cell.fwd[0][1]*dv20;
  y:=_cell.fwd[1][0]*du20+_cell.fwd[1][1]*dv20;
  w:=_cell.fwd[2][0]*du20+_cell.fwd[2][1]*dv20+_cell.fwd[2][2];
  a02 := a02 + dx20*w-x;
  a12 := a12 + dy20*w-y;
  x:=_cell.fwd[0][0]*du30+_cell.fwd[0][1]*dv30;
  y:=_cell.fwd[1][0]*du30+_cell.fwd[1][1]*dv30;
  w:=_cell.fwd[2][0]*du30+_cell.fwd[2][1]*dv30+_cell.fwd[2][2];
  a02 := a02 + dx30*w-x;
  a12 := a12 + dy30*w-y;
  _cell.fwd[0][2]:=(a02+2) div 4{>>2};
  _cell.fwd[1][2]:=(a12+2) div 4{>>2};
  _cell.x0:=_x0;
  _cell.y0:=_y0;
  _cell.u0:=_u0;
  _cell.v0:=_v0;
end;

(*Finish a partial projection, converting from homogeneous coordinates to the
   normal 2-D representation.
  In loops, we can avoid many multiplies by computing the homogeneous _x, _y,
   and _w incrementally, but we cannot avoid the divisions, done here.*)
procedure qr_hom_cell_fproject(out _p: qr_point;  _cell: pqr_hom_cell;  _x, _y, _w: integer);
begin
  if _w=0 then begin
    if _x<0 then
      _p[0] := Low(integer)
    else
      _p[0] := High(integer);
    if _y<0 then
      _p[1] := Low(integer)
    else
      _p[1] := High(integer);
  end
  else begin
    if _w<0 then begin
      _x := -_x;
      _y := -_y;
      _w := -_w;
    end;
    _p[0] := QR_DIVROUND(_x,_w) + _cell^.x0;
    _p[1] := QR_DIVROUND(_y,_w) + _cell^.y0;
  end;
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_hom_cell_fproject: (%d,%d),%d -> (%d,%d); (%d,%d)', [_x,_y,_w, _p[0],_p[1], _p[0] shr QR_FINDER_SUBPREC, _p[1] shr QR_FINDER_SUBPREC]);{$ENDIF}
end;


(*Retrieves the bits corresponding to the alignment pattern template centered
   at the given location in the original image (at subpel precision).*)
type
  T55_qr_point = array[0..4, 0..4] of qr_point;
function qr_alignment_pattern_fetch(const _p: T55_qr_point; _x0,_y0: integer; _img: PByte; _width,_height: integer): unsigned;
var
  i, j, k, dx, dy: integer;
begin
  dx:=_x0-_p[2,2][0];
  dy:=_y0-_p[2,2][1];
  Result:=0;
  k:=0;
  for i:=0 to 4 do
    for j:=0 to 4 do begin
      Result := Result or (qr_img_get_bit(_img,_width,_height, _p[i,j][0]+dx, _p[i,j][1]+dy) shl k);
      Inc(k);
    end;
end;

function qr_finder_locate_crossing(_img: PByte; _width,_height: integer;  _x0,_y0, _x1,_y1, _v: integer; out _p: qr_point): integer;
  forward;

(*Searches for an alignment pattern  the given location.*)
function qr_alignment_pattern_search(out _p: qr_point;  _cell: pqr_hom_cell;  _u, _v, _r: integer;  _img: pbyte;  _width, _height: integer): integer;
const
  MASK_TESTS: array[0..7,0..1] of unsigned = (
    ($1040041,$1000001),($0041040,$0001000),
    ($0110110,$0100010),($0011100,$0001000),
    ($0420084,$0400004),($0021080,$0001000),
    ($0006C00,$0004400),($0003800,$0001000));
  MASK_COORDS: array[0..7,0..1] of unsigned = (
    (0,0),(1,1),(4,0),(3,1),(2,0),(2,1),(0,2),(1,2));
var
  c: array[0..3] of qr_point;
  nc: array[0..3] of integer;
  p: T55_qr_point;
  pc: qr_point;
  best_match, match: unsigned;
  best_dist, bestx, besty: integer;
  dist, u, v, x0,y0,w0, x,y,w, dxdu, dydu, dwdu, dxdv, dydv, dwdv, dx, dy, i, j: integer;
  side_len, dir: integer;
  x1, y1, cx,cy, a,b: integer;
begin
  (*Build up a basic template using _cell to control shape and scale.
   We project the points in the template back to the image just once, since if
    the alignment pattern has moved, we don't really know why.
   If it's because of radial distortion, or the code wasn't flat, or something
    else, there's no reason to expect that a re-projection around each
    subsequent search point would be any closer to the actual shape than our
    first projection.
   Therefore we simply slide this template around, as is.*)
  u:=(_u-2)-_cell^.u0;
  v:=(_v-2)-_cell^.v0;
  x0:=_cell^.fwd[0][0]*u + _cell^.fwd[0][1]*v + _cell^.fwd[0][2];
  y0:=_cell^.fwd[1][0]*u + _cell^.fwd[1][1]*v + _cell^.fwd[1][2];
  w0:=_cell^.fwd[2][0]*u + _cell^.fwd[2][1]*v + _cell^.fwd[2][2];
  dxdu:=_cell^.fwd[0][0];
  dydu:=_cell^.fwd[1][0];
  dwdu:=_cell^.fwd[2][0];
  dxdv:=_cell^.fwd[0][1];
  dydv:=_cell^.fwd[1][1];
  dwdv:=_cell^.fwd[2][1];

  for i:=0 to 4 do begin
   x:=x0;
   y:=y0;
   w:=w0;
   for j:=0 to 4 do begin
     qr_hom_cell_fproject(p[i,j],_cell,x,y,w);
     Inc(x, dxdu);
     Inc(y, dydu);
     Inc(w, dwdu);
   end;
   Inc(x0, dxdv);
   Inc(y0, dydv);
   Inc(w0, dwdv);
  end;

  bestx:=p[2,2][0];
  besty:=p[2,2][1];
  best_match:=qr_alignment_pattern_fetch(p, bestx, besty, _img,_width,_height);
  best_dist:=qr_hamming_dist(best_match,$1F8D63F,25);
  if(best_dist>0) then begin
    u:=_u-_cell^.u0;
    v:=_v-_cell^.v0;
    x:=(_cell^.fwd[0][0]*u+_cell^.fwd[0][1]*v+_cell^.fwd[0][2]) shl QR_ALIGN_SUBPREC;
    y:=(_cell^.fwd[1][0]*u+_cell^.fwd[1][1]*v+_cell^.fwd[1][2]) shl QR_ALIGN_SUBPREC;
    w:=(_cell^.fwd[2][0]*u+_cell^.fwd[2][1]*v+_cell^.fwd[2][2]) shl QR_ALIGN_SUBPREC;
   (*Search an area at most _r modules around the target location, in
      concentric squares..*)
   for i:=1 to (_r shl QR_ALIGN_SUBPREC) - 1 do begin
     side_len:=(i shl 1)-1;
     Dec(x, dxdu+dxdv);
     Dec(y, dydu+dydv);
     Dec(w, dwdu+dwdv);
     for j:=0 to 4*side_len-1 do begin
       qr_hom_cell_fproject(pc,_cell,x,y,w);
       match:=qr_alignment_pattern_fetch(p, pc[0], pc[1], _img,_width,_height);
       dist:=qr_hamming_dist(match,$1F8D63F,best_dist+1);
       if (dist<best_dist) then begin
         best_match:=match;
         best_dist:=dist;
         bestx:=pc[0];
         besty:=pc[1];
       end;
       if (j<2*side_len) then begin
         dir := Ord(j>=side_len);
         Inc(x, _cell^.fwd[0][dir]);
         Inc(y, _cell^.fwd[1][dir]);
         Inc(w, _cell^.fwd[2][dir]);
       end
       else begin
         dir := Ord(j>=3*side_len);
         Dec(x, _cell^.fwd[0][dir]);
         Dec(y, _cell^.fwd[1][dir]);
         Dec(w, _cell^.fwd[2][dir]);
       end;
       if (best_dist=0) then break;
     end;
     if (best_dist=0) then break;
   end;
  end;

  (*If the best result we got was sufficiently bad, reject the match.
   If we're wrong and we include it, we can grossly distort the nearby
    region, whereas using the initial starting point should at least be
    consistent with the geometry we already have.*)
  if(best_dist>6) then begin
    _p[0]:=p[2][2][0];
    _p[1]:=p[2][2][1];
    Exit(-1);
  end;

  (*Now try to get a more accurate location of the pattern center.*)
  dx:=bestx-p[2][2][0];
  dy:=besty-p[2][2][1];
  FillByte(nc,sizeof(nc),0);
  FillByte(c,sizeof(c),0);
  (*We consider 8 lines across the finder pattern in turn.
   If we actually found a symmetric pattern along that line, search for its
    exact center in the image.
   There are plenty more lines we could use if these don't work, but if we've
    found anything remotely close to an alignment pattern, we should be able
    to use most of these.*)
  for i:=0 to 7 do begin
   if ((best_match and MASK_TESTS[i][0]) = MASK_TESTS[i][1]) then begin
     x0 := (p[MASK_COORDS[i][1]][MASK_COORDS[i][0]][0]+dx) div (1 shl QR_FINDER_SUBPREC);
     if (x0<0) or (x0>=_width) then continue;
     y0 := (p[MASK_COORDS[i][1]][MASK_COORDS[i][0]][1]+dy) div (1 shl QR_FINDER_SUBPREC);
     if (y0<0) or (y0>=_height) then continue;
     x1 := (p[4-MASK_COORDS[i][1]][4-MASK_COORDS[i][0]][0]+dx) div (1 shl QR_FINDER_SUBPREC);
     if (x1<0) or (x1>=_width) then continue;
     y1 := (p[4-MASK_COORDS[i][1]][4-MASK_COORDS[i][0]][1]+dy) div (1 shl QR_FINDER_SUBPREC);
     if (y1<0) or (y1>=_height) then continue;
     if (qr_finder_locate_crossing(_img,_width,_height, x0,y0, x1,y1, i and 1, pc)=0) then begin
       cx:=pc[0]-bestx;
       cy:=pc[1]-besty;
       if (i and 1 <> 0) then begin
         (*Weight crossings around the center dot more highly, as they are
            generally more reliable.*)
         w:=3;
         Inc(cx, cx shl 1);
         Inc(cy, cy shl 1);
       end
       else
         w:=1;
       Inc(nc[i shr 1], w);
       Inc(c[i shr 1][0], cx);
       Inc(c[i shr 1][1], cy);
     end;
   end;
  end;

  (*Sum offsets from lines in orthogonal directions.*)
  for i:=0 to 1 do begin
   a:=nc[i shl 1];
   b:=nc[(i shl 1) or 1];
   if (a<>0) and (b<>0) then begin
     w:=QR_MAXI(a,b);
     c[i shl 1][0] := QR_DIVROUND(w*(b*c[i shl 1][0]+a*c[(i shl 1) or 1][0]), a*b);
     c[i shl 1][1] := QR_DIVROUND(w*(b*c[i shl 1][1]+a*c[(i shl 1) or 1][1]), a*b);
     nc[i shl 1] := w shl 1;
   end
   else begin
     Inc(c[i shl 1][0], c[(i shl 1) or 1][0]);
     Inc(c[i shl 1][1], c[(i shl 1) or 1][1]);
     Inc(nc[i shl 1], b);
   end;
  end;

  (*Average offsets from pairs of orthogonal lines.*)
  Inc(c[0][0], c[2][0]);
  Inc(c[0][1], c[2][1]);
  Inc(nc[0], nc[2]);
  (*If we actually found any such lines, apply the adjustment.*)
  if (nc[0]<>0) then begin
   dx:=QR_DIVROUND(c[0][0],nc[0]);
   dy:=QR_DIVROUND(c[0][1],nc[0]);
   (*But only if it doesn't make things too much worse.*)
   match:=qr_alignment_pattern_fetch(p,bestx+dx,besty+dy,_img,_width,_height);
   dist:=qr_hamming_dist(match,$1F8D63F,best_dist+1);
   if(dist<=best_dist+1) then begin
     Inc(bestx, dx);
     Inc(besty, dy);
   end;
  end;
  _p[0]:=bestx;
  _p[1]:=besty;
  // best_dist = 0 idealne
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_alignment_pattern_search: best (%d,%d) = %d', [bestx div (1 shl QR_FINDER_SUBPREC), besty div (1 shl QR_FINDER_SUBPREC), best_dist]);{$ENDIF}
  Exit(0);
end;

{$ELSE}

procedure qr_hom_init(out _hom: qr_hom; _x0,_y0, _x1,_y1, _x2,_y2, _x3,_y3, _res: integer);
begin
  FillByte(_hom, SizeOf(_hom), 0);
  _hom.fwd[0][0{x}] := 1;
  _hom.fwd[1][1{y}] := 1;
end;

// square -> image
procedure qr_hom_fproject(out _p: qr_point;  _hom: pqr_hom;  _x, _y, _w: integer); inline;
begin
  // we does not care about real coordinates
  // qr_img_get_bit() shifts >> QR_FINDER_SUBPREC
  _p[0] := _x shl QR_FINDER_SUBPREC;
  _p[1] := _y shl QR_FINDER_SUBPREC;
end;

// image -> square
function qr_hom_unproject(out _q: qr_point; _hom: pqr_hom; _x,_y: integer): integer; inline;
begin
  _q[0] := _x shr QR_FINDER_SUBPREC;
  _q[1] := _y shr QR_FINDER_SUBPREC;
end;

procedure qr_hom_cell_init(out _cell: qr_hom_cell;  _u0, _v0, _u1, _v1, _u2, _v2, _u3, _v3, _x0, _y0, _x1, _y1, _x2, _y2, _x3, _y3: integer);
begin
  FillByte(_cell, SizeOf(_cell), 0);
  _cell.fwd[0,0] := 1;
  _cell.fwd[1,1] := 1;
end;

procedure qr_hom_cell_fproject(out _p: qr_point;  _cell: pqr_hom_cell;  _x, _y, _w: integer); inline;
begin
  _p[0] := _x shl QR_FINDER_SUBPREC;
  _p[1] := _y shl QR_FINDER_SUBPREC;
end;

function qr_alignment_pattern_search(out _p: qr_point;  _cell: pqr_hom_cell;  _u, _v, _r: integer;  _img: pbyte;  _width, _height: integer): integer; inline;
begin
  Result := 0;
end;
{$ENDIF}


procedure qr_finder_init(out fp: qr_finder; _hom: pqr_hom; _x, _y: integer);
var p1: qr_point;
begin
  FillByte(fp, SizeOf(fp), 0);
  // The center of the finder pattern (in the square domain)
  qr_hom_unproject(fp.o, _hom, _x, _y); // image -> square

  // velkost 1 bodu v stvorci
  qr_hom_unproject(p1, _hom, _hom^.x0 + 1 shl QR_FINDER_SUBPREC, _hom^.y0 + 1 shl QR_FINDER_SUBPREC);
  fp.size[0] := p1[0]; // 1
  fp.size[1] := p1[1]; // 1

  New(fp.c);
  fp.c^.pos[0] := _x;
  fp.c^.pos[1] := _y;
end;

procedure qr_finder_clear(var fp: qr_finder);
begin
  Dispose(fp.c);
end;


(*The BCH(18,6,3) codes are only used for version information, which must lie
   between 7 and 40 (inclusive).*)
const
  BCH18_6_CODES: array [0..33] of unsigned = ($07C94,$085BC,$09A99,$0A4D3,$0BBF6,$0C762,$0D847,$0E60D,$0F928,$10B78,$1145D,$12A17,$13532,$149A6,$15683,$168C9,$177EC,$18EC4,$191E1,$1AFAB,$1B08E,$1CC1A,$1D33F,$1ED75,$1F250,$209D5,$216F0,$228BA,$2379F,$24B0B,$2542E,$26A64,$27541,$28C69);

(*Corrects a BCH(18,6,3) code word.
  _y: Contains the code word to be checked on input, and the corrected value on
      output.
  Return: The number of errors.
          If more than 3 errors are detected, returns a negative value and
          performs no correction.*)
function bch18_6_correct(var _y: unsigned): integer;
var
  x: unsigned;
begin
  (*Check the easy case first: see if the data bits were uncorrupted.*)
  x := _y shr 12;
  if (x>=7) and (x<=40) then begin
    Result := qr_hamming_dist(_y, BCH18_6_CODES[x-7], 4);
    if Result < 4 then begin
      _y := BCH18_6_CODES[x-7];
      Exit;
    end;
  end;

  (*Exhaustive search is faster than field operations in GF(19).*)
  for x:=0 to 33 do
    if x+7 <> _y shr 12 then begin
      Result := qr_hamming_dist(_y, BCH18_6_CODES[x], 4);
      if Result < 4 then begin
        _y := BCH18_6_CODES[x];
        Exit;
      end;
    end;

  Result := -1;
end;


(*Reads the version bits  a finder module and decodes the version number.*)
// _dir = 0 for Top-Roght
//        1 for Bottom-Left
// If we have a small version (less than 7), there's no encoded version information.
function qr_finder_version_decode(_f: pqr_finder;  _hom: pqr_hom;  _img: pbyte;  _width, _height: integer;  _dir: integer): integer;
var
  p,q: qr_point;
  v: unsigned;
  x0, y0, w0: integer;
  dxi, dyi, dwi, dxj, dyj, dwj: integer;
  i,j,k,x,y,w: integer;
begin
  v := 0;
  q[_dir] := _f^.o[_dir]-7*_f^.size[_dir];
  q[1-_dir] := _f^.o[1-_dir]-3*_f^.size[1-_dir];
  x0:= _hom^.fwd[0][0]*q[0]+_hom^.fwd[0][1]*q[1];
  y0:= _hom^.fwd[1][0]*q[0]+_hom^.fwd[1][1]*q[1];
  w0:= _hom^.fwd[2][0]*q[0]+_hom^.fwd[2][1]*q[1] + _hom^.fwd22;
  dxi:= _hom^.fwd[0][1-_dir]*_f^.size[1-_dir];
  dyi:= _hom^.fwd[1][1-_dir]*_f^.size[1-_dir];
  dwi:= _hom^.fwd[2][1-_dir]*_f^.size[1-_dir];
  dxj:= _hom^.fwd[0][_dir]*_f^.size[_dir];
  dyj:= _hom^.fwd[1][_dir]*_f^.size[_dir];
  dwj:= _hom^.fwd[2][_dir]*_f^.size[_dir];
  k:=0;
  // 3x6 bit version information
  for i:=0 to Pred(6) do begin
    x := x0;
    y := y0;
    w := w0;
    j:= 0;
    while j<3 do begin
      qr_hom_fproject(p,_hom,x,y,w);
      v := v or (qr_img_get_bit(_img,_width,_height,p[0],p[1]) shl k);
      Inc(x, dxj);
      Inc(y, dyj);
      Inc(w, dwj);
      Inc(j);
      Inc(k);
    end;
    Inc(x0, dxi);
    Inc(y0, dyi);
    Inc(w0, dwi);
  end;
  Result := bch18_6_correct(v);

  (*TODO: I'd certainly hope the order the version bits is accessed in is
     well-defined, but I seem to have images for two different codes with the
     same version using two different orders?
    Maybe the other one is a model 1 code?
    Even if I parse the version here, I can't decode the rest of the code.
    If this is really needed, we should just re-order the bits.*)
  if Result >= 0 then
    Result := v shr 12;
end;


(*Reads the format info bits  the finder modules and decodes them.*)
function qr_finder_fmt_info_decode(_ul, _ur, _dl: pqr_finder;  _hom: pqr_hom;  _img: pbyte;  _width, _height: integer): integer;
var
  p: qr_point;
  lo, hi: array [0..1] of unsigned;
  u,v,x,y,w,dx,dy,dw: integer;
  fmt_info, count, nerrs: array [0..3] of integer;
  nfmt_info: integer;
  besti, maxi, di, i, j, k, ret: integer;
  v2: unsigned;
begin
  // suradnice UL,UR a DL su v ramci stvorca so stranou 2^_hom^.res
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_finder_fmt_info_decode: UL=(%.2f,%.2f), UR=(%.2f,%.2f), DL=(%.2f,%.2f), UL size=(%.2f,%2f), DL size=(%.2f,%2f)',[_ul^.o[0]/(1 shl _hom^.res),_ul^.o[1]/(1 shl _hom^.res), _ur^.o[0]/(1 shl _hom^.res),_ur^.o[1]/(1 shl _hom^.res), _dl^.o[0]/(1 shl _hom^.res),_dl^.o[1]/(1 shl _hom^.res), _ul^.size[0]/(1 shl _hom^.res),_ul^.size[1]/(1 shl _hom^.res), _dl^.size[0]/(1 shl _hom^.res),_dl^.size[1]/(1 shl _hom^.res)]);{$ENDIF}
  (*Read the bits around the UL corner.*)
  lo[0] := 0;
  u := _ul^.o[0] + 5*_ul^.size[0]; // x-shift from center
  v := _ul^.o[1] - 3*_ul^.size[1]; // y-shift from center
  x := _hom^.fwd[0][0]*u + _hom^.fwd[0][1]*v;
  y := _hom^.fwd[1][0]*u + _hom^.fwd[1][1]*v;
  w := _hom^.fwd[2][0]*u + _hom^.fwd[2][1]*v + _hom^.fwd22;
  dx:= _hom^.fwd[0][1]*_ul^.size[1];
  dy:= _hom^.fwd[1][1]*_ul^.size[1];
  dw:= _hom^.fwd[2][1]*_ul^.size[1];
  k:=0;
  i:=0;
  // read 8 bits on right side of top-left FP
  while True do begin
    (*Skip the timing pattern row.*)
    if i<>6 then begin
      qr_hom_fproject(p,_hom,x,y,w);
      lo[0] := lo[0] or (qr_img_get_bit(_img, _width, _height, p[0], p[1]) shl k{++});
      Inc(k);
      (*Don't advance q in the last iteration... we'll start the next loop from
        the current position.*)
      if i>=8 then break;
    end;
    Inc(x, dx);
    Inc(y, dy);
    Inc(w, dw);
    Inc(i);
  end;

  hi[0] := 0;
  dx := -_hom^.fwd[0][0]*_ul^.size[0];
  dy := -_hom^.fwd[1][0]*_ul^.size[0];
  dw := -_hom^.fwd[2][0]*_ul^.size[0];
  // read next 8 bits on bottom side of top-left FP
  while i > 0 do begin
    Inc(x, dx);
    Inc(y, dy);
    Inc(w, dw);
    (*Skip the timing pattern column.*)
    if i<>6 then begin
      qr_hom_fproject(p,_hom,x,y,w);
      hi[0] := hi[0] or (qr_img_get_bit(_img, _width, _height, p[0], p[1]) shl k{++});
      Inc(k);
    end;
    Dec(i);
  end;

  // Top-Right
  lo[1] := 0;
  u := _ur^.o[0]+3*_ur^.size[0];
  v := _ur^.o[1]+5*_ur^.size[1];
  x:= _hom^.fwd[0][0]*u+_hom^.fwd[0][1]*v;
  y:= _hom^.fwd[1][0]*u+_hom^.fwd[1][1]*v;
  w:= _hom^.fwd[2][0]*u+_hom^.fwd[2][1]*v + _hom^.fwd22;
  dx:= -_hom^.fwd[0][0]*_ur^.size[0];
  dy:= -_hom^.fwd[1][0]*_ur^.size[0];
  dw:= -_hom^.fwd[2][0]*_ur^.size[0];
  (*Read the 8 bits next to the UR corner.*)
  for k:=0 to 7 {k++} do begin
    qr_hom_fproject(p,_hom,x,y,w);
    lo[1] := lo[1] or (qr_img_get_bit(_img,_width,_height,p[0],p[1]) shl k);
    Inc(x, dx);
    Inc(y, dy);
    Inc(w, dw);
  end;

  hi[1] := 0;
  u := _dl^.o[0]+5*_dl^.size[0];
  v := _dl^.o[1]-3*_dl^.size[1];
  x := _hom^.fwd[0][0]*u+_hom^.fwd[0][1]*v;
  y := _hom^.fwd[1][0]*u+_hom^.fwd[1][1]*v;
  w := _hom^.fwd[2][0]*u+_hom^.fwd[2][1]*v + _hom^.fwd22;
  dx:= _hom^.fwd[0][1]*_dl^.size[1];
  dy:= _hom^.fwd[1][1]*_dl^.size[1];
  dw:= _hom^.fwd[2][1]*_dl^.size[1];
  (*Read the bits next to the DL corner.*)
  for k:=8 to 14 {k++} do begin
    qr_hom_fproject(p,_hom,x,y,w);
    hi[1] := hi[1] or (qr_img_get_bit(_img,_width,_height,p[0],p[1]) shl k);
    Inc(x, dx);
    Inc(y, dy);
    Inc(w, dw);
  end;

  (*For the 8th bit we have 3 samples... use the majority value.
        TODO: The DL bit appears to be wrong as much as right? Guess it's not
         really a third copy after all, but doesn't appear to be used for data.
      i=((lo[0]>>7&1)+(lo[1]>>7&1)+(hi[1]>>7&1)>>1)<<7;
      lo[0]=lo[0]&~0x80|i;
      lo[1]=lo[1]&~0x80|i;
      hi[1]&=~0x80;*)
  maxi := 2 shl Ord(hi[0]<>hi[1]); // 2 or 4
  di := 1 + Ord(lo[0]=lo[1]); // 1 or 2
  nfmt_info := 0;
  (*For the remaining bits we have two samples... try them in all
       combinations and pick the most popular valid code, breaking ties using
       the number of bit errors.*)
  i := 0;
  while i < maxi do begin
    v2 := (lo[i and 1] or hi[i shr 1]) xor $5412;
    ret := bch15_5_correct(v2);
    v2 := v2 shr 10;
    if ret < 0 then ret := 4;
    j := 0;
    while j<=3 do begin
      if j>=nfmt_info then begin
        fmt_info[j] := v2;
        count[j] := 1;
        nerrs[j] := ret;
        Inc(nfmt_info);
        break;
      end;
      if fmt_info[j]=integer(v2) then begin
        Inc(count[j]);
        if nerrs[j] > ret then nerrs[j] := ret;
        break;
      end;
      Inc(j);
    end;
    Inc(i, di);
  end;

  besti := 0;
  for i:=1 to Pred(nfmt_info) do
    if (nerrs[besti]>3) and (nerrs[i]<=3) or (count[i]>count[besti]) or (count[i]=count[besti]) and (nerrs[i]<nerrs[besti]) then
      besti := i;

  if nerrs[besti] < 4 then
    Result := fmt_info[besti]
  else
    Result := -1;
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_finder_fmt_info_decode: FmtInfo:%d (EccLevel:%d, Mask:%d)',[Result, Result shr 3 xor 1, Result and 7]);{$ENDIF}
end;


procedure qr_hom_cell_project(out _p: qr_point;  _cell: pqr_hom_cell;  _u, _v, _res: integer);
begin
  Dec(_u, _cell^.u0 shl _res);
  Dec(_v, _cell^.v0 shl _res);
  qr_hom_cell_fproject(_p, _cell,
                           _cell^.fwd[0,0]*_u+_cell^.fwd[0,1]*_v+(_cell^.fwd[0,2] shl _res),
                           _cell^.fwd[1,0]*_u+_cell^.fwd[1,1]*_v+(_cell^.fwd[1,2] shl _res),
                           _cell^.fwd[2,0]*_u+_cell^.fwd[2,1]*_v+(_cell^.fwd[2,2] shl _res));
end;

(*Mark a given region as belonging to the function pattern.*)
procedure qr_sampling_grid_fp_mask_rect(var _grid: qr_sampling_grid;  _dim, _u, _v, _w, _h: integer);
var
  i, j, stride: integer;
begin
  stride := (_dim+QR_INT_BITS-1) shr QR_INT_LOGBITS;
  (*Note that we store bits column-wise, since that's how they're read out of
    the grid.*)
  for j:=_u to Pred(_u+_w) do
    for i:=_v to Pred(_v+_h) do
      _grid.fpmask[j*stride+(i shr QR_INT_LOGBITS)] := _grid.fpmask[j*stride+(i shr QR_INT_LOGBITS)] or (unsigned(1) shl (i and (QR_INT_BITS-1)));
end;

(*Determine if a given grid location is inside the function pattern.*)
function qr_sampling_grid_is_in_fp(_grid: pqr_sampling_grid;  _dim, _u, _v: integer): integer;
begin
  Result := _grid^.fpmask[_u*((_dim+QR_INT_BITS-1) shr QR_INT_LOGBITS)+(_v shr QR_INT_LOGBITS)] shr (_v and (QR_INT_BITS-1)) and 1;
end;

const
  QR_ALIGNMENT_SPACING: array [0..Pred(34)] of byte = (16,18,20,22,24,26,28,20,22,24,24,26,28,28,22,24,24,26,26,28,28,24,24,26,26,26,28,28,24,26,26,26,28,28);

(*Initialize the sampling grid for each region of the code.
  _version:  The (decoded) version number.
  _ul_pos:   The location of the UL finder pattern.
  _ur_pos:   The location of the UR finder pattern.
  _dl_pos:   The location of the DL finder pattern.
  _p:        On input, contains estimated positions of the four corner modules.
             On output, contains a bounding quadrilateral for the code.
  _img:      The binary input image.
  _width:    The width of the input image.
  _height:   The height of the input image.
  Return: 0 on success, or a negative value on error.*)
procedure qr_sampling_grid_init(var _grid: qr_sampling_grid;  _version: integer;  _ul_pos, _ur_pos, _dl_pos: qr_point;  _p: pqr_point;  _img: pbyte;  _width,  _height: integer);
var
  base_cell: qr_hom_cell;
  align_pos: array [0..6] of integer;
  dim, nalign, i, j, k, k2, d, jmin, jmax: integer;
  q, p: pqr_point;
  cell: pqr_hom_cell;
  u, v: integer;
  p0, p1, p2: qr_point;
begin
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_sampling_grid_init: version=%d, UL=(%d,%d), UR=(%d,%d), DL=(%d,%d)',[_version, _ul_pos[0] shr QR_FINDER_SUBPREC,_ul_pos[1] shr QR_FINDER_SUBPREC, _ur_pos[0] shr QR_FINDER_SUBPREC,_ur_pos[1] shr QR_FINDER_SUBPREC, _dl_pos[0] shr QR_FINDER_SUBPREC,_dl_pos[1] shr QR_FINDER_SUBPREC]);{$ENDIF}
  (*Each predictor is basically a straight-line extrapolation from two
    neighboring alignment patterns (except possibly  the opposing
    finder patterns).*)
  dim := 17 + (_version shl 2);   // size of code in modules
  nalign := (_version div 7) + 2; // count of align patterns
  (*Create a base cell to bootstrap the alignment pattern search.*)
  qr_hom_cell_init(base_cell, 0,0, dim-1,0, 0,dim-1, dim-1,dim-1, _p[0][0],_p[0][1], _p[1][0],_p[1][1], _p[2][0],_p[2][1], _p[3][0],_p[3][1]);

  (*Allocate the array of cells.*)
  _grid.ncells := nalign-1;
  _grid.cells[0] := GetMem((nalign-1)*(nalign-1)*SizeOf(_grid.cells[0]^));
  for i:=1 to Pred(_grid.ncells) do
    _grid.cells[i] := _grid.cells[i-1] + _grid.ncells;

  (*Initialize the function pattern mask.*)
  _grid.fpmask := AllocMem(dim * ((dim+QR_INT_BITS-1) shr QR_INT_LOGBITS) * SizeOf(_grid.fpmask^));

  (*Mask out the finder patterns (and separators and format info bits).*)
  qr_sampling_grid_fp_mask_rect(_grid,dim,0,0,9,9);
  qr_sampling_grid_fp_mask_rect(_grid,dim,0,dim-8,9,8);
  qr_sampling_grid_fp_mask_rect(_grid,dim,dim-8,0,8,9);

  (*Mask out the version number bits.*)
  if _version>6 then begin
    qr_sampling_grid_fp_mask_rect(_grid,dim,0,dim-11,6,3);
    qr_sampling_grid_fp_mask_rect(_grid,dim,dim-11,0,3,6);
  end;

  (*Mask out the timing patterns.*)
  qr_sampling_grid_fp_mask_rect(_grid,dim,9,6,dim-17,1);
  qr_sampling_grid_fp_mask_rect(_grid,dim,6,9,1,dim-17);

  (*If we have no alignment patterns (e.g., this is a version 1 code), just use
       the base cell and hope it's good enough.*)
  if _version < 2 then
    Move(base_cell, _grid.cells[0]^, SizeOf(base_cell))
  else begin
    q := GetMem(nalign*nalign*SizeOf(q^));
    p := GetMem(nalign*nalign*SizeOf(p^));
    (*Initialize the alignment pattern position list.*)
    align_pos[0] := 6;
    align_pos[nalign-1] := dim-7;
    if _version > 6 then begin
      d := QR_ALIGNMENT_SPACING[_version-7];
      i := nalign-1;
      while i>1 do begin
        Dec(i);
        align_pos[i] := align_pos[i+1]-d;
      end;
    end;
    (*Three of the corners use a finder pattern instead of a separate
           alignment pattern.*)
    q[0][0]:= 3;
    q[0][1]:= 3;
    p[0][0]:= _ul_pos[0];
    p[0][1]:= _ul_pos[1];

    q[nalign-1][0]:= dim-4;
    q[nalign-1][1]:= 3;
    p[nalign-1][0]:= _ur_pos[0];
    p[nalign-1][1]:= _ur_pos[1];

    q[(nalign-1)*nalign][0]:= 3;
    q[(nalign-1)*nalign][1]:= dim-4;
    p[(nalign-1)*nalign][0]:= _dl_pos[0];
    p[(nalign-1)*nalign][1]:= _dl_pos[1];

    (*Scan for alignment patterns using a diagonal sweep.*)
    for k:=1 to Pred(2*nalign-1) do begin
      jmax := QR_MINI(k,nalign-1) - Ord(k=nalign-1);
      jmin := QR_MAXI(0,k-(nalign-1)) + Ord(k=nalign-1);
      for j:=jmin to jmax do begin
        i:= jmax-(j-jmin);
        k2:= i*nalign+j;
        u:= align_pos[j];
        v:= align_pos[i];
        q[k2][0]:= u;
        q[k2][1]:= v;
        (*Mask out the alignment pattern.*)
        qr_sampling_grid_fp_mask_rect(_grid,dim,u-2,v-2,5,5);
        (*Pick a cell to use to govern the alignment pattern search.*)
        if (i>1)and(j>1) then begin
          qr_hom_cell_project(p0,_grid.cells[i-2]+j-1,u,v,0);
          qr_hom_cell_project(p1,_grid.cells[i-2]+j-2,u,v,0);
          qr_hom_cell_project(p2,_grid.cells[i-1]+j-2,u,v,0);
          (*Take the median of the predictions as the search center.*)
          QR_SORT2I(p0[0],p1[0]);
          QR_SORT2I(p0[1],p1[1]);
          QR_SORT2I(p1[0],p2[0]);
          QR_SORT2I(p1[1],p2[1]);
          QR_SORT2I(p0[0],p1[0]);
          QR_SORT2I(p0[1],p1[1]);
          (*We need a cell that has the target point at a known (u,v) location.
            Since our cells don't have inverses, just construct one from our
              neighboring points.*)
          cell:= _grid.cells[i-1]+j-1;
          qr_hom_cell_init(cell^, q[k2-nalign-1][0], q[k2-nalign-1][1], q[k2-nalign][0], q[k2-nalign][1], q[k2-1][0], q[k2-1][1], q[k2][0], q[k2][1], p[k2-nalign-1][0], p[k2-nalign-1][1], p[k2-nalign][0], p[k2-nalign][1], p[k2-1][0], p[k2-1][1], p1[0], p1[1]);
        end
        else if (i>1)and(j>0) then
          cell := _grid.cells[i-2]+j-1
        else if (i>0)and(j>1) then
          cell := _grid.cells[i-1]+j-2
        else
          cell := @base_cell;
        (*Use a very small search radius.
          A large displacement here usually means a false positive (e.g., when
            the real alignment pattern is damaged or missing), which can
            severely distort the projection.*)
        qr_alignment_pattern_search(p[k2], cell, u, v, 2, _img, _width,_height);
        if (i>0)and(j>0) then
          qr_hom_cell_init((_grid.cells[i-1]+j-1)^, q[k2-nalign-1][0], q[k2-nalign-1][1], q[k2-nalign][0], q[k2-nalign][1], q[k2-1][0], q[k2-1][1], q[k2][0], q[k2][1], p[k2-nalign-1][0], p[k2-nalign-1][1], p[k2-nalign][0], p[k2-nalign][1], p[k2-1][0], p[k2-1][1], p[k2][0], p[k2][1]);
      end;
    end;
    //qr_svg_points('align',p,nalign*nalign);
    FreeMem(q);
    FreeMem(p);
  end;
  (*Set the limits over which each cell is used.*)
  Move(align_pos[1], _grid.cell_limits[0], (_grid.ncells-1)*SizeOf(_grid.cell_limits[0]));
  _grid.cell_limits[_grid.ncells-1] := dim;
  (*Produce a bounding square for the code (to mark finder centers with).
      Because of non-linear distortion, this might not actually bound the code,
       but it should be good enough.
      I don't think it's worth computing a convex hull or anything silly like
       that.*)
  qr_hom_cell_project(_p[0], _grid.cells[0]+0, -1, -1, 1);
  qr_hom_cell_project(_p[1], _grid.cells[0]+_grid.ncells-1, (dim shl 1)-1, -1, 1);
  qr_hom_cell_project(_p[2], _grid.cells[_grid.ncells-1]+0, -1, (dim shl 1)-1, 1);
  qr_hom_cell_project(_p[3], _grid.cells[_grid.ncells-1]+_grid.ncells-1, (dim shl 1)-1, (dim shl 1)-1, 1);
  (*Clamp the points somewhere  the image (this is really just in case a
       corner is  the plane at infinity).*)
  for i:=0 to Pred(4) do begin
    _p[i][0] := QR_CLAMPI((-_width) shl QR_FINDER_SUBPREC, _p[i][0], _width shl (QR_FINDER_SUBPREC+1));
    _p[i][1] := QR_CLAMPI((-_height) shl QR_FINDER_SUBPREC, _p[i][1], _height shl (QR_FINDER_SUBPREC+1));
  end;
  (*TODO: Make fine adjustments using the timing patterns.
    Possible strategy: scan the timing pattern at QR_ALIGN_SUBPREC (or finer)
     resolution, use dynamic programming to match midpoints between
     transitions to the ideal grid locations.*)
end;


procedure qr_sampling_grid_clear(var _grid: qr_sampling_grid);
begin
  FreeMem(_grid.fpmask);
  FreeMem(_grid.cells[0]);
end;


procedure qr_data_mask_fill(_mask: punsigned;  _dim: integer;  _pattern: integer);
var
  stride, i,j, m, mi, mj: unsigned;
begin
  stride := (_dim+QR_INT_BITS-1) shr QR_INT_LOGBITS;
  (*Note that we store bits column-wise, since that's how they're read out of
    the grid.*)
  case _pattern of
    (*10101010 i+j+1&1
      01010101
      10101010
      01010101*)
    0:
      begin
        m := $55;
        for j:=0 to Pred(_dim) do begin
          FillByte((_mask+j*stride)^, stride*SizeOf(_mask^), m);
          m := m xor $FF;
        end;
      end;

    1:
    (*11111111 i+1&1
      00000000
      11111111
      00000000*)
      begin
      FillByte(_mask^, _dim*stride*SizeOf(_mask^), $55);
      end;

    2:
    (*10010010 (j+1)%3&1
      10010010
      10010010
      10010010*)
      begin
        m := $FF;
        for j:=0 to Pred(_dim) do begin
          FillByte((_mask+j*stride)^, stride*sizeof(_mask^), m and $FF);
          m := (m shl 8) or (m shr 16);
        end;
      end;

    3:
    (*10010010 (i+j+1)%3&1
      00100100
      01001001
      10010010*)
      begin
        mj := 0;
        for i:=0 to Pred((QR_INT_BITS+2) div 3) do
          mj := mj or (1 shl (3*i));
        for j:=0 to Pred(_dim) do begin
          mi := mj;
          for i:=0 to Pred(stride) do begin
            _mask[j*stride+i] := mi;
            mi := (mi shr (QR_INT_BITS mod 3)) or (mi shl (3-(QR_INT_BITS mod 3)));
          end;
          mj := (mj shr 1) or (mj shl 2);
        end;
      end;

    4:
    (*11100011 (i>>1)+(j/3)+1&1
      11100011
      00011100
      00011100*)
      begin
        m := 7;
        for j:=0 to Pred(_dim) do begin
          FillByte((_mask+j*stride)^, stride*sizeof(_mask^), ($CC xor -(m and 1)) and $FF);
          m := (m shr 1) or (m shl 5);
        end;
      end;

    5:
    (*11111111 !((i*j)%6)
      10000010
      10010010
      10101010*)
      begin
        for j:=0 to Pred(_dim) do begin
          m := 0;
          for i:=0 to Pred(6) do
            m := m or (Ord((i*j) mod 6 = 0) shl i);
          i := 6;
          while i<QR_INT_BITS do begin
            m := m or (m shl i);
            i := i shl 1;
          end;
          for i:=0 to Pred(stride) do begin
            _mask[j*stride+i] := m;
            m := (m shr (QR_INT_BITS mod 6)) or (m shl (6-QR_INT_BITS mod 6));
          end;
        end;
      end;

    6:
    (*11111111 (i*j)%3+i*j+1&1
      11100011
      11011011
      10101010*)
      begin
        for j:=0 to Pred(_dim) do begin
          m := 0;
          for i:=0 to Pred(6) do
            m := m or ((((i*j) mod 3 + i*j + 1) and 1) shl i);
          i := 6;
          while i<QR_INT_BITS do begin
            m := m or (m shl i);
            i := i shl 1;
          end;
          for i:=0 to Pred(stride) do begin
            _mask[j*stride+i] := m;
            m := (m shr (QR_INT_BITS mod 6)) or (m shl (6-QR_INT_BITS mod 6));
          end;
        end;
      end;

    else
    (*10101010 (i*j)%3+i+j+1&1
      00011100
      10001110
      01010101*)
      begin
        for j:=0 to Pred(_dim) do begin
          m := 0;
          for i:=0 to Pred(6) do
            m := m or ((((i*j) mod 3+i+j+1) and 1) shl i);
          i := 6;
          while i<QR_INT_BITS do begin
            m := m or (m shl i);
            i := i shl 1;
          end;
          for i:=0 to Pred(stride) do begin
            _mask[j*stride+i] := m;
            m := (m shr (QR_INT_BITS mod 6)) or (m shl (6-QR_INT_BITS mod 6));
          end;
        end;
      end;

  end;{case}
end;


procedure qr_sampling_grid_sample(_grid: pqr_sampling_grid;  _data_bits: punsigned;  _dim, _fmt_info: integer;  _img: pbyte;  _width, _height: integer);
var
  stride, u0, u1: integer;
  i,j: integer;
  v0, v1: integer;
  cell: pqr_hom_cell;
  x0, y0, w0, u, v, du, dv, x, y, w: integer;
  p: qr_point;
begin
  (*We initialize the buffer with the data mask and XOR bits into it as we read
       them out of the image instead of unmasking in a separate step.*)
  qr_data_mask_fill(_data_bits, _dim, _fmt_info and 7);
  stride := (_dim+QR_INT_BITS-1) shr QR_INT_LOGBITS;
  u0 := 0;

  (*We read data cell-by-cell to avoid having to constantly change which
       projection we're using as we read each bit.
    This (and the position-dependent data mask) is the reason we buffer the
       bits we read instead of converting them directly to codewords here.
    Note that bits are stored column-wise, since that's how we'll scan them.*)
  for j:=0 to Pred(_grid^.ncells) do begin
    u1 := _grid^.cell_limits[j];
    v0 := 0;
    for i:=0 to Pred(_grid^.ncells) do begin
      v1 := _grid^.cell_limits[i];
      cell := _grid^.cells[i]+j;
      du := u0-cell^.u0;
      dv := v0-cell^.v0;
      x0 := cell^.fwd[0][0]*du + cell^.fwd[0][1]*dv + cell^.fwd[0][2];
      y0 := cell^.fwd[1][0]*du + cell^.fwd[1][1]*dv + cell^.fwd[1][2];
      w0 := cell^.fwd[2][0]*du + cell^.fwd[2][1]*dv + cell^.fwd[2][2];
      for u:=u0 to Pred(u1) do begin
        x := x0;
        y := y0;
        w := w0;
        for v:=v0 to Pred(v1) do begin
          (*Skip doing all the divisions and bounds checks if the bit is in the
            function pattern.*)
          if qr_sampling_grid_is_in_fp(_grid,_dim,u,v) = 0 then begin
            qr_hom_cell_fproject(p,cell,x,y,w);
            _data_bits[u*stride+(v shr QR_INT_LOGBITS)] := _data_bits[u*stride+(v shr QR_INT_LOGBITS)] xor (qr_img_get_bit(_img,_width,_height, p[0], p[1]) shl (v and (QR_INT_BITS-1)));
          end;
          Inc(x, cell^.fwd[0][1]);
          Inc(y, cell^.fwd[1][1]);
          Inc(w, cell^.fwd[2][1]);
        end;
        Inc(x0, cell^.fwd[0][0]);
        Inc(y0, cell^.fwd[1][0]);
        Inc(w0, cell^.fwd[2][0]);
      end;
      v0 := v1;
    end;
    u0 := u1;
  end;
end;


(*Arranges the sample bits read by qr_sampling_grid_sample() into bytes and
  groups those bytes into Reed-Solomon blocks.
  The individual block pointers are destroyed by this routine.*)
procedure qr_samples_unpack(_blocks: PPByte;  _nblocks, _nshort_data, _nshort_blocks: integer;  _data_bits, _fp_mask: punsigned;  _dim: integer);
var
  bits, data1, data2, fp_mask1, fp_mask2: unsigned;
  biti, stride, blocki, blockj, i, j, nbits, l: integer;
begin
  stride := (_dim+QR_INT_BITS-1) shr QR_INT_LOGBITS;
  (*If _all_ the blocks are short, don't skip anything (see below).*)
  if _nshort_blocks >= _nblocks then _nshort_blocks := 0;
  (*Scan columns in pairs from right to left.*)
  bits := 0;
  blocki := 0;
  blockj := 0;
  biti := 0;
  j := _dim-1; // column
  while j>0 do begin
    (*Scan up a pair of columns.*)
    nbits := ((_dim-1) and (QR_INT_BITS-1))+1;
    l := j*stride;
    i := stride;
    while i{--} > 0 do begin
      Dec(i);
      data1 := _data_bits[l+i];
      fp_mask1 := _fp_mask[l+i];
      data2 := _data_bits[l+i-stride];
      fp_mask2 := _fp_mask[l+i-stride];
      while nbits{--} > 0 do begin
        Dec(nbits);
        (*Pull a bit from the right column.*)
        if ((fp_mask1 shr nbits) and 1) = 0 then begin
          bits := (bits shl 1) or (data1 shr nbits) and 1;
          Inc(biti);
        end;
        (*Pull a bit from the left column.*)
        if ((fp_mask2 shr nbits) and 1) = 0 then begin
          bits := (bits shl 1) or (data2 shr nbits) and 1;
          Inc(biti);
        end;
        (*If we finished a byte, drop it in a block.*)
        if biti>=8 then begin
          Dec(biti, 8);
          _blocks[blocki{++}]^{++} := Byte(bits shr biti);
          Inc(_blocks[blocki]);
          Inc(blocki);
          (*For whatever reason, the long blocks are at the _end_ of the list,
              instead of the beginning.
            Even worse, the extra bytes they get come at the end of the data
              bytes, before the parity bytes.
            Hence the logic here: when we've filled up the data portion of the
              short blocks, skip directly to the long blocks for the next byte.
            It's also the reason we increment _blocks[blocki] on each store,
              instead of just indexing with blockj (after this iteration the
              number of bytes in each block differs).*)
          if blocki >= _nblocks then begin
            Inc(blockj);
            if {++}blockj = _nshort_data then
              blocki := _nshort_blocks
            else
              blocki := 0;
          end;
        end;
      end;
      nbits := QR_INT_BITS;
    end;
    Dec(j, 2);

    (*Skip the column with the vertical timing pattern.*)
    if j=6 then Dec(j);

    (*Scan down a pair of columns.*)
    l := j*stride;
    for i:=0 to Pred(stride) do begin
      data1 := _data_bits[l+i];
      fp_mask1 := _fp_mask[l+i];
      data2 := _data_bits[l+i-stride];
      fp_mask2 := _fp_mask[l+i-stride];
      nbits := QR_MINI(_dim-(i shl QR_INT_LOGBITS), QR_INT_BITS);
      while nbits{--} > 0 do begin
        Dec(nbits);
        (*Pull a bit from the right column.*)
        if (fp_mask1 and 1) = 0 then begin
          bits := (bits shl 1) or data1 and 1;
          inc(biti);
        end;
        data1 := data1 shr 1;
        fp_mask1 := fp_mask1 shr 1;
        (*Pull a bit from the left column.*)
        if (fp_mask2 and 1) = 0 then begin
          bits := (bits shl 1) or data2 and 1;
          inc(biti);
        end;
        data2 := data2 shr 1;
        fp_mask2 := fp_mask2 shr 1;
        (*If we finished a byte, drop it in a block.*)
        if biti>=8 then begin
          Dec(biti, 8);
          _blocks[blocki{++}]^{++} := Byte(bits shr biti);
          Inc(_blocks[blocki]);
          Inc(blocki);
          (*See comments on the "up" loop for the reason behind this mess.*)
          if blocki>=_nblocks then begin
            Inc(blockj);
            if {++}blockj = _nshort_data then
              blocki := _nshort_blocks
            else
              blocki := 0;
          end;
        end;
      end;
    end;
    Dec(j, 2);
  end;
end;


(*Bit reading code blatantly stolen^W^Wadapted from libogg/libtheora (because
   I've already debugged it and I know it works).
  Portions (C) Xiph.Org Foundation 1994-2008, BSD-style license.*)
type
  qr_pack_buf = record
    buf: pbyte;
    endbyte: integer;
    endbit: integer;
    storage: integer;
end;

procedure qr_pack_buf_init(out _b: qr_pack_buf;  _data: pbyte;  _ndata: integer);
begin
  _b.buf := _data;
  _b.storage := _ndata;
  _b.endbyte := 0;
  _b.endbit := 0;
end;

(*Assumes 0<=_bits<=16.*)
function qr_pack_buf_read(var _b: qr_pack_buf;  _bits: integer): integer;
var
  p: PByte;
  ret: unsigned;
  m, d: integer;
begin
  m := 16-_bits;
  Inc(_bits, _b.endbit);
  d := _b.storage-_b.endbyte;
  if d<=2 then begin
    (*Not the main path.*)
    if d*8<_bits then begin
      Inc(_b.endbyte, _bits shr 3);
      _b.endbit := _bits and 7;
      Exit(-1);
    end
    (*Special case to avoid reading p[0] below, which might be past the end of
           the buffer; also skips some useless accounting.*)
    else if _bits = 0 then
      Exit(0);
  end;
  p := _b.buf+_b.endbyte;
  ret := p[0] shl (8+_b.endbit);
  if _bits>8 then begin
    ret := ret or (p[1] shl _b.endbit);
    if _bits>16 then
      ret := ret or (p[2] shr (8-_b.endbit));
  end;
  Inc(_b.endbyte, _bits shr 3);
  _b.endbit := _bits and 7;
  Result := (ret and $FFFF) shr m;
end;

function qr_pack_buf_avail(constref _b: qr_pack_buf): integer;
begin
  Result := ((_b.storage-_b.endbyte) shl 3) - _b.endbit;
end;


(*The characters available in QR_MODE_ALNUM.*)
const
  QR_ALNUM_TABLE: array [0..44] of char = ('0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',' ','$','%','*','+','-','.','/',':');

function qr_code_data_parse(var _qrdata: qr_code_data;  _version: integer;  _data: pbyte;  _ndata: integer): integer;
const
  (*The number of bits used to encode the character count for each version
    range and each data mode.*)
  LEN_BITS: array [0..2,0..3] of Byte = ((10,9,8,8),(12,11,16,10),(14,13,16,12));
var
  qpb: qr_pack_buf;
  centries, len_bits_idx: integer;
  entry: pqr_code_data_entry;
  mode: qr_mode;
  buf: pbyte;
  self_parity, val, bits, c: unsigned;
  bits2, len, count, rem: integer;
begin
  (*Entries are stored directly in the struct during parsing.
    Caller cleans up any allocated data on failure.*)
  _qrdata.entries := nil;
  _qrdata.nentries := 0;
  _qrdata.sa_size := 0;
  self_parity := 0;
  centries := 0;
  (*The versions are divided into 3 ranges that each use a different number of
       bits for length fields.*)
  len_bits_idx := Ord(_version>9) + Ord(_version>26);

  qr_pack_buf_init(qpb, _data, _ndata);
  (*While we have enough bits to read a mode...*)
  while qr_pack_buf_avail(qpb) >= 4 do begin
    mode := qr_mode(qr_pack_buf_read(qpb,4));
    (*Mode 0 is a terminator.*)
    if mode = QR_MODE_TERM then break;
    if _qrdata.nentries >= centries then begin
      centries := (centries shl 1) or 1;
      _qrdata.entries := ReallocMem(_qrdata.entries, centries*sizeof(_qrdata.entries^));
    end;
    entry := _qrdata.entries + _qrdata.nentries{++}; Inc(_qrdata.nentries);
    (*Set the mode to an invalid value until we allocate a buffer for it.
      This ensures we don't try to free it on clean-up until then.*)
    //entry^.mode := QR_MODE_INVALID;
    entry^.mode := mode;
    (*Set the buffer to NULL, because if parsing fails, we might try to free it
       on clean-up.*)
    entry^.payload.data.buf := nil;
    case mode of
      QR_MODE_NUM:
        begin
          len := qr_pack_buf_read(qpb, LEN_BITS[len_bits_idx][0]);
          if len<0 then Exit(-1);
          (*Check to see if there are enough bits left now, so we don't have to
            in the decode loop.*)
          count := len div 3;
          rem := len mod 3;
          if qr_pack_buf_avail(qpb) < 10*count + 7*((rem shr 1) and 1) + 4*(rem and 1) then Exit(-1);
          buf := GetMem(len*sizeof(buf^));
          entry^.payload.data.buf := buf;
          entry^.payload.data.len := len;
          (*Read groups of 3 digits encoded in 10 bits.*)
          while count{--} > 0 do begin
            Dec(count);
            bits := qr_pack_buf_read(qpb,10);
            if bits>=1000 then Exit(-1);
            c := Ord('0') + bits div 100;
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
            bits := bits mod 100;
            c := Ord('0') + bits div 10;
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
            c := Ord('0') + bits mod 10;
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
          end;
          (*Read the last two digits encoded in 7 bits.*)
          if rem>1 then begin
            bits := qr_pack_buf_read(qpb, 7);
            if bits>=100 then Exit(-1);
            c := Ord('0') + bits div 10;
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
            c := Ord('0') + bits mod 10;
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
          end
          (*Or the last one digit encoded in 4 bits.*)
          else if rem>0 then begin
            bits := qr_pack_buf_read(qpb, 4);
            if bits>=10 then Exit(-1);
            c := Ord('0') + bits;
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
          end;
        end;

      QR_MODE_ALNUM:
        begin
          len := qr_pack_buf_read(qpb, LEN_BITS[len_bits_idx][1]);
          if len<0 then Exit(-1);
          (*Check to see if there are enough bits left now, so we don't have to
              in the decode loop.*)
          count := len shr 1;
          rem := len and 1;
          if qr_pack_buf_avail(qpb) < 11*count+6*rem then Exit(-1);
          buf := GetMem(len*sizeof(buf^));
          entry^.payload.data.buf := buf;
          entry^.payload.data.len := len;
          (*Read groups of two characters encoded in 11 bits.*)
          while count{--} > 0 do begin
            Dec(count);
            bits := qr_pack_buf_read(qpb, 11);
            if bits >= 2025 then Exit(-1);
            c := Ord(QR_ALNUM_TABLE[bits div 45]);
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
            c := Ord(QR_ALNUM_TABLE[bits mod 45]);
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
            Dec(len, 2);
          end;
          (*Read the last character encoded in 6 bits.*)
          if rem<>0 then begin
            bits := qr_pack_buf_read(qpb, 6);
            if bits>=45 then Exit(-1);
            buf^{++} := Ord(QR_ALNUM_TABLE[bits]);
            Inc(buf);
          end;
        end;

      QR_MODE_STRUCT:
        (*Structured-append header.*)
        begin
          bits2 := qr_pack_buf_read(qpb,16);
          if bits2<0 then Exit(-1);
          (*We save a copy of the data in _qrdata for easy reference when
           grouping structured-append codes.
          If for some reason the code has multiple S-A headers, first one wins,
           since it is supposed to come before everything else (TODO: should we
           return an error instead?).*)
          if _qrdata.sa_size = 0 then begin
            entry^.payload.sa.sa_index  := (bits2 shr 12 and $F);
            entry^.payload.sa.sa_size   := ((bits2 shr 8 and $F) + 1);
            entry^.payload.sa.sa_parity := (bits2 and $FF);
            _qrdata.sa_index  := entry^.payload.sa.sa_index;
            _qrdata.sa_size   := entry^.payload.sa.sa_size;
            _qrdata.sa_parity := entry^.payload.sa.sa_parity;
          end;
        end;

      QR_MODE_BYTE:
        begin
          len := qr_pack_buf_read(qpb, LEN_BITS[len_bits_idx][2]);
          if len<0 then Exit(-1);
          (*Check to see if there are enough bits left now, so we don't have to
            in the decode loop.*)
          if qr_pack_buf_avail(qpb) < len shl 3 then Exit(-1);
          buf := GetMem(len*sizeof(buf^));
          entry^.payload.data.buf := buf;
          entry^.payload.data.len := len;
          while len{--} > 0 do begin
            Dec(len);
            c := qr_pack_buf_read(qpb,8);
            self_parity := self_parity xor c;
            buf^{++} := c;
            Inc(buf);
          end;
        end;

      QR_MODE_FNC1_1ST:
      (*FNC1 first position marker.*)
        begin
        end;

      QR_MODE_ECI:
      (*Extended Channel Interpretation data.*)
        begin
          bits2 := qr_pack_buf_read(qpb,8);
          if bits2<0 then Exit(-1);
          (*One byte:*)
          if (bits2 and $80) = 0 then
            val := bits2
          (*Two bytes:*)
          else if (bits2 and $40) = 0 then begin
            val := bits2 and ($3F shl 8);
            bits2 := qr_pack_buf_read(qpb,8);
            if bits2<0 then Exit(-1);
            val := val or bits2;
          end
          (*Three bytes:*)
          else if (bits2 and $20) = 0 then begin
            val := bits2 and ($1F shl 16);
            bits2 := qr_pack_buf_read(qpb,16);
            if bits2<0 then Exit(-1);
            val := val or bits2;
            (*Valid ECI values are 0...999999.*)
            if val>=1000000 then Exit(-1);
          end
          (*Invalid lead byte.*)
          else
            Exit(-1);
          entry^.payload.eci := val;
        end;

      QR_MODE_KANJI:
        begin
          len := qr_pack_buf_read(qpb, LEN_BITS[len_bits_idx][3]);
          if len<0 then Exit(-1);
          (*Check to see if there are enough bits left now, so we don't have to
              in the decode loop.*)
          if qr_pack_buf_avail(qpb) < 13*len then Exit(-1);
          buf := GetMem(2*len*sizeof(buf^));
          entry^.payload.data.buf := buf;
          entry^.payload.data.len := 2*len;
          (*Decode 2-byte SJIS characters encoded in 13 bits.*)
          while len{--} > 0 do begin
            Dec(len);
            bits := qr_pack_buf_read(qpb,13);
            bits := (((bits div $C0) shl 8) or (bits mod $C0)) + $8140;
            if bits >= $A000 then Inc(bits, $4000);
            (*TODO: Are values 0xXX7F, 0xXXFD...0xXXFF always invalid?
                        Should we reject them here?*)
            self_parity := self_parity xor bits;
            buf^{++} := (bits shr 8);
            Inc(buf);
            buf^{++} := (bits and $FF);
            Inc(buf);
          end;
        end;

      QR_MODE_FNC1_2ND:
      (*FNC1 second position marker.*)
        begin
          (*FNC1 in the 2nd position encodes an Application Indicator in one
             byte, which is either a letter (A...Z or a...z) or a 2-digit number.
            The letters are encoded with their ASCII value plus 100, the numbers
             are encoded directly with their numeric value.
            Values 100...164, 191...196, and 223...255 are invalid, so we reject
             them here.*)
          bits := qr_pack_buf_read(qpb,8);
          if not((bits>=0) and (bits<100) or (bits>=165) and (bits<191) or (bits>=197) and (bits<223)) then
            Exit(-1);
          entry^.payload.ai := bits;
        end;

      else
      (*Unknown mode number:*)
        begin
          (*Unfortunately, because we have to understand the format of a mode to
                     know how many bits it occupies, we can't skip unknown modes.
                    Therefore we have to fail.*)
          Exit(-1);
        end;
    end;{case?}
  end;
  (*Store the parity of the data from this code, for S-A.
    The final parity is the 8-bit XOR of all the decoded bytes of literal data.
    We don't combine the 2-byte kanji codes into one byte in the loops above,
     because we can just do it here instead.*)
  _qrdata.self_parity := ((self_parity shr 8) xor self_parity) and $FF;
  (*Success.*)
  _qrdata.entries := ReallocMem(_qrdata.entries, _qrdata.nentries*sizeof(_qrdata.entries^));
  (*TODO: If there was a S-A header, we should compute the parity of this
       code; how are non-data modes handled (ECI, FNC1)?*)
  Result := 0;
end;

(*Check if a mode has a data buffer associated with it.
  Currently this is only modes with exactly one bit set.*)
//#define QR_MODE_HAS_DATA(_mode) (!((_mode)&(_mode)-1))
function QR_MODE_HAS_DATA(_mode: qr_mode): boolean;
begin
  Result := (Ord(_mode) and (Ord(_mode)-1)) = 0;
end;

procedure qr_code_data_clear(var _qrdata: qr_code_data);
var
  i: integer;
begin
  for i:=0 to Pred(_qrdata.nentries) do begin
    if QR_MODE_HAS_DATA(_qrdata.entries[i].mode) then
      FreeMem(_qrdata.entries[i].payload.data.buf);
  end;
  FreeMem(_qrdata.entries);
end;


(*The total number of codewords in a QR code.*)
function qr_code_ncodewords(_version: unsigned): integer;
var
  nalign: unsigned;
begin
  (*This is 24-27 instructions on ARM in thumb mode, or a 26-32 byte savings
    over just using a table (not counting the instructions that would be
    needed to do the table lookup).*)
  if _version = 1 then
    Result := 26
  else begin
    nalign := (_version div 7)+2;
    Result := ((_version shl 4)*(_version+8)-(5*nalign)*(5*nalign-2)+36*Ord(_version<7)+83) shr 3;
  end;
end;


const
  (*Bulk data for the number of parity bytes per Reed-Solomon block.*)
  QR_RS_NPAR_VALS: array [0..70] of byte = (
    7,10,13,17,
    10,16,22,28,26,26,26,22,24,22,22,26,24,18,22,
    15,26,18,22,24,30,24,20,24,
    18,16,24,28,28,28,28,30,24,
    20,18,18,26,24,28,24,30,26,28,28,26,28,30,30,22,20,24,
    20,18,26,16,
    20,30,28,24,22,26,28,26,30,28,30,30);

  (*An offset into QR_RS_NPAR_DATA for each version that gives the number of
    parity bytes per Reed-Solomon block for each error correction level.*)
  QR_RS_NPAR_OFFS: array [0..39] of byte = (0,4,19,55,15,28,37,12,51,39,59,62,10,24,22,41,31,44,7,65,47,33,67,67,48,32,67,67,67,67,67,67,67,67,67,67,67,67,67,67);

  (*The number of Reed-Solomon blocks for each version and error correction level.*)
  QR_RS_NBLOCKS: array [0..39,0..3] of byte = ((1,1,1,1),(1,1,1,1),(1,1,2,2),(1,2,2,4),(1,2,4,4),(2,4,4,4),(2,4,6,5),(2,4,6,6),(2,5,8,8),(4,5,8,8),(4,5,8,11),(4,8,10,11),(4,9,12,16),(4,9,16,16),(6,10,12,18),(6,10,17,16),(6,11,16,19),(6,13,18,21),(7,14,21,25),(8,16,20,25),(8,17,23,25),(9,17,23,34),(9,18,25,30),(10,20,27,32),(12,21,29,35),(12,23,34,37),(12,25,34,40),(13,26,35,42),(14,28,38,45),(15,29,40,48),(16,31,43,51),(17,33,45,54),(18,35,48,57),(19,37,51,60),(19,38,53,63),(20,40,56,66),(21,43,59,70),(22,45,62,74),(24,47,65,77),(25,49,68,81));

(*Attempts to fully decode a QR code.
  _qrdata:   Returns the parsed code data.
  _gf:       Used for Reed-Solomon error correction.
  _ul_pos:   The location of the UL finder pattern.
  _ur_pos:   The location of the UR finder pattern.
  _dl_pos:   The location of the DL finder pattern.
  _version:  The (decoded) version number.
  _fmt_info: The decoded format info.
  _img:      The binary input image.
  _width:    The width of the input image.
  _height:   The height of the input image.
  Return: 0 on success, or a negative value on error.
*)
function qr_code_decode(var _qrdata: qr_code_data;  _gf: prs_gf256;  const _ul_pos, _ur_pos, _dl_pos: qr_point;  _version, _fmt_info: integer;  _img: pbyte;  _width, _height: integer): integer;
var
  grid: qr_sampling_grid;
  data_bits: punsigned;
  blocks: PPByte;
  block_data: PByte;
  nblocks, nshort_blocks, ncodewords: integer;
  block_sz, block_szi: integer;
  ecc_level: integer;
  i, ndata, npar, dim, ndatai: integer;
begin
  if (_version < 1) or (_version > 40) then Exit(-1);
  // _qrdata.bbox
  qr_sampling_grid_init(grid, _version, _ul_pos,_ur_pos,_dl_pos, @_qrdata.bbox, _img, _width, _height);
  dim := 17 + (_version shl 2);
  data_bits := GetMem(dim * ((dim+QR_INT_BITS-1) shr QR_INT_LOGBITS) * SizeOf(data_bits^));
  qr_sampling_grid_sample(@grid, data_bits, dim, _fmt_info, _img, _width, _height); // _img -> data_bits
  (*Group those bits into Reed-Solomon codewords.*)
  ecc_level := (_fmt_info shr 3) xor 1;
  nblocks := QR_RS_NBLOCKS[_version-1, ecc_level];
  npar := QR_RS_NPAR_VALS[QR_RS_NPAR_OFFS[_version-1]+ecc_level];
  ncodewords := qr_code_ncodewords(_version);
  block_sz := ncodewords div nblocks;
  nshort_blocks := nblocks - (ncodewords mod nblocks);
  blocks := GetMem(nblocks*SizeOf(blocks^));
  block_data := GetMem(ncodewords*SizeOf(block_data^));
  blocks[0] := block_data;
  for i:=1 to Pred(nblocks) do
    blocks[i] := blocks[i-1] + block_sz + Ord(i>nshort_blocks);
  qr_samples_unpack(blocks, nblocks, block_sz-npar, nshort_blocks, data_bits, grid.fpmask, dim); // _data_bits -> block_data
  qr_sampling_grid_clear(grid);
  FreeMem(blocks);
  FreeMem(data_bits);

  (*Perform the error correction.*)
  ndata := 0;
  ncodewords := 0;
  Result := 0;
  for i:=0 to Pred(nblocks) do begin
    block_szi := block_sz + Ord(i>=nshort_blocks);
    Result := rs_correct(_gf, QR_M0, block_data+ncodewords, block_szi, npar, nil, 0);
    (* For version 1 symbols and version 2-L and 3-L symbols, we aren't allowed
        to use all the parity bytes for correction.
       They are instead used to improve detection.
       Version 1-L reserves 3 parity bytes for detection.
       Versions 1-M and 2-L reserve 2 parity bytes for detection.
       Versions 1-Q, 1-H, and 3-L reserve 1 parity byte for detection.
       We can ignore the version 3-L restriction because it has an odd number of
        parity bytes, and we don't support erasure detection. *)
    if (Result<0) or ((_version=1) and (Result>(ecc_level+1)<<1)) or ((_version=2) and (ecc_level=0) and (Result>4)) then begin
      Result := -1;
      break;
    end;
    ndatai := block_szi-npar;
    Move((block_data+ncodewords)^, (block_data+ndata)^, ndatai*sizeof(block_data^));
    Inc(ncodewords, block_szi);
    Inc(ndata, ndatai);
  end;
  (*Parse the corrected bitstream.*)
  if Result >= 0 then begin
    Result := qr_code_data_parse(_qrdata, _version, block_data, ndata); // block_data -> _qrdata.entries
    (*We could return any partially decoded data, but then we'd have to have
      API support for that; a mode ignoring ECC errors might also be useful.*)
    if Result < 0 then
      qr_code_data_clear(_qrdata);
    _qrdata.version := _version;
    _qrdata.ecc_level := ecc_level;
  end
  else begin
    _qrdata.nentries:=0;
    _qrdata.entries:=nil;
  end;
  FreeMem(block_data);
end;


{$IFDEF HAVE_zbar_qr_decode}
function _zbar_qr_create(): pqr_reader; forward;
procedure _zbar_qr_destroy(reader: pqr_reader); forward;
procedure _zbar_qr_reset(reader: pqr_reader); forward;
function _zbar_qr_decode(reader: pqr_reader;  iscn: pzbar_image_scanner_t;  img: pzbar_image_t): integer; forward;

(**********
 * util.c *
 **********)
(*Computes floor(sqrt(_val)) exactly.*)
function qr_isqrt(_val: unsigned): unsigned;
var
  b,t: unsigned;
  bshift: integer;
begin
 (*Uses the second method from
    http://www.azillionmonkeys.com/qed/sqroot.html
   The main idea is to search for the largest binary digit b such that
    (g+b)*(g+b) <= _val, and add it to the solution g.*)
 Result:=0;
 b:=$8000;
 bshift:=16;
 while bshift>0 do begin
   Dec(bshift);
   t:=((Result shl 1) + b) shl bshift;
   if(t<=_val) then begin
     Inc(Result, b);
     Dec(_val, t);
   end;
   b := b shr 1;
 end;
end;

(* Computes sqrt(_x*_x+_y*_y) *)
function qr_ihypot(_x,_y: integer): integer;
var x,y: int64;
begin
  x := _x;
  y := _y;
  //DebugLn('qr_ihypot: %d^2 + %d^2 = %d',[_x,_y, Sqr(x)+Sqr(y)]);
  Result := Round(Sqrt(Sqr(x)+Sqr(y)));
end;


(***********
 * error.c *
 ***********)
function zbar_version(var major, minor: cardinal): integer;
const
  ZBAR_VERSION_MAJOR = 0;
  ZBAR_VERSION_MINOR = 10;
begin
  major := ZBAR_VERSION_MAJOR;
  minor := ZBAR_VERSION_MINOR;
  Result := 0;
end;

procedure zbar_set_verbosity(level: integer);
begin
  //_zbar_verbosity := level;
end;

procedure err_cleanup(var err: errinfo_t);
begin
    assert(err.magic = $5252457a{ERRINFO_MAGIC});
    if Assigned(err.buf) then begin
        FreeMem(err.buf);
        err.buf := nil;
    end;
    if Assigned(err.arg_str) then begin
        FreeMem(err.arg_str);
        err.arg_str := nil;
    end;
end;

(************
 * refcnt.h *
 ************)
function _zbar_refcnt(var cnt: refcnt_t; delta: integer): integer; inline;
begin
  Result := -1;
  if (delta > 0) then
    while delta{--}<>0 do begin
      Dec(delta);
      // Increments (increases by one) the value of the specified 32-bit variable as an atomic operation.
      // The function returns the resulting incremented value
      //rc = InterlockedIncrement(cnt);
      Inc(cnt);
      Result := cnt;
    end
  else if (delta < 0) then
    while delta{++}<>0 do begin
      Inc(Delta);
      //rc = InterlockedDecrement(cnt);
      Dec(cnt);
      Result := cnt;
    end;
  Assert(Result >= 0);
end;

(************
 * symbol.c *
 ************)
procedure _zbar_symbol_free(sym: pzbar_symbol_t); forward;

procedure _zbar_symbol_refcnt(sym: pzbar_symbol_t; delta: integer); inline;
begin
  if (_zbar_refcnt(sym^.refcnt, delta)=0) and (delta <= 0) then
    _zbar_symbol_free(sym);
end;

procedure _zbar_symbol_set_free(syms: pzbar_symbol_set_t); inline;
var
  sym, next: pzbar_symbol_t;
begin
  sym := syms^.head;
  while sym<>nil do begin
    next := sym^.next;
    sym^.next := nil;
    _zbar_symbol_refcnt(sym, -1);
    sym := next;
  end;
  syms^.head := nil;
  FreeMem(syms);
end;

procedure zbar_symbol_set_ref(syms: pzbar_symbol_set_t; delta: integer);
var
  ncsyms: pzbar_symbol_set_t;
begin
  ncsyms := syms;
  if (_zbar_refcnt(ncsyms^.refcnt, delta)=0) and (delta <= 0) then
    _zbar_symbol_set_free(ncsyms);
end;

procedure _zbar_symbol_free(sym: pzbar_symbol_t);
begin
  if Assigned(sym^.syms) then begin
     zbar_symbol_set_ref(sym^.syms, -1);
     sym^.syms := nil;
  end;

  if Assigned(sym^.pts) then
    FreeMem(sym^.pts);
  if (sym^.data_alloc<>0) and Assigned(sym^.data) then
    FreeMem(sym^.data);
  FreeMem(sym);
  sym:=nil;
end;

function _zbar_symbol_set_create(): pzbar_symbol_set_t;
begin
  Result := AllocMem(sizeof(Result^));
  _zbar_refcnt(Result^.refcnt, 1);
end;

procedure sym_add_point(var sym: zbar_symbol_t; x,y: integer);
var i: integer;
begin
  i := sym.npts;
  Inc(sym.npts);
  if (sym.npts >= sym.pts_alloc) then begin
    Inc(sym.pts_alloc);
    sym.pts := ReallocMem(sym.pts, sym.pts_alloc * sizeof(point_t));
  end;
  sym.pts[i].x := x;
  sym.pts[i].y := y;
end;

function zbar_symbol_get_type(sym: pzbar_symbol_t): zbar_symbol_type_t;
begin
    Result := sym^.&type;
end;

function zbar_symbol_get_loc_x(sym: pzbar_symbol_t; idx: unsigned): integer;
begin
    if (idx < sym^.npts) then
        Result :=  sym^.pts[idx].x
    else
        Result := -1;
end;
function zbar_symbol_get_loc_y (sym: pzbar_symbol_t; idx: unsigned): integer;
begin
    if (idx < sym^.npts) then
        Result :=  sym^.pts[idx].y
    else
        Result := -1;
end;

function zbar_symbol_get_data(sym: pzbar_symbol_t): PAnsiChar;
begin
    Result := PAnsiChar(sym^.data);
end;

function zbar_symbol_next(sym: pzbar_symbol_t): pzbar_symbol_t;
begin
    if sym<>nil then
      Result := sym^.next
    else
      Result := nil;
end;

function zbar_get_symbol_name(sym: zbar_symbol_type_t): PAnsiChar;
begin
  Result := 'QR-Code';
end;


(***********
 * image.c *
 ***********)
procedure _zbar_image_free(img: pzbar_image_t);
begin
    if Assigned(img^.syms) then begin
        zbar_symbol_set_ref(img^.syms, -1);
        img^.syms := nil;
    end;
    FreeMem(img);
end;

procedure _zbar_image_refcnt(var img: zbar_image_t; delta: integer);
begin
  if (_zbar_refcnt(img.refcnt, delta)=0) and (delta <= 0) then begin
    if Assigned(img.cleanup) then img.cleanup(@img);
    if not Assigned(img.src) then _zbar_image_free(@img);
  end;
end;

procedure zbar_image_destroy(img: pzbar_image_t);
begin
    _zbar_image_refcnt(img^, -1);
end;

function zbar_image_create(): pzbar_image_t;
begin
  Result := AllocMem(sizeof(zbar_image_t));
  //_zbar_refcnt_init(); // calls pthread_once()
  _zbar_image_refcnt(Result^, 1);
  Result^.srcidx := -1;
end;

procedure zbar_image_set_format(img: pzbar_image_t; fmt: PAnsiChar);
begin
  img^.format := PLongWord(fmt)^;
end;
function zbar_image_get_format(img: pzbar_image_t): unsigned;
begin
  Result := img^.format;
end;

procedure zbar_image_set_size(img: pzbar_image_t; w,h: unsigned);
begin
  img^.crop_x := 0;
  img^.crop_y := 0;
  img^.width  := w;
  img^.crop_w := w;
  img^.height := h;
  img^.crop_h := h;
end;

procedure zbar_image_set_crop (img: pzbar_image_t; x, y, w, h: unsigned);
var
  img_w, img_h: unsigned;
begin
    img_w := img^.width;
    if (x > img_w) then x := img_w;
    if (x + w > img_w) then w := img_w - x;
    img^.crop_x := x;
    img^.crop_w := w;

    img_h := img^.height;
    if (y > img_h) then y := img_h;
    if (y + h > img_h) then h := img_h - y;
    img^.crop_y := y;
    img^.crop_h := h;
end;

procedure zbar_image_free_data(img: pzbar_image_t);
var
  newimg: pzbar_image_t;
  cleanup: zbar_image_cleanup_handler_t;
begin
    if (img=nil) then Exit;

    if img^.src<>nil then begin
        (* replace video image w/new copy *)
        Assert(img^.refcnt<>0); (* FIXME needs lock *)
        newimg := zbar_image_create();
        Move(img^, newimg^, sizeof(zbar_image_t));
        (* recycle video image *)
        newimg^.cleanup(newimg);
        (* detach old image from src *)
        img^.cleanup := nil;
        img^.src := nil;
        img^.srcidx := -1;
    end
    else if Assigned(img^.cleanup) and Assigned(img^.data) then begin
        if (img^.cleanup <> @zbar_image_free_data) then begin
            (* using function address to detect this case is a bad idea;
             * windows link libraries add an extra layer of indirection...
             * this works around that problem (bug #2796277)
             *)
            cleanup := img^.cleanup;
            img^.cleanup := @zbar_image_free_data;
            cleanup(img);
        end
        else
            FreeMem(img^.data);
    end;
    img^.data := nil;
end;

procedure zbar_image_set_data(img: pzbar_image_t; data: Pointer; len: unsigned; cleanup: zbar_image_cleanup_handler_t);
begin
  zbar_image_free_data(img);
  img^.data := data;
  img^.datalen := len;
  img^.cleanup := cleanup;
end;

procedure _zbar_image_swap_symbols(a,b: pzbar_image_t); inline;
var tmp: pzbar_symbol_set_t;
begin
    tmp := a^.syms;
    a^.syms := b^.syms;
    b^.syms := tmp;
end;

function zbar_image_first_symbol(img: pzbar_image_t): pzbar_symbol_t;
begin
    if Assigned(img^.syms) then
      Result := img^.syms^.head
    else
      Result := nil;
end;


(*************
 * decoder.c *
 *************)
(* reset QR finder specific state *)
procedure qr_finder_reset(var qrf: qr_finder_t); inline;
begin
  qrf.s5 := 0;
end;

procedure zbar_decoder_reset(dcode: pzbar_decoder_t);
begin
    FillByte(dcode^, @dcode^.buf_alloc - dcode, 0);
    qr_finder_reset(dcode^.qrf);
end;

function zbar_decoder_create(): pzbar_decoder_t;
const BUFFER_MIN=$20;
begin
    Result := AllocMem(sizeof(zbar_decoder_t));
    Result^.buf_alloc := BUFFER_MIN;
    Result^.buf := GetMem(Result^.buf_alloc);

    (* initialize default configs *)
    Result^.qrf.config := 1 shl Ord(ZBAR_CFG_ENABLE);

    zbar_decoder_reset(Result);
end;

procedure zbar_decoder_destroy(dcode: pzbar_decoder_t);
begin
    if Assigned(dcode^.buf) then
        FreeMem(dcode^.buf);
    FreeMem(dcode);
end;

procedure zbar_decoder_new_scan(dcode: pzbar_decoder_t); //ok
begin
    (* soft reset decoder *)
    FillByte(dcode^.w, sizeof(dcode^.w), 0);
    dcode^.lock := ZBAR_NONE;
    dcode^.idx := 0;
    dcode^.s6 := 0;
    qr_finder_reset(dcode^.qrf);
end;

function zbar_decoder_set_handler(dcode: pzbar_decoder_t; handler: zbar_decoder_handler_t): zbar_decoder_handler_t;
begin
    Result := dcode^.handler;
    dcode^.handler := handler;
end;

procedure zbar_decoder_set_userdata(dcode: pzbar_decoder_t; userdata: pointer);
begin
    dcode^.userdata := userdata;
end;
function zbar_decoder_get_userdata(dcode: pzbar_decoder_t): pointer;
begin
    Result := dcode^.userdata;
end;

function zbar_decoder_get_type(dcode: pzbar_decoder_t): zbar_symbol_type_t;
begin
    Result := dcode^.&type;
end;

(* retrieve i-th previous element width *)
function get_width(dcode: pzbar_decoder_t; offset: byte): unsigned;
begin
    Result := dcode^.w[(dcode^.idx - offset) and (DECODE_WINDOW - 1)];
end;

(* return current element color *)
function get_color(dcode: pzbar_decoder_t): byte; inline;
begin
    Result := dcode^.idx and 1;
end;

function decode_e(e,s,n: unsigned): integer;
var _E: integer;
begin
    (* result is encoded number of units - 2
     * (for use as zero based index)
     * or -1 if invalid
     *)
    _E := ((e * n * 2 + 1) div s - 3) div 2;
    if _E >= n - 3 then
      Result := -1
    else
      Result := _E;
end;

(* retrieve bar+space pair width starting at offset i *)
function pair_width(dcode: pzbar_decoder_t; offset: Byte): unsigned;
begin
    Result := get_width(dcode, offset) + get_width(dcode, offset + 1);
end;

function _zbar_find_qr(dcode: pzbar_decoder_t): zbar_symbol_type_t;
var
  qrf: pqr_finder_t;
  s, qz, w: unsigned;
  ei: integer;
begin
    qrf := @dcode^.qrf;

    (* update latest finder pattern width *)
    Dec(qrf^.s5, get_width(dcode, 6));
    Inc(qrf^.s5, get_width(dcode, 1));
    s := qrf^.s5;
    //DebugLn('find_qr: s=%d; pair1=%d, pair2=%d, pair3=%d',[s, pair_width(dcode, 1), pair_width(dcode, 2), pair_width(dcode, 3)]);

    (*TODO: The 2005 standard allows reflectance-reversed codes (light on dark
       instead of dark on light).
      If we find finder patterns with the opposite polarity, we should invert
       the final binarized image and use them to search for QR codes in that.*)
    if (get_color(dcode) <> Ord(ZBAR_SPACE)) or (s < 7) then
        Exit(ZBAR_NONE);

    ei := decode_e(pair_width(dcode, 1), s, 7);
    if (ei <> 0) then Exit(ZBAR_NONE);

    ei := decode_e(pair_width(dcode, 2), s, 7);
    if (ei <> 2) then Exit(ZBAR_NONE);

    ei := decode_e(pair_width(dcode, 3), s, 7);
    if(ei <> 2) then Exit(ZBAR_NONE);

    ei := decode_e(pair_width(dcode, 4), s, 7);
    if (ei <> 0) then Exit(ZBAR_NONE);

    (* valid QR finder symbol
     * mark positions needed by decoder
     *)
    qz := get_width(dcode, 0);
    w := get_width(dcode, 1);
    qrf^.line.eoffs := qz + (w + 1) div 2;
    qrf^.line.len := qz + w + get_width(dcode, 2);
    qrf^.line.pos[0] := qrf^.line.len + get_width(dcode, 3);
    qrf^.line.pos[1] := qrf^.line.pos[0];
    w := get_width(dcode, 5);
    qrf^.line.boffs := qrf^.line.pos[0] + get_width(dcode, 4) + (w + 1) div 2;

    dcode^.direction := 0;
    dcode^.buflen := 0;
    Result := ZBAR_QRCODE_;
end;

function zbar_decode_width(dcode: pzbar_decoder_t; w: unsigned): zbar_symbol_type_t;  //ok
var
  tmp: zbar_symbol_type_t;
begin
    Result := ZBAR_NONE;

    dcode^.w[dcode^.idx and (DECODE_WINDOW - 1)] := w;
    //DebugLn('    decode[%d]: w=%d (%.1f); s6=%d', [dcode^.idx, w, w/32, dcode^.s6]);

    (* update shared character width *)
    Dec(dcode^.s6, get_width(dcode, 7));
    Inc(dcode^.s6, get_width(dcode, 1));

    (* each decoder processes width stream in parallel *)

    // #define TEST_CFG(config, cfg) (((config) >> (cfg)) & 1)
    if ({TEST_CFG}((dcode^.qrf.config{config} shr Ord(ZBAR_CFG_ENABLE){cfg}) and 1) <> 0) then begin
      tmp := _zbar_find_qr(dcode);
      if (tmp > ZBAR_PARTIAL) then
        Result := tmp;
    end;

    {$PUSH}
    {$R-}
    Inc(dcode^.idx);
    {$POP}
    dcode^.&type := Result;
    if (Result <> ZBAR_NONE) then begin
        //if (dcode^.lock<>0) and (Result > ZBAR_PARTIAL) and (Result <> ZBAR_QRCODE) then
        //    release_lock(dcode, sym);
        if Assigned(dcode^.handler) then
            dcode^.handler(dcode);
    end;
end;

(*************
 * scanner.c *
 *************)
const
  ZBAR_FIXED=5;
  ZBAR_SCANNER_THRESH_MIN=4;
  ZBAR_SCANNER_THRESH_INIT_WEIGHT=0.44;
  THRESH_INIT = Trunc((ZBAR_SCANNER_THRESH_INIT_WEIGHT * (1 shl (ZBAR_FIXED + 1)) + 1) / 2);
  ZBAR_SCANNER_THRESH_FADE = 8;
  ZBAR_SCANNER_EWMA_WEIGHT = 0.78;
  EWMA_WEIGHT:unsigned = Trunc((ZBAR_SCANNER_EWMA_WEIGHT * (1 shl (ZBAR_FIXED + 1)) + 1) / 2);

procedure zbar_scanner_memset(var scn: zbar_scanner_t);
begin
  scn.x := 0;
  FillByte(scn.y0, Sizeof(scn.y0), 0);
  scn.y1_sign:=0;
  scn.y1_thresh:=0;
  scn.cur_edge:=0;
  scn.last_edge:=0;
  scn.width:=0;
end;

function zbar_scanner_reset(var scn: zbar_scanner_t): zbar_symbol_type_t;
begin
    zbar_scanner_memset(scn);
    scn.y1_thresh := scn.y1_min_thresh;
    if Assigned(scn.decoder) then
        zbar_decoder_reset(scn.decoder);
    Result := ZBAR_NONE;
end;

function zbar_scanner_create (dcode: pzbar_decoder_t): pzbar_scanner_t;
begin
    Result := GetMem(sizeof(zbar_scanner_t));
    Result^.decoder := dcode;
    Result^.y1_min_thresh := ZBAR_SCANNER_THRESH_MIN;
    zbar_scanner_reset(Result^);
end;

procedure zbar_scanner_destroy(scn: pzbar_scanner_t);
begin
    FreeMem(scn);
end;

function zbar_scanner_get_edge(scn: pzbar_scanner_t; offset: unsigned; prec: integer): unsigned;
var edge: unsigned;
begin
    // #define ROUND (1 << (ZBAR_FIXED - 1))
    edge := scn^.last_edge - offset - (1 shl ZBAR_FIXED) - {ROUND}(1 shl (ZBAR_FIXED - 1));
    prec := ZBAR_FIXED - prec;
    if (prec > 0) then
        Result := edge shr prec
    else if (prec=0) then
        Result := edge
    else
        Result := edge shl -prec;
end;

function process_edge(scn: pzbar_scanner_t; y1: integer): zbar_symbol_type_t;
begin
    if (scn^.y1_sign=0) then begin
        scn^.cur_edge := (1 shl ZBAR_FIXED) + {ROUND}(1 shl (ZBAR_FIXED - 1));
        scn^.last_edge := scn^.cur_edge;
    end
    else if (scn^.last_edge=0) then
        scn^.last_edge := scn^.cur_edge;

    scn^.width := scn^.cur_edge - scn^.last_edge;
    scn^.last_edge := scn^.cur_edge;
    //DebugLn('  process_edge: sgn=%d cur=%d.%d w=%d (SPACE/BAR:%d)',[scn^.y1_sign, scn^.cur_edge shr ZBAR_FIXED, scn^.cur_edge and ((1 shl ZBAR_FIXED) - 1), scn^.width, y1]);

    (* pass to decoder *)
    if Assigned(scn^.decoder) then
      Result := zbar_decode_width(scn^.decoder, scn^.width)
    else
      Result := ZBAR_PARTIAL;
end;

function zbar_scanner_flush(scn: pzbar_scanner_t): zbar_symbol_type_t;
var
  x: unsigned;
begin
    if (scn^.y1_sign = 0) then
        Exit(ZBAR_NONE);

    x := (scn^.x shl ZBAR_FIXED) + {ROUND}(1 shl (ZBAR_FIXED - 1));

    if (scn^.cur_edge <> x) or (scn^.y1_sign > 0) then begin
        Result := process_edge(scn, -scn^.y1_sign);
        scn^.cur_edge := x;
        scn^.y1_sign := -scn^.y1_sign;
        Exit;
    end;

    scn^.y1_sign := 0;
    scn^.width := 0;
    if Assigned(scn^.decoder) then
       Result := zbar_decode_width(scn^.decoder, 0)
    else
       Result := ZBAR_PARTIAL;
end;

function zbar_scanner_new_scan(scn: pzbar_scanner_t): zbar_symbol_type_t; //ok
var
  tmp: zbar_symbol_type_t;
begin
    Result := ZBAR_NONE;
    while (scn^.y1_sign <> 0) do begin
        tmp := zbar_scanner_flush(scn);
        if (tmp < ZBAR_NONE) or (tmp > Result) then
            Result := tmp;
    end;

    (* reset scanner and associated decoder *)
    zbar_scanner_memset(scn^);
    scn^.y1_thresh := scn^.y1_min_thresh;
    if Assigned(scn^.decoder) then
        zbar_decoder_new_scan(scn^.decoder);
end;

function calc_thresh(scn: pzbar_scanner_t): unsigned; //ok
var
    dx, thresh, t: unsigned;
begin
    (* threshold 1st to improve noise rejection *)
    thresh := scn^.y1_thresh;
    if (thresh <= scn^.y1_min_thresh) or (scn^.width=0) then begin
        Exit(scn^.y1_min_thresh);
    end;
    (* slowly return threshold to min *)
    dx := (scn^.x shl ZBAR_FIXED) - scn^.last_edge;
    t := thresh * dx;
    t := t div scn^.width;
    t := t div ZBAR_SCANNER_THRESH_FADE;
    if (thresh > t) then begin
        thresh := thresh - t;
        if (thresh > scn^.y1_min_thresh) then
            Exit(thresh);
    end;
    scn^.y1_thresh := scn^.y1_min_thresh;
    Exit(scn^.y1_min_thresh);
end;

// detekcia hran (parameter y: sedotonova intenzita bodu)
// zbar_scan_y -> process_edge -> zbar_decode_width -> dcode^.handler
function zbar_scan_y(scn: pzbar_scanner_t; y: integer): zbar_symbol_type_t;
var
  x, y0_0, y0_1, y0_2, y0_3, y1_1, y1_2, y2_1, y2_2, d: integer;
  y1_rev: boolean;
begin
    (* FIXME calc and clip to max y range... *)
    (* retrieve short value history *)
    x := scn^.x;
    y0_1 := scn^.y0[(x - 1) and 3];
    y0_0 := y0_1;

    if (x<>0) then begin
        (* update weighted moving average *)
        // y0_0 += ((int)((y - y0_1) * EWMA_WEIGHT)) >> ZBAR_FIXED;
        // unlike the >> operator in the C language, the shr operator is a logical (not arithmetic) bit shift, even if the left operand is a signed integer.
        //DebugLn(' zbar_scan_y: y=%d; x=%d; y0_0=%d; y0_1=%d (EWMA=%d)', [y,x,y0_0,y0_1,EWMA_WEIGHT]);
        Inc(y0_0, (y - y0_1) * EWMA_WEIGHT div (1 shl ZBAR_FIXED));
        scn^.y0[x and 3] := y0_0;
    end
    else begin
        y0_0 := y;
        y0_1 := y;
        scn^.y0[0] := y;
        scn^.y0[1] := y;
        scn^.y0[2] := y;
        scn^.y0[3] := y;
    end;
    y0_2 := scn^.y0[(x - 2) and 3];
    y0_3 := scn^.y0[(x - 3) and 3];

    (* 1st differential @ x-1 *)
    y1_1 := y0_1 - y0_2;
    y1_2 := y0_2 - y0_3;
    if (abs(y1_1) < abs(y1_2)) and
       ((y1_1 >= 0) = (y1_2 >= 0)) then y1_1 := y1_2;

    (* 2nd differentials @ x-1 & x-2 *)
    y2_1 := y0_0 - (y0_1 * 2) + y0_2;
    y2_2 := y0_1 - (y0_2 * 2) + y0_3;

    Result := ZBAR_NONE;
    (* 2nd zero-crossing is 1st local min/max - could be edge *)
    if ((y2_1=0) or ((y2_1>0) and (y2_2 < 0) or (y2_1<=0) and (y2_2>0))) and
       (calc_thresh(scn) <= abs(y1_1)) then begin
        (* check for 1st sign change *)
        if (scn^.y1_sign > 0) then
          y1_rev := y1_1 < 0
        else
          y1_rev := y1_1 > 0;

        if (y1_rev) then
            (* intensity change reversal - finalize previous edge *)
            Result := process_edge(scn, y1_1);

        if (y1_rev or (abs(scn^.y1_sign) < abs(y1_1))) then begin
            scn^.y1_sign := y1_1;

            (* adaptive thresholding *)
            (* start at multiple of new min/max *)
            scn^.y1_thresh := (abs(y1_1) * THRESH_INIT + {ROUND}(1 shl (ZBAR_FIXED - 1))) shr ZBAR_FIXED;
            //dbprintf(1, "\tthr=%d", scn^.y1_thresh);
            if (scn^.y1_thresh < scn^.y1_min_thresh) then
                scn^.y1_thresh := scn^.y1_min_thresh;

            (* update current edge *)
            d := y2_1 - y2_2;
            scn^.cur_edge := 1 shl ZBAR_FIXED;
            if(d=0) then
                scn^.cur_edge := scn^.cur_edge shr 1
            else if(y2_1<>0) then
                (* interpolate zero crossing *)
                Dec(scn^.cur_edge, ((y2_1 shl ZBAR_FIXED) + 1) div d);
            Inc(scn^.cur_edge, x shl ZBAR_FIXED);
            //dbprintf(1, "\n");
        end;
    end
    else
        ; //dbprintf(1, "\n");

    (* FIXME add fall-thru pass to decoder after heuristic "idle" period
       (eg, 6-8 * last width) *)
    scn^.x := x + 1;
end;

(*****************
 * img_scanner.c *
 *****************)
function zbar_image_scanner_set_config(iscn: pzbar_image_scanner_t; sym: zbar_symbol_type_t; cfg: zbar_config_t; val: integer): integer;
var
  c, i: integer;
begin
  if ((sym = ZBAR_NONE) or (sym = ZBAR_COMPOSITE)) and (cfg = ZBAR_CFG_ENABLE) then begin
     iscn^.ean_config := not(not val);
     if sym <> ZBAR_NONE then Exit(0);
  end;

  //if (cfg < ZBAR_CFG_UNCERTAINTY) then
  //   return(zbar_decoder_set_config(iscn^.dcode, sym, cfg, val));

  if (cfg < ZBAR_CFG_POSITION) then begin
     if(cfg > ZBAR_CFG_UNCERTAINTY) then Exit(1);

     c := Ord(cfg) - Ord(ZBAR_CFG_UNCERTAINTY);
     Assert(c=0);
     if (sym > ZBAR_PARTIAL) then begin
         //i := _zbar_get_symbol_hash(sym);
         //iscn^.sym_configs[c][i] = val;
     end
     else begin
       for i:=0 to NUM_SYMS-1 do
         iscn^.sym_configs[c][i] := val;
     end;
     Exit(0);
  end;

  if(sym > ZBAR_PARTIAL) then Exit(1);


  if (cfg >= ZBAR_CFG_X_DENSITY) and (cfg <= ZBAR_CFG_Y_DENSITY) then begin
    //#define CFG(iscn, cfg) ((iscn)^.configs[(cfg) - ZBAR_CFG_X_DENSITY])
    iscn^.configs[Ord(cfg) - Ord(ZBAR_CFG_X_DENSITY)] := val;
    Exit(0);
  end;

  if (cfg > ZBAR_CFG_POSITION) then
    Exit(1);
  Dec(cfg, Ord(ZBAR_CFG_POSITION));

  if (val=0) then
     iscn^.config := iscn^.config and (not (1 shl Ord(cfg)))
  else if (val = 1) then
     iscn^.config := iscn^.config or (1 shl Ord(cfg))
  else
     Exit(1);

  Exit(0);
end;

procedure zbar_image_scanner_destroy(iscn: pzbar_image_scanner_t);
var
  i: integer;
  sym, next: pzbar_symbol_t;
begin
  //dump_stats(iscn);
  if Assigned(iscn^.syms) then begin
     if iscn^.syms^.refcnt > 0 then
       zbar_symbol_set_ref(iscn^.syms, -1)
     else
       _zbar_symbol_set_free(iscn^.syms);
     iscn^.syms := nil;
  end;

  if Assigned(iscn^.scn) then
    zbar_scanner_destroy(iscn^.scn);
  iscn^.scn := nil;

  if Assigned(iscn^.dcode) then
    zbar_decoder_destroy(iscn^.dcode);
  iscn^.dcode := nil;

  for i:=0 to RECYCLE_BUCKETS-1 do begin
    sym := iscn^.recycle[i].head;
    while sym<>nil do begin
      next := sym^.next;
      _zbar_symbol_free(sym);
      sym := next
    end;
  end;

  if Assigned(iscn^.qr) then begin
     _zbar_qr_destroy(iscn^.qr);
     iscn^.qr := nil;
  end;

  FreeMem(iscn);
end;

(* at this point lengths are all decode unit offsets from the decode edge
 * NB owned by finder *)
function _zbar_decoder_get_qr_finder_line(dcode: pzbar_decoder_t): pqr_finder_line;
begin
    Result := @dcode^.qrf.line;
end;

function _zbar_qr_found_line(reader: pqr_reader; dir: integer; line: pqr_finder_line): integer;
var
  lines: pqr_finder_lines;
begin
    (* minimally intrusive brute force version *)
    lines := @reader^.finder_lines[dir];

    if (lines^.nlines >= lines^.clines) then begin
        lines^.clines := lines^.clines * 2;
        Inc(lines^.clines);
        lines^.lines := ReallocMem(lines^.lines, {++}lines^.clines * sizeof(lines^.lines^));
    end;

    Move(line^, lines^.lines[lines^.nlines{++}], sizeof(line^));
    Inc(lines^.nlines);

    Result := 0;
end;

procedure qr_handler(iscn: pzbar_image_scanner_t);
var
  u: unsigned;
  tmp, vert: integer;
  line: pqr_finder_line;
begin
    line := _zbar_decoder_get_qr_finder_line(iscn^.dcode);
    Assert(line<>nil);
    u := zbar_scanner_get_edge(iscn^.scn, line^.pos[0], QR_FINDER_SUBPREC);
    line^.boffs := u - zbar_scanner_get_edge(iscn^.scn, line^.boffs, QR_FINDER_SUBPREC);
    line^.len := zbar_scanner_get_edge(iscn^.scn, line^.len, QR_FINDER_SUBPREC);
    line^.eoffs := zbar_scanner_get_edge(iscn^.scn, line^.eoffs, QR_FINDER_SUBPREC) - line^.len;
    line^.len := line^.len - u;

    // # define QR_FIXED(v, rnd) ((((v) << 1) + (rnd)) << (QR_FINDER_SUBPREC - 1))
    u := (((iscn^.umin{v} shl 1) + 0{rnd}) shl (QR_FINDER_SUBPREC - 1)) + iscn^.du * u;
    if(iscn^.du < 0) then begin
        tmp := line^.boffs;
        line^.boffs := line^.eoffs;
        line^.eoffs := tmp;
        Dec(u, line^.len);
    end;
    vert := (iscn^.dx and 1) xor 1; // vert=!iscn^.dx
    line^.pos[vert] := u;
    line^.pos[vert xor 1] := (((iscn^.v shl 1){v} + 1{rnd}) shl (QR_FINDER_SUBPREC - 1)); //QR_FIXED(iscn^.v, 1);

    _zbar_qr_found_line(iscn^.qr, Ord(vert), line);
end;

// symbol_handler -> qr_handler -> _zbar_qr_found_line
procedure symbol_handler(dcode: pzbar_decoder_t);
var
  iscn: pzbar_image_scanner_t;
  _type: zbar_symbol_type_t;
  x,y,dir: integer;
  data: PByte;
begin
    iscn := zbar_decoder_get_userdata(dcode);
    _type := zbar_decoder_get_type(dcode);

    if (_type = ZBAR_QRCODE_) then begin
        qr_handler(iscn);
        Exit;
    end
    else
        Assert(False);
(*
    if(TEST_CFG(iscn, ZBAR_CFG_POSITION)) {
        /* tmp position fixup */
        int w = zbar_scanner_get_width(iscn->scn);
        int u = iscn->umin + iscn->du * zbar_scanner_get_edge(iscn->scn, w, 0);
        if(iscn->dx) {
            x = u;
            y = iscn->v;
        }
        else {
            x = iscn->v;
            y = u;
        }
    }

    /* FIXME debug flag to save/display all PARTIALs */
    if(type <= ZBAR_PARTIAL) {
        zprintf(256, "partial symbol @(%d,%d)\n", x, y);
        return;
    }

    data = zbar_decoder_get_data(dcode);
    datalen = zbar_decoder_get_data_length(dcode);

    /* FIXME need better symbol matching */
    for(sym = iscn->syms->head; sym; sym = sym->next)
        if(sym->type == type &&
           sym->datalen == datalen &&
           !memcmp(sym->data, data, datalen)) {
            sym->quality++;
            zprintf(224, "dup symbol @(%d,%d): dup %s: %.20s\n",
                    x, y, zbar_get_symbol_name(type), data);
            if(TEST_CFG(iscn, ZBAR_CFG_POSITION))
                /* add new point to existing set */
                /* FIXME should be polygon */
                sym_add_point(sym, x, y);
            return;
        }

    sym = _zbar_image_scanner_alloc_sym(iscn, type, datalen + 1);
    sym->configs = zbar_decoder_get_configs(dcode, type);
    sym->modifiers = zbar_decoder_get_modifiers(dcode);
    /* FIXME grab decoder buffer */
    memcpy(sym->data, data, datalen + 1);

    /* initialize first point */
    if(TEST_CFG(iscn, ZBAR_CFG_POSITION)) {
        zprintf(192, "new symbol @(%d,%d): %s: %.20s\n",
                x, y, zbar_get_symbol_name(type), data);
        sym_add_point(sym, x, y);
    }

    dir = zbar_decoder_get_direction(dcode);
    if(dir)
        sym->orient = (iscn->dy != 0) + ((iscn->du ^ dir) & 2);

    _zbar_image_scanner_add_sym(iscn, sym);
*)
end;

function zbar_image_scanner_create(): pzbar_image_scanner_t;
begin
  Result := AllocMem(sizeof(zbar_image_scanner_t));
  if Result = nil then Exit;

  Result^.dcode := zbar_decoder_create();
  Result^.scn := zbar_scanner_create(Result^.dcode);
  if (Result^.dcode=nil) or (Result^.scn=nil) then begin
    zbar_image_scanner_destroy(Result);
    Exit(nil);
  end;
  zbar_decoder_set_userdata(Result^.dcode, Result);
  zbar_decoder_set_handler(Result^.dcode, @symbol_handler);

  Result^.qr := _zbar_qr_create();

  (* apply default configuration *)
  //#define CFG(iscn, cfg) ((iscn)^.configs[(cfg) - ZBAR_CFG_X_DENSITY])
  Result^.configs[Ord(ZBAR_CFG_X_DENSITY) - Ord(ZBAR_CFG_X_DENSITY)] := 1;
  Result^.configs[Ord(ZBAR_CFG_Y_DENSITY) - Ord(ZBAR_CFG_X_DENSITY)] := 1;
  zbar_image_scanner_set_config(Result, ZBAR_NONE, ZBAR_CFG_POSITION, 1);
  zbar_image_scanner_set_config(Result, ZBAR_NONE, ZBAR_CFG_UNCERTAINTY, 2);
  zbar_image_scanner_set_config(Result, ZBAR_QRCODE_, ZBAR_CFG_UNCERTAINTY, 0);
  zbar_image_scanner_set_config(Result, ZBAR_CODE128, ZBAR_CFG_UNCERTAINTY, 0);
  zbar_image_scanner_set_config(Result, ZBAR_CODE93, ZBAR_CFG_UNCERTAINTY, 0);
  zbar_image_scanner_set_config(Result, ZBAR_CODE39, ZBAR_CFG_UNCERTAINTY, 0);
  zbar_image_scanner_set_config(Result, ZBAR_CODABAR, ZBAR_CFG_UNCERTAINTY, 1);
  zbar_image_scanner_set_config(Result, ZBAR_COMPOSITE, ZBAR_CFG_UNCERTAINTY, 0);
end;

function _zbar_image_scanner_alloc_sym(
  iscn: pzbar_image_scanner_t;
  &type: zbar_symbol_type_t;
  datalen: integer): pzbar_symbol_t;
var
  i: integer;
begin
  (* recycle old or alloc new symbol *)
  Result := nil;
  for i:=0 to RECYCLE_BUCKETS - 2 do
    if (datalen <= 1 shl (i * 2)) then break;

  while i > 0 do begin
    if (Result = iscn^.recycle[i].head) then begin
      //# define STAT(x) iscn^.stat_##x++
      Inc(iscn^.stat_sym_recycle[i]);
      break;
    end;
    Dec(i);
  end;

  if Assigned(Result) then begin
    iscn^.recycle[i].head := Result^.next;
    Result^.next := nil;
    Assert(iscn^.recycle[i].nsyms<>0);
    Dec(iscn^.recycle[i].nsyms);
  end
  else begin
    Result := AllocMem(sizeof(zbar_symbol_t));
    Inc(iscn^.stat_sym_new);
  end;

  (* init new symbol *)
  Result^.&type := &type;
  Result^.quality := 1;
  Result^.npts := 0;
  Result^.orient := ZBAR_ORIENT_UNKNOWN;
  Result^.cache_count := 0;
  Result^.time := iscn^.time;
  assert(Result^.syms=nil);

  if (datalen > 0) then begin
    Result^.datalen := datalen - 1;
    if (Result^.data_alloc < datalen) then begin
      if Assigned(Result^.data) then FreeMem(Result^.data);
      Result^.data_alloc := datalen;
      Result^.data := GetMem(datalen);
    end;
  end
  else begin
    if Assigned(Result^.data) then FreeMem(Result^.data);
    Result^.data := nil;
    Result^.datalen := 0;
    Result^.data_alloc := 0;
  end;
end;

procedure _zbar_image_scanner_recycle_syms(iscn: pzbar_image_scanner_t; sym: pzbar_symbol_t);
var
  next: pzbar_symbol_t;
  i: integer;
  bucket: precycle_bucket_t;
begin
    next := nil;
    while sym<>nil do begin
        next := sym^.next;
        if (sym^.refcnt<>0) and (_zbar_refcnt(sym^.refcnt, -1)<>0) then begin
            (* unlink referenced symbol *)
            (* FIXME handle outstanding component refs (currently unsupported) *)
            assert(sym^.data_alloc<>0);
            sym^.next := nil;
        end
        else begin
            (* recycle unreferenced symbol *)
            if(sym^.data_alloc=0) then begin
                sym^.data := nil;
                sym^.datalen := 0;
            end;
            if Assigned(sym^.syms) then begin
                if(_zbar_refcnt(sym^.syms^.refcnt, -1)<>0) then
                  assert(False);
                _zbar_image_scanner_recycle_syms(iscn, sym^.syms^.head);
                sym^.syms^.head := nil;
                _zbar_symbol_set_free(sym^.syms);
                sym^.syms := nil;
            end;
            i:=0;
            while i < RECYCLE_BUCKETS do begin
                if (sym^.data_alloc < (1 shl (i * 2))) then
                    break;
                Inc(i);
            end;
            if (i = RECYCLE_BUCKETS) then begin
                assert(sym^.data<>nil);
                FreeMem(sym^.data);
                sym^.data := nil;
                sym^.data_alloc := 0;
                i := 0;
            end;
            bucket := @iscn^.recycle[i];
            (* FIXME cap bucket fill *)
            Inc(bucket^.nsyms);
            sym^.next := bucket^.head;
            bucket^.head := sym;
        end;
        sym := next
    end;
end;

procedure _zbar_image_scanner_add_sym(iscn: pzbar_image_scanner_t; sym: pzbar_symbol_t);
var
  syms: pzbar_symbol_set_t;
begin
  //cache_sym(iscn, sym);

  syms := iscn^.syms;
  if (sym^.cache_count<>0) or (syms^.tail=nil) then begin
      sym^.next := syms^.head;
      syms^.head := sym;
  end
  else begin
      sym^.next := syms^.tail^.next;
      syms^.tail^.next := sym;
  end;

  if (sym^.cache_count=0) then
      Inc(syms^.nsyms)
  else if Assigned(syms^.tail) then
      syms^.tail := sym;

  _zbar_symbol_refcnt(sym, 1);
end;

function recycle_syms(iscn: pzbar_image_scanner_t; syms: pzbar_symbol_set_t): integer;
begin
    if _zbar_refcnt(syms^.refcnt, -1) <> 0 then
        Exit(1);

    _zbar_image_scanner_recycle_syms(iscn, syms^.head);
    syms^.head := nil;
    syms^.tail := nil;
    syms^.nsyms := 0;
    Exit(0);
end;

procedure zbar_image_scanner_recycle_image(iscn: pzbar_image_scanner_t; img: pzbar_image_t);
var
  syms: pzbar_symbol_set_t;
begin
    syms := iscn^.syms;
    if Assigned(syms) and (syms^.refcnt<>0) then begin
        if (recycle_syms(iscn, syms)<>0) then begin
            //STAT(iscn_syms_inuse);
            Inc(iscn^.stat_iscn_syms_inuse);
            iscn^.syms := nil;
        end
        else
            //STAT(iscn_syms_recycle);
            Inc(iscn^.stat_iscn_syms_recycle);
    end;

    syms := img^.syms;
    img^.syms := nil;
    if Assigned(syms) and (recycle_syms(iscn, syms)<>0) then
        //STAT(img_syms_inuse);
        Inc(iscn^.stat_img_syms_inuse)
    else if Assigned(syms) then begin
        //STAT(img_syms_recycle);
        Inc(iscn^.stat_iscn_syms_recycle);

        (* select one set to resurrect, destroy the other *)
        if Assigned(iscn^.syms) then
            _zbar_symbol_set_free(syms)
        else
            iscn^.syms := syms;
    end;
end;

procedure zbar_image_scanner_enable_cache(iscn: pzbar_image_scanner_t; enable: integer);
begin
    if Assigned(iscn^.cache) then begin
        (* recycle all cached syms *)
        _zbar_image_scanner_recycle_syms(iscn, iscn^.cache);
        iscn^.cache := nil;
    end;
    if enable <> 0 then
      iscn^.enable_cache := 1
    else
      iscn^.enable_cache := 0;
end;

procedure quiet_border(iscn: pzbar_image_scanner_t); //ok
begin
    (* flush scanner pipeline *)
    zbar_scanner_flush(iscn^.scn);
    zbar_scanner_flush(iscn^.scn);
    zbar_scanner_new_scan(iscn^.scn);
end;

(** scan for symbols in provided image.  The image format must be
 * "Y800" or "GRAY".
 * @returns >0 if symbols were successfully decoded from the image,
 * 0 if no symbols were found or -1 if an error occurs
 * @see zbar_image_convert()
 * @since 0.9 - changed to only accept grayscale images
 *)
// prechadza obraz po riadkoch a nasledno po stlpcoch
function zbar_scan_image(iscn: pzbar_image_scanner_t; img: pzbar_image_t): integer;
var
  syms: pzbar_symbol_set_t;
  scn: pzbar_scanner_t;
  data, p: PByte;
  w,h, cx1, cy1: unsigned;
  density, border, cx0,cy0, x,y: integer;
  d: Byte;
begin
    scn := iscn^.scn;
    //iscn^.time := _zbar_timer_now();

    (* image must be in grayscale format *)
    _zbar_qr_reset(iscn^.qr);
    iscn^.img := img;

    (* recycle previous scanner and image results *)
    zbar_image_scanner_recycle_image(iscn, img);
    syms := iscn^.syms;
    if (syms=nil) then begin
        syms := _zbar_symbol_set_create();
        iscn^.syms := syms;
        Inc(iscn^.stat_syms_new);
        zbar_symbol_set_ref(syms, 1);
    end
    else
        zbar_symbol_set_ref(syms, 2);
    img^.syms := syms;

    w := img^.width;
    h := img^.height;
    cx1 := img^.crop_x + img^.crop_w;
    assert(cx1 <= w);
    cy1 := img^.crop_y + img^.crop_h;
    assert(cy1 <= h);
    data := img^.data;

    zbar_scanner_new_scan(scn);

    // #define CFG(iscn, cfg) ((iscn)^.configs[(cfg) - ZBAR_CFG_X_DENSITY])
    density := iscn^.configs[Ord(ZBAR_CFG_Y_DENSITY) - Ord(ZBAR_CFG_X_DENSITY)];
    if density>0 then begin
      p:=data;
      x:=0;
      y:=0;
      border :=(((img^.crop_h-1) mod density)+1) div 2;
      if border > img^.crop_h div 2 then
        border := img^.crop_h div 2;
      Inc(border, img^.crop_y);
      assert(border<=h);
      iscn^.dy := 0;

      //movedelta(img^.crop_x, border);
      Inc(x, img^.crop_x{dx});
      Inc(y, border{dy});
      Inc(p, img^.crop_x{dx} + border{dy}*w);

      iscn^.v := y;

      // prechadza obraz po riadkoch
      while y<cy1 do begin
        cx0:=img^.crop_x;
        iscn^.dx := 1;
        iscn^.du := 1;
        iscn^.umin := cx0;
        // z lava do prava
        while x<cx1 do begin
          d:={*}p^;
          //DebugLn('img_x+: %d,%d (%d,%d) point=%d; density=%d', [x, y, cx1,cy1, d, density]);
          //movedelta(1,0);
          Inc(x, 1{dx});
          Inc(y, 0{dy});
          Inc(p, 1{dx} + 0{dy}*w);
          zbar_scan_y(scn,d);
        end;
        // # define ASSERT_POS assert(p == data + x + y * (intptr_t)w)
        assert(p = data + x + y * w);
        quiet_border(iscn);
        //svg_path_end();

        //movedelta(-1,density);
        Inc(x, -1{dx});
        Inc(y, density{dy});
        Inc(p, -1{dx} + density{dy}*w);
        iscn^.v := y;
        if y>=cy1 then break;

        //zprintf(128,'img_x-: %04d,%04d @%p'#13#10'',x,y,p);
        //svg_path_start('vedge',-1. div 32,w,y+0.5);
        iscn^.dx := -1;
        iscn^.du := -1;
        iscn^.umin := cx1;
        // z prava do lava
        while x>=cx0 do begin
          d:={*}p^;
          //DebugLn('img_x+: %d,%d (%d,%d) point=%d; density=%d', [x, y, cx1,cy1, d, density]);
          //movedelta(-1,0);
          Inc(x, -1{dx});
          Inc(y, 0{dy});
          Inc(p, -1{dx} + 0{dy}*w);
          zbar_scan_y(scn,d);
        end;
        //ASSERT_POS;
        assert(p = data + x + y * w);
        quiet_border(iscn);
        //svg_path_end();

        //movedelta(1,density);
        Inc(x, 1{dx});
        Inc(y, density{dy});
        Inc(p, 1{dx} + density{dy}*w);
        iscn^.v := y;
      end;
      //svg_group_end();
    end;
    iscn^.dx := 0;

    //density := CFG(iscn,ZBAR_CFG_X_DENSITY);
    density := iscn^.configs[Ord(ZBAR_CFG_X_DENSITY) - Ord(ZBAR_CFG_X_DENSITY)];
    if density>0 then begin
      p:=data;
      x:=0;
      y:=0;
      border :=(((img^.crop_w-1) mod density)+1) div 2;
      if border>img^.crop_w div 2 then
        border:= img^.crop_w div 2;
      Inc(border, img^.crop_x);
      assert(border<=w);
      //svg_group_start('scanner',90,1,-1,0,0);
      //movedelta(border,img^.crop_y);
      Inc(x, border{dx});
      Inc(y, img^.crop_y{dy});
      Inc(p, border{dx} + img^.crop_y{dy}*w);
      iscn^.v := x;

      // prechadza obraz po stlpcoch
      while x<cx1 do begin
        //DebugLn('img_y+: %d,%d (%d,%d); density=%d', [x, y, cx1,cy1, density]);
        cy0:=img^.crop_y;
        //zprintf(128,'img_y+: %04d,%04d @%p'#13#10'',x,y,p);
        //svg_path_start('vedge',1. div 32,0,x+0.5);
        iscn^.dy := 1;
        iscn^.du := 1;
        iscn^.umin := cy0;
        // z hora dole
        while y<cy1 do begin
          d:={*}p^;
          //movedelta(0,1);
          Inc(x, 0{dx});
          Inc(y, 1{dy});
          Inc(p, 0{dx} + 1{dy}*w);
          zbar_scan_y(scn,d);
        end;
        //ASSERT_POS;
        assert(p = data + x + y * w);
        quiet_border(iscn);
        //svg_path_end();

        //movedelta(density,-1);
        Inc(x, density{dx});
        Inc(y, -1{dy});
        Inc(p, density{dx} + -1{dy}*w);
        iscn^.v := x;
        if x>=cx1 then break;

        //zprintf(128,'img_y-: %04d,%04d @%p'#13#10'',x,y,p);
        //svg_path_start('vedge',-1. div 32,h,x+0.5);
        iscn^.dy := -1;
        iscn^.du := -1;
        iscn^.umin := cy1;
        // z dola hore
        while y>=cy0 do begin
          d:={*}p^;
          //movedelta(0,-1);
          Inc(x, 0{dx});
          Inc(y, -1{dy});
          Inc(p, 0{dx} + -1{dy}*w);
          zbar_scan_y(scn,d);
        end;
        //ASSERT_POS;
        assert(p = data + x + y * w);
        quiet_border(iscn);
        //svg_path_end();

        //movedelta(density,1);
        Inc(x, density{dx});
        Inc(y, 1{dy});
        Inc(p, density{dx} + 1{dy}*w);
        iscn^.v := x;
      end;
      //svg_group_end();
    end;
    iscn^.dy := 0;
    iscn^.img := nil;

    _zbar_qr_decode(iscn^.qr, iscn, img);

    if (syms^.nsyms<>0) and Assigned(iscn^.handler) then
        iscn^.handler(img, iscn^.userdata);

    Result := syms^.nsyms;
end;

function zbar_image_scanner_get_results(iscn: pzbar_image_scanner_t): pzbar_symbol_set_t; inline;
begin
    Result := iscn^.syms;
end;

(***************
 * processor.c *
***************)
function _zbar_processor_init(proc: pzbar_processor_t): integer;
begin
    Result := 0;
end;

function zbar_processor_create(threaded: integer): pzbar_processor_t;
begin
  Result := AllocMem(sizeof(zbar_processor_t));
  if Result = nil then Exit;

  //err_init(&proc^.err, ZBAR_MOD_PROCESSOR);
  Result^.err.magic := $5252457a; //ERRINFO_MAGIC;
  Result^.err.module := ZBAR_MOD_PROCESSOR;

  Result^.scanner := zbar_image_scanner_create();
  if Result^.scanner = nil then begin
    FreeMem(Result);
    Result := nil;
  end
  else begin
    Result^.threaded := {!_zbar_mutex_init(&proc^.mutex) and} threaded;
    _zbar_processor_init(Result);
  end;
end;

function zbar_processor_init(processor: pzbar_processor_t; video_device: PAnsiChar; enable_display: Integer): Integer;
begin
  Result := 0;
end;

procedure zbar_processor_destroy(proc: pzbar_processor_t);
begin
    zbar_processor_init(proc, nil, 0);

    if Assigned(proc^.syms) then begin
        zbar_symbol_set_ref(proc^.syms, -1);
        proc^.syms := nil;
    end;

    if Assigned(proc^.scanner) then begin
        zbar_image_scanner_destroy(proc^.scanner);
        proc^.scanner := nil;
    end;

    //_zbar_mutex_destroy(&proc^.mutex);
    //_zbar_processor_cleanup(proc); // no-op in Windows

    err_cleanup(proc^.err);
    FreeMem(proc);
end;

function zbar_processor_set_visible(processor: pzbar_processor_t; visible: Integer): Integer;
begin
  Result := 0;
end;

(* cleanup linked image by unrefing *)
procedure cleanup_ref(img: pzbar_image_t);
begin
    if Assigned(img^.next) then
        _zbar_image_refcnt(img^.next^, -1);
end;

function _zbar_process_image(proc: pzbar_processor_t; img: pzbar_image_t): integer;
  (* make new image w/reference to the same image data *)
  procedure convert_copy(dst: pzbar_image_t; src: pzbar_image_t);
  begin
    if (src^.width = dst^.width) and (src^.height = dst^.height) then begin
        dst^.data := src^.data;
        dst^.datalen := src^.datalen;
        dst^.cleanup := @cleanup_ref;
        dst^.next := src;
        _zbar_image_refcnt(src^, 1);
    end;
  end;

  function zbar_image_convert(src: pzbar_image_t; fmt: unsigned): pzbar_image_t;
  begin
    Result := zbar_image_create();
    Result^.format := fmt;
    Result^.width := src^.width;
    Result^.height := src^.height;
    zbar_image_set_crop(Result, src^.crop_x, src^.crop_y, src^.crop_w, src^.crop_h);
    // netestujeme, ale predpokladame, ze src^.format = fmt
    convert_copy(Result, src); ;
  end;

var
  force_fmt, format: uint32_t;
  tmp: pzbar_image_t;
  nsyms: integer;
  syms: pzbar_symbol_set_t;
begin
    force_fmt := proc^.force_output;
    if Assigned(img) then begin
        format := zbar_image_get_format(img);
        (* FIXME locking all other interfaces while processing is conservative
         * but easier for now and we don't expect this to take long...
         *)
        tmp := zbar_image_convert(img, 0{fourcc('Y','8','0','0')});

        if Assigned(proc^.syms) then begin
            zbar_symbol_set_ref(proc^.syms, -1);
            proc^.syms := nil;
        end;
        zbar_image_scanner_recycle_image(proc^.scanner, img);
        nsyms := zbar_scan_image(proc^.scanner, tmp);
        _zbar_image_swap_symbols(img, tmp);

        zbar_image_destroy(tmp);
        tmp := nil;
        //if(nsyms < 0) goto error;
        assert(nsyms>=0);

        proc^.syms := zbar_image_scanner_get_results(proc^.scanner);
        if Assigned(proc^.syms) then
            zbar_symbol_set_ref(proc^.syms, 1);

        if (nsyms<>0) then begin
            (* FIXME only call after filtering *)
            if Assigned(proc^.handler) then
                proc^.handler(img, proc^.userdata);
        end;

        if (force_fmt<>0) then begin
            syms := img^.syms;
            img := zbar_image_convert(img, force_fmt);
            //if (!img) goto error;
            img^.syms := syms;
            zbar_symbol_set_ref(syms, 1);
        end;
    end;

    if (force_fmt<>0) and Assigned(img) then
        zbar_image_destroy(img);

    Result := 0;
end;

function zbar_process_image(proc: pzbar_processor_t; img: pzbar_image_t): integer;
begin
  zbar_image_scanner_enable_cache(proc^.scanner, 0);
  Result := _zbar_process_image(proc, img);
  if (proc^.streaming <> 0) then
    zbar_image_scanner_enable_cache(proc^.scanner, 1);
end;


(**************
 * qrdectxt.c *
 **************)
{ $I zbar_qrdectxt.inc}


{$IF FPC_FULLVERSION > 30202}
// /rtl/inc/sortbase.pp
procedure QSort(base: Pointer; L, R : integer; size: integer; Compare: TListSortComparer_Context);
begin
  QuickSort_ItemList_Context(base, R-L+1, size, Compare, nil);
end;
{$ELSE}
// Quick Sort (/rtl/inc/sortbase.pp)
type
  TListSortComparer_Context = function(Item1, Item2, Context: Pointer): Integer;
procedure QSort(base: Pointer; L, R : integer; size: integer; Compare: TListSortComparer_Context);
var
  Q : Pointer;

  procedure QuickSort(L, R : integer);
  var
    I, J, PivotIdx : integer;
    P : Pointer;
  begin
    repeat
      I := L;
      J := R;
      PivotIdx := L + ((R - L) shr 1); { same as ((L + R) div 2), but without the possibility of overflow }
      P := base + size*PivotIdx;
      repeat
        while (I < PivotIdx) and (Compare(P, base + size*I, nil) > 0) do
          Inc(I);
        while (J > PivotIdx) and (Compare(P, base + size*J, nil) < 0) do
          Dec(J);
        if I < J then
        begin
          Move((base + size*I)^, Q^, size);
          Move((base + size*J)^, (base + size*I)^, size);
          Move(Q^, (base + size*J)^, size);
          if PivotIdx = I then
          begin
            PivotIdx := J;
            P := base + size*PivotIdx;
            Inc(I);
          end
          else if PivotIdx = J then
          begin
            PivotIdx := I;
            P := base + size*PivotIdx;
            Dec(J);
          end
          else
          begin
            Inc(I);
            Dec(J);
          end;
        end;
      until I >= J;
      // sort the smaller range recursively
      // sort the bigger range via the loop
      // Reasons: memory usage is O(log(n)) instead of O(n) and loop is faster than recursion
      if (PivotIdx - L) < (R - PivotIdx) then
      begin
        if (L + 1) < PivotIdx then
          QuickSort(L, PivotIdx - 1);
        L := PivotIdx + 1;
      end
      else
      begin
        if (PivotIdx + 1) < R then
          QuickSort(PivotIdx + 1, R);
        if (L + 1) < PivotIdx then
          R := PivotIdx - 1
        else
          exit;
      end;
    until L >= R;
  end;

begin
  Q := GetMem(size);
  QuickSort(L, R);
  FreeMem(Q);
end;
{$ENDIF}


(***********
 * qrdec.c *
 ***********)
const
  (*TODO: Perhaps these thresholds should be on the module size instead?
   Unfortunately, I'd need real-world images of codes with larger versions to
    see if these thresholds are still effective, but such versions aren't used
    often.*)

  (*The amount that the estimated version numbers are allowed to differ from the
     real version number and still be considered valid.*)
  QR_SMALL_VERSION_SLACK = 1;
  (*Since cell phone cameras can have severe radial distortion, the estimated
     version for larger versions can be off by larger amounts.*)
  QR_LARGE_VERSION_SLACK = 3;

type
  qr_line = array[0..2] of integer;


(* reset finder state between scans *)
procedure _zbar_qr_reset(reader: pqr_reader);
begin
    reader^.finder_lines[0].nlines := 0;
    reader^.finder_lines[1].nlines := 0;
end;

function qr_finder_vline_cmp(_a,_b,context: Pointer): integer;
var
  a,b: pqr_finder_line;
begin
  a:=pqr_finder_line(_a);
  b:=pqr_finder_line(_b);
  Result := ((Ord(a^.pos[0] > b^.pos[0]) - Ord(a^.pos[0] < b^.pos[0])) shl 1) +
              Ord(a^.pos[1] > b^.pos[1]) - Ord(a^.pos[1] < b^.pos[1]);
end;

(*Clusters adjacent lines into groups that are large enough to be crossing a
   finder pattern (relative to their length).
  _clusters:  The buffer in which to store the clusters found.
  _neighbors: The buffer used to store the lists of lines in each cluster.
  _lines:     The list of lines to cluster.
              Horizontal lines must be sorted in ascending order by Y
               coordinate, with ties broken by X coordinate.
              Vertical lines must be sorted in ascending order by X coordinate,
               with ties broken by Y coordinate.
  _nlines:    The number of lines in the set of lines to cluster.
  _v:         0 for horizontal lines, or 1 for vertical lines.
  Return: The number of clusters.*)

function qr_finder_cluster_lines(
  _clusters: pqr_finder_cluster;
  _neighbors: ppqr_finder_line;
  _lines: pqr_finder_line;
  _nlines, _v: integer): integer;
var
  a,b: pqr_finder_line;
  mark: pbyte;
  neighbors: ppqr_finder_line;
  nneighbors: integer;
  i,j, len, thresh: integer;
begin
  mark := AllocMem(_nlines*sizeof(mark^));
  neighbors := _neighbors;
  Result := 0;
  for i:=0 to Pred(_nlines-1) do
    if mark[i]=0 then begin
      nneighbors:= 1;
      neighbors[0] := _lines+i;
      len := _lines[i].len;
      for j:=i+1 to Pred(_nlines) do
        if mark[j]=0 then begin
          a:= neighbors[nneighbors-1];
          b:= _lines+j;
          (*The clustering threshold is proportional to the size of the lines,
            since minor noise in large areas can interrupt patterns more easily
            at high resolutions.*)
          thresh := (a^.len+7) shr 2;
          if abs(a^.pos[1-_v] - b^.pos[1-_v])>thresh then break;
          if abs(a^.pos[_v] - b^.pos[_v])>thresh then continue;
          if abs(a^.pos[_v] + a^.len - b^.pos[_v] - b^.len)>thresh then continue;
          if (a^.boffs>0) and (b^.boffs>0) and (abs(a^.pos[_v]-a^.boffs-b^.pos[_v]+b^.boffs)>thresh) then continue;
          if (a^.eoffs>0) and (b^.eoffs>0) and (abs(a^.pos[_v]+a^.len+a^.eoffs-b^.pos[_v]-b^.len-b^.eoffs)>thresh) then continue;
          neighbors[nneighbors] := _lines+j;
          Inc(nneighbors);
          Inc(len, b^.len);
    end;
    (*We require at least three lines to form a cluster, which eliminates a
      large number of false positives, saving considerable decoding time.
      This should still be sufficient for 1-pixel codes with no noise.*)
    if nneighbors<3 then continue;
    (*The expected number of lines crossing a finder pattern is equal to their average length.
      We accept the cluster if size is at least 1/3 their average length (this
      is a very small threshold, but was needed for some test images).*)
    len := ((len shl 1)+nneighbors) div (nneighbors shl 1);

    if nneighbors*(5 shl QR_FINDER_SUBPREC) >= len then begin
      _clusters[Result].lines := neighbors;
      _clusters[Result].nlines:= nneighbors;
      for j:=0 to Pred(nneighbors) do
        mark[neighbors[j]-_lines] := 1;
      Inc(neighbors, nneighbors);
      Inc(Result);
    end;
  end;
  FreeMem(mark);
end;

(*Adds the coordinates of the edge points from the lines contained in the
   given list of clusters to the list of edge points for a finder center.
  Only the edge point position is initialized.
  The edge label and extent are set by qr_finder_edge_pts_aff_classify()
   or qr_finder_edge_pts_hom_classify().

  _edge_pts:   The buffer in which to store the edge points.
  _nedge_pts:  The current number of edge points in the buffer.
  _neighbors:  The list of lines in the cluster.
  _nneighbors: The number of lines in the list of lines in the cluster.
  _v:          0 for horizontal lines and 1 for vertical lines.
  Return: The new total number of edge points.
*)
function qr_finder_edge_pts_fill(_edge_pts: pqr_finder_edge_pt; _nedge_pts: integer;
 _neighbors: ppqr_finder_cluster; _nneighbors: integer; _v: integer): integer;
var
  i,j: integer;
  c: pqr_finder_cluster;
  l: pqr_finder_line;
begin
  for i:=0 to _nneighbors-1 do begin
    c := _neighbors[i];
    for j:=0 to c^.nlines-1 do begin
      l := c^.lines[j];
      if l^.boffs > 0 then begin
        _edge_pts[_nedge_pts].pos[0] := l^.pos[0];
        _edge_pts[_nedge_pts].pos[1] := l^.pos[1];
        Dec(_edge_pts[_nedge_pts].pos[_v], l^.boffs);
        Inc(_nedge_pts);
      end;
      if l^.eoffs>0 then begin
        _edge_pts[_nedge_pts].pos[0] := l^.pos[0];
        _edge_pts[_nedge_pts].pos[1] := l^.pos[1];
        Inc(_edge_pts[_nedge_pts].pos[_v], l^.len + l^.eoffs);
        Inc(_nedge_pts);
      end;
    end;
  end;
  Result := _nedge_pts;
end;

function qr_finder_center_cmp(_a, _b, context: Pointer): integer;
var
  a, b: pqr_finder_center;
begin
  a:=pqr_finder_center(_a);
  b:=pqr_finder_center(_b);
  Result := ((Ord(b^.nedge_pts > a^.nedge_pts) - Ord(b^.nedge_pts < a^.nedge_pts)) shl 2) +
            ((Ord(a^.pos[1] > b^.pos[1]) - Ord(a^.pos[1] < b^.pos[1])) shl 1) +
              Ord(a^.pos[0] > b^.pos[0]) - Ord(a^.pos[0] < b^.pos[0]);
end;

(*Determine if a horizontal line crosses a vertical line.
  _hline: The horizontal line.
  _vline: The vertical line.
  Return: A non-zero value if the lines cross, or zero if they do not.
*)
function qr_finder_lines_are_crossing(_hline, _vline: pqr_finder_line): boolean;
begin
  Result :=
   (_hline^.pos[0] <= _vline^.pos[0]) and (_vline^.pos[0] < _hline^.pos[0]+_hline^.len) and
   (_vline^.pos[1] <= _hline^.pos[1]) and (_hline^.pos[1] < _vline^.pos[1]+_vline^.len);
end;

(* Finds horizontal clusters that cross corresponding vertical clusters,
   presumably corresponding to a finder center.

  _center:     The buffer in which to store putative finder centers.
  _edge_pts:   The buffer to use for the edge point lists for each finder
                center.
  _hclusters:  The clusters of horizontal lines crossing finder patterns.
  _nhclusters: The number of horizontal line clusters.
  _vclusters:  The clusters of vertical lines crossing finder patterns.
  _nvclusters: The number of vertical line clusters.
  Return: The number of putative finder centers.
*)
function qr_finder_find_crossings(_centers: pqr_finder_center;
  _edge_pts: pqr_finder_edge_pt; _hclusters: pqr_finder_cluster; _nhclusters: integer;
  _vclusters: pqr_finder_cluster; _nvclusters: integer): integer;
var
  hneighbors, vneighbors: ^pqr_finder_cluster;
  hmark, vmark: PByte;
  i,j: integer;
  a,b: pqr_finder_line;
  nvneighbors, nhneighbors, nedge_pts, x, y: integer;
  c: pqr_finder_center;
begin
  hneighbors := GetMem(_nhclusters*sizeof(hneighbors^));
  vneighbors := GetMem(_nvclusters*sizeof(vneighbors^));
  hmark := AllocMem(_nhclusters*sizeof(hmark^));
  vmark := AllocMem(_nvclusters*sizeof(vmark^));
  Result := 0;
  (*TODO: This may need some re-working.
    We should be finding groups of clusters such that _all_ horizontal lines in
     _all_ horizontal clusters in the group cross _all_ vertical lines in _all_
     vertical clusters in the group.
    This is equivalent to finding the maximum bipartite clique in the
     connectivity graph, which requires linear progamming to solve efficiently.
    In principle, that is easy to do, but a realistic implementation without
     floating point is a lot of work (and computationally expensive).
    Right now we are relying on a sufficient border around the finder patterns
     to prevent false positives.*)
  for i:=0 to _nhclusters-1 do if hmark[i]=0 then begin
    a := _hclusters[i].lines[_hclusters[i].nlines shr 1];
    y := 0;
    nvneighbors :=0;
    for j:=0 to _nvclusters-1 do if vmark[j]=0 then begin
      b := _vclusters[j].lines[_vclusters[j].nlines shr 1];
      if qr_finder_lines_are_crossing(a,b) then begin
        vmark[j]:=1;
        y := y + (b^.pos[1] shl 1) + b^.len;
        if (b^.boffs > 0) and (b^.eoffs>0) then Inc(y, b^.eoffs - b^.boffs);
        vneighbors[nvneighbors{++}] := _vclusters+j;
        Inc(nvneighbors);
      end;
    end;
    if nvneighbors>0 then begin
      x := (a^.pos[0] shl 1) + a^.len;
      if (a^.boffs>0) and (a^.eoffs>0) then Inc(x, a^.eoffs - a^.boffs);
      hneighbors[0] := _hclusters+i;
      nhneighbors := 1;
      j := nvneighbors shr 1;
      b := vneighbors[j]^.lines[vneighbors[j]^.nlines shr 1];
      for j:=i+1 to _nhclusters-1 do if hmark[j]=0 then begin
        a := _hclusters[j].lines[_hclusters[j].nlines shr 1];
        if qr_finder_lines_are_crossing(a,b) then begin
          hmark[j] := 1;
          Inc(x, (a^.pos[0] shl 1) + a^.len);
          if (a^.boffs>0) and (a^.eoffs>0) then Inc(x, a^.eoffs - a^.boffs);
          hneighbors[nhneighbors{++}] := _hclusters+j;
          Inc(nhneighbors);
        end;
      end;
      c := _centers+Result; Inc(Result);
      c^.pos[0] := (x+nhneighbors) div (nhneighbors shl 1);
      c^.pos[1] := (y+nvneighbors) div (nvneighbors shl 1);
      c^.edge_pts := _edge_pts;
      nedge_pts := qr_finder_edge_pts_fill(_edge_pts, 0, hneighbors,nhneighbors, 0);
      nedge_pts := qr_finder_edge_pts_fill(_edge_pts, nedge_pts, vneighbors,nvneighbors, 1);
      c^.nedge_pts := nedge_pts;
      Inc(_edge_pts, nedge_pts);
    end;
  end;
  FreeMem(vmark);
  FreeMem(hmark);
  FreeMem(vneighbors);
  FreeMem(hneighbors);
  (*Sort the centers by decreasing numbers of edge points.*)
  qsort(_centers, 0, Result-1, sizeof(_centers^), @qr_finder_center_cmp);
end;

(* Locates a set of putative finder centers in the image.
   First we search for horizontal and vertical lines that have
   (dark:light:dark:light:dark) runs with size ratios of roughly (1:1:3:1:1).
   Then we cluster them into groups such that each subsequent pair of endpoints
    is close to the line before it in the cluster.
   This will locate many line clusters that don't cross a finder pattern, but
    qr_finder_find_crossings() will filter most of them out.
   Where horizontal and vertical clusters cross, a prospective finder center is
    returned.

  _centers:  Returns a pointer to a freshly-allocated list of finder centers.
             This must be freed by the caller.
  _edge_pts: Returns a pointer to a freshly-allocated list of edge points
              around those centers.
             This must be freed by the caller.
  _img:      The binary image to search.
  _width:    The width of the image.
  _height:   The height of the image.
  Return: The number of putative finder centers located.
*)

function qr_finder_centers_locate(
  var _centers: pqr_finder_center;
  var _edge_pts: pqr_finder_edge_pt;
  reader: pqr_reader;
  _width,_height: integer): integer;
var
  hlines: pqr_finder_line;
  vlines: pqr_finder_line;
  nhlines, nvlines: integer;
  hclusters: pqr_finder_cluster;
  nhclusters: integer;
  vclusters: pqr_finder_cluster;
  nvclusters: integer;
  hneighbors: ^pqr_finder_line;
  vneighbors: ^pqr_finder_line;
  edge_pts: pqr_finder_edge_pt;
  centers: pqr_finder_center;
  nedge_pts: integer;
  i: integer;
begin
  hlines:=reader^.finder_lines[0].lines;
  nhlines:=reader^.finder_lines[0].nlines;
  vlines:=reader^.finder_lines[1].lines;
  nvlines:=reader^.finder_lines[1].nlines;

  (*Cluster the detected lines.*)
  hneighbors := GetMem(nhlines*sizeof(hneighbors^));
  (*We require more than one line per cluster, so there are at most nhlines/2.*)
  hclusters := GetMem((nhlines shr 1)*sizeof(hclusters^));
  nhclusters := qr_finder_cluster_lines(hclusters,hneighbors,hlines,nhlines,0);
  (*We need vertical lines to be sorted by X coordinate, with ties broken by Y
    coordinate, for clustering purposes.
    We scan the image in the opposite order for cache efficiency, so sort the
    lines we found here.*)
  qsort(vlines, 0, nvlines-1, sizeof(vlines^), @qr_finder_vline_cmp);

  vneighbors := GetMem(nvlines*sizeof(vneighbors^));
  (*We require more than one line per cluster, so there are at most nvlines/2.*)
  vclusters := GetMem((nvlines shr 1)*sizeof(vclusters^));
  nvclusters:= qr_finder_cluster_lines(vclusters,vneighbors,vlines,nvlines,1);

  (*Find line crossings among the clusters.*)
  if (nhclusters>=3) and (nvclusters>=3) then begin
    nedge_pts := 0;
    for i:=0 to Pred(nhclusters) do
      nedge_pts:= nedge_pts + (hclusters[i].nlines);
    for i:=0 to Pred(nvclusters) do
      nedge_pts:= nedge_pts + (vclusters[i].nlines);
    nedge_pts:= nedge_pts shl 1;
    edge_pts := GetMem(nedge_pts*sizeof(edge_pts^));
    centers := GetMem(QR_MINI(nhclusters,nvclusters)*sizeof(centers^));
    Result := qr_finder_find_crossings(centers,edge_pts,hclusters,nhclusters,vclusters,nvclusters);
    _centers:=centers;
    _edge_pts:=edge_pts;
  end
  else
    Result:= 0;
  FreeMem(vclusters);
  FreeMem(vneighbors);
  FreeMem(hclusters);
  FreeMem(hneighbors);
end;

procedure qr_point_translate(var _point: qr_point; _dx,_dy: integer);
begin
  Inc(_point[0], _dx);
  Inc(_point[1], _dy);
end;

function qr_point_distance2(const _p1,_p2: qr_point): unsigned;
begin
  Result := Sqr(_p1[0]-_p2[0]) + Sqr(_p1[1]-_p2[1]);
end;

(*Returns the cross product of the three points, which is positive if they are
   in CCW order (in a right-handed coordinate system), and 0 if they're
   colinear.*)
function qr_point_ccw(const _p0, _p1, _p2: qr_point): integer;
begin
  Result := (_p1[0]-_p0[0])*(_p2[1]-_p0[1]) - (_p1[1]-_p0[1])*(_p2[0]-_p0[0]);
end;


(*Evaluates a line equation at a point.
  _line: The line to evaluate.
  _x:    The X coordinate of the point.
  _y:    The y coordinate of the point.
  Return: The value of the line equation _line[0]*_x+_line[1]*_y+_line[2].
*)
function qr_line_eval(_line: qr_line; _x,_y: integer): integer;
begin
  Result := _line[0]*_x + _line[1]*_y + _line[2];
end;

(*Computes a line passing through the given point using the specified second
   order statistics.
  Given a line defined by the equation
    A*x+B*y+C = 0 ,
   the least squares fit to n points (x_i,y_i) must satisfy the two equations
    A^2 + (Syy - Sxx)/Sxy*A*B - B^2 = 0 ,
    C = -(xbar*A+ybar*B) ,
   where
    xbar = sum(x_i)/n ,
    ybar = sum(y_i)/n ,
    Sxx = sum((x_i-xbar)**2) ,
    Sxy = sum((x_i-xbar)*(y_i-ybar)) ,
    Syy = sum((y_i-ybar)**2) .
  The quadratic can be solved for the ratio (A/B) or (B/A):
    A/B = (Syy + sqrt((Sxx-Syy)**2 + 4*Sxy**2) - Sxx)/(-2*Sxy) ,
    B/A = (Sxx + sqrt((Sxx-Syy)**2 + 4*Sxy**2) - Syy)/(-2*Sxy) .
  We pick the one that leads to the larger ratio to avoid destructive
   cancellation (and e.g., 0/0 for horizontal or vertical lines).
  The above solutions correspond to the actual minimum.
  The other solution of the quadratic corresponds to a saddle point of the
   least squares objective function.
  _l:   Returns the fitted line values A, B, and C.
  _x0:  The X coordinate of the point the line is supposed to pass through.
  _y0:  The Y coordinate of the point the line is supposed to pass through.
  _sxx: The sum Sxx.
  _sxy: The sum Sxy.
  _syy: The sum Syy.
  _res: The maximum number of bits occupied by the product of any two of
         _l[0] or _l[1].
        Smaller numbers give less angular resolution, but allow more overhead
         room for computations.
*)
procedure qr_line_fit(out _l: qr_line; _x0,_y0, _sxx,_sxy,_syy,_res: integer);
var
  dshift,dround, u,v,w: integer;
begin
  u:=abs(_sxx-_syy);
  v:=(-_sxy) shl 1;
  w:=qr_ihypot(u,v);
  (*Computations in later stages can easily overflow with moderate sizes, so we
     compute a shift factor to scale things down into a managable range.
    We ensure that the product of any two of _l[0] and _l[1] fits within _res
     bits, which allows computation of line intersections without overflow.*)
  dshift:=QR_MAXI(0, QR_MAXI(qr_ilog(u), qr_ilog(abs(v)))+1-((_res+1) shr 1));
  dround:=(1 shl dshift) shr 1;
  if(_sxx>_syy) then begin
    _l[0]:=(v+dround)   div (1 shl dshift);
    _l[1]:=(u+w+dround) div (1 shl dshift);
  end
  else begin
    _l[0]:=(u+w+dround) div (1 shl dshift);
    _l[1]:=(v+dround)   div (1 shl dshift)
  end;
  _l[2]:=-(_x0*_l[0]+_y0*_l[1]);
end;

(*Perform a least-squares line fit to a list of points.
  At least two points are required.*)
procedure qr_line_fit_points(out _l: qr_line; _p: pqr_point; _np,_res: integer);
var
  sx, sy, xmin,xmax,ymin,ymax,xbar,ybar,dx,dy, sxx,sxy,syy, sshift, sround, i: integer;
begin
  sx:=0;
  sy:=0;
  ymax:=Low(integer);
  xmax:=Low(integer);
  ymin:=High(integer);
  xmin:=High(integer);
  for i:=0 to _np-1 do begin
    Inc(sx, _p[i][0]);
    xmin:=QR_MINI(xmin,_p[i][0]);
    xmax:=QR_MAXI(xmax,_p[i][0]);
    Inc(sy, _p[i][1]);
    ymin:=QR_MINI(ymin,_p[i][1]);
    ymax:=QR_MAXI(ymax,_p[i][1]);
  end;
  xbar:=(sx+(_np shr 1)) div _np;
  ybar:=(sy+(_np shr 1)) div _np;
  sshift:=QR_MAXI(0,qr_ilog(_np*QR_MAXI(QR_MAXI(xmax-xbar,xbar-xmin),
          QR_MAXI(ymax-ybar,ybar-ymin)))-((QR_INT_BITS-1) shr 1));
  sround:=(1 shl sshift) shr 1;
  sxx:=0;
  sxy:=0;
  syy:=0;
  for i:=0 to _np-1 do begin
    dx:=(_p[i][0]-xbar+sround) div (1 shl sshift);
    dy:=(_p[i][1]-ybar+sround) div (1 shl sshift);
    Inc(sxx, dx*dx);
    Inc(sxy, dx*dy);
    Inc(syy, dy*dy);
  end;
  qr_line_fit(_l,xbar,ybar,sxx,sxy,syy,_res);
end;

procedure qr_line_orient(var _l: qr_line; _x,_y: integer);
begin
  if(qr_line_eval(_l,_x,_y)<0) then begin
    _l[0]:=-_l[0];
    _l[1]:=-_l[1];
    _l[2]:=-_l[2];
  end;
end;

function qr_line_isect(out _p: qr_point; const _l0, _l1: qr_line): integer;
var
  d,x,y: integer;
begin
  d:=_l0[0]*_l1[1]-_l0[1]*_l1[0];
  if (d=0) then Exit(-1);
  x:=_l0[1]*_l1[2]-_l1[1]*_l0[2];
  y:=_l1[0]*_l0[2]-_l0[0]*_l1[2];
  if(d<0) then begin
    x:=-x;
    y:=-y;
    d:=-d;
  end;
  _p[0]:=QR_DIVROUND(x,d);
  _p[1]:=QR_DIVROUND(y,d);
  Exit(0);
end;


(*An affine homography.
  This maps from the image (at subpel resolution) to a square domain with
   power-of-two sides (of res bits) and back.*)
type
  qr_aff = record
    fwd, inv: array[0..1,0..1] of integer;
    x0,y0,res,ires: integer;
  end;
  pqr_aff = ^qr_aff;

procedure qr_aff_init(out _aff: qr_aff; const _p0,_p1,_p2: qr_point; _res: integer);
var
  det, ires, dx1,dy1,dx2,dy2: integer;
begin
  (*det is ensured to be positive by our caller.*)
  dx1:=_p1[0]-_p0[0];
  dx2:=_p2[0]-_p0[0];
  dy1:=_p1[1]-_p0[1];
  dy2:=_p2[1]-_p0[1];
  det:=dx1*dy2-dy1*dx2;
  ires:=QR_MAXI((qr_ilog(abs(det)) shr 1)-2, 0);
  _aff.fwd[0][0]:=dx1;
  _aff.fwd[0][1]:=dx2;
  _aff.fwd[1][0]:=dy1;
  _aff.fwd[1][1]:=dy2;
  Assert(det>=0);
  _aff.inv[0][0]:=QR_DIVROUND( dy2 shl _res, det shr ires);
  _aff.inv[0][1]:=QR_DIVROUND(-dx2 shl _res, det shr ires);
  _aff.inv[1][0]:=QR_DIVROUND(-dy1 shl _res, det shr ires);
  _aff.inv[1][1]:=QR_DIVROUND( dx1 shl _res, det shr ires);
  _aff.x0:=_p0[0];
  _aff.y0:=_p0[1];
  _aff.res:=_res;
  _aff.ires:=ires;
end;

(*Map from the image (at subpel resolution) into the square domain.*)
procedure qr_aff_unproject(out _q: qr_point; _aff: pqr_aff; _x,_y: integer);
begin
  _q[0] := (_aff^.inv[0][0]*(_x-_aff^.x0) + _aff^.inv[0][1]*(_y-_aff^.y0) + ((1 shl _aff^.ires) shr 1)) {>>}div (1 shl _aff^.ires);
  _q[1] := (_aff^.inv[1][0]*(_x-_aff^.x0) + _aff^.inv[1][1]*(_y-_aff^.y0) + ((1 shl _aff^.ires) shr 1)) {>>}div (1 shl _aff^.ires);
end;

(*Map from the square domain into the image (at subpel resolution).*)
procedure qr_aff_project(out _p: qr_point; _aff: pqr_aff; _u,_v: integer);
begin
  _p[0] := (_aff^.fwd[0][0]*_u + _aff^.fwd[0][1]*_v + (1 shl (_aff^.res-1))) {>>}div (1 shl _aff^.res) + _aff^.x0;
  _p[1] := (_aff^.fwd[1][0]*_u + _aff^.fwd[1][1]*_v + (1 shl (_aff^.res-1))) {>>}div (1 shl _aff^.res) + _aff^.y0;
end;


function qr_cmp_edge_pt(_a,_b,context: pointer): integer;
var
  a,b: pqr_finder_edge_pt;
begin
  a:=pqr_finder_edge_pt(_a);
  b:=pqr_finder_edge_pt(_b);
  Result := ((Ord(a^.edge > b^.edge)-Ord(a^.edge < b^.edge)) shl 1)+
              Ord(a^.extent > b^.extent)-Ord(a^.extent<b^.extent);
end;

(*Computes the index of the edge each edge point belongs to, and its (signed)
   distance along the corresponding axis from the center of the finder pattern
   (in the square domain).
  The resulting list of edge points is sorted by edge index, with ties broken
   by extent.*)
procedure qr_finder_edge_pts_aff_classify(_f: pqr_finder; const _aff: pqr_aff);
var
  c: pqr_finder_center;
  i, e, d: integer;
  q: qr_point;
begin
  c := _f^.c;
  for e:=0 to 3 do _f^.nedge_pts[e]:=0;
  for i:=0 to c^.nedge_pts-1 do begin
    qr_aff_unproject(q, _aff, c^.edge_pts[i].pos[0], c^.edge_pts[i].pos[1]);
    qr_point_translate(q, -_f^.o[0], -_f^.o[1]);
    d:=Ord(abs(q[1])>abs(q[0]));
    e:=(d shl 1) or Ord(q[d]>=0);
    Inc(_f^.nedge_pts[e]);
    c^.edge_pts[i].edge := e;
    c^.edge_pts[i].extent := q[d];
  end;
  qsort(c^.edge_pts, 0,c^.nedge_pts-1, sizeof(c^.edge_pts^), @qr_cmp_edge_pt);
  _f^.edge_pts[0] := c^.edge_pts;
  for e:=1 to 3 do _f^.edge_pts[e]:=_f^.edge_pts[e-1]+_f^.nedge_pts[e-1];
end;

(*Computes the index of the edge each edge point belongs to, and its (signed)
   distance along the corresponding axis from the center of the finder pattern
   (in the square domain).
  The resulting list of edge points is sorted by edge index, with ties broken
   by extent.*)
procedure qr_finder_edge_pts_hom_classify(_f: pqr_finder; const _hom: pqr_hom);
var
  c: pqr_finder_center;
  i, e, d: integer;
  q: qr_point;
begin
  c := _f^.c;
  for e:=0 to 3 do _f^.nedge_pts[e]:=0;
  for i:=0 to c^.nedge_pts-1 do begin
    if(qr_hom_unproject(q, _hom, c^.edge_pts[i].pos[0], c^.edge_pts[i].pos[1])>=0) then begin
      qr_point_translate(q, -_f^.o[0], -_f^.o[1]);
      d:=Ord(abs(q[1])>abs(q[0]));
      e:=(d shl 1) or Ord(q[d]>=0);
      Inc(_f^.nedge_pts[e]);
      c^.edge_pts[i].edge := e;
      c^.edge_pts[i].extent := q[d];
    end
    else begin
      c^.edge_pts[i].edge := 4;
      c^.edge_pts[i].extent := q[0];
    end;
  end;
  qsort(c^.edge_pts, 0, c^.nedge_pts-1, sizeof(c^.edge_pts^), @qr_cmp_edge_pt);
  _f^.edge_pts[0] := c^.edge_pts;
  for e:=1 to 3 do _f^.edge_pts[e]:=_f^.edge_pts[e-1]+_f^.nedge_pts[e-1];
end;


(*Estimates the size of a module after classifying the edge points.
  _width:  The distance between UL and UR in the square domain.
  _height: The distance between UL and DL in the square domain.*)
function qr_finder_estimate_module_size_and_version(_f: pqr_finder; _width,_height: integer): integer;
var
  offs: qr_point;
  sums, nsums: array[0..3] of integer;
  usize, nusize, vsize, nvsize, uversion, vversion, e: integer;
  edge_pts: pqr_finder_edge_pt;
  sum, mean, n, i: integer;
begin
  offs[0]:=0;
  offs[1]:=0;
  for e:=0 to 3 do if(_f^.nedge_pts[e]>0) then begin
    (*Average the samples for this edge, dropping the top and bottom 25%.*)
    edge_pts:=_f^.edge_pts[e];
    n:=_f^.nedge_pts[e];
    sum:=0;
    for i:=(n shr 2) to n-(n shr 2)-1 do Inc(sum, edge_pts[i].extent);
    n := n - ((n shr 2) shl 1);
    mean:=QR_DIVROUND(sum,n);
    Inc(offs[e shr 1], mean);
    sums[e]:=sum;
    nsums[e]:=n;
  end
  else begin
    nsums[e]:=0;
    sums[e]:=0;
  end;
  (*If we have samples on both sides of an axis, refine our idea of where the
     unprojected finder center is located.*)
  if(_f^.nedge_pts[0]>0) and (_f^.nedge_pts[1]>0) then begin
    Dec(_f^.o[0], offs[0] {shr 1}div 2);
    Dec(sums[0], (offs[0]*nsums[0]) {shr 1}div 2);
    Dec(sums[1], (offs[0]*nsums[1]) {shr 1}div 2);
  end;
  if(_f^.nedge_pts[2]>0) and (_f^.nedge_pts[3]>0) then begin
    Dec(_f^.o[1], offs[1] {shr 1}div 2);
    Dec(sums[2], (offs[1]*nsums[2]) {shr 1}div 2);
    Dec(sums[3], (offs[1]*nsums[3]) {shr 1}div 2);
  end;

  (*We must have _some_ samples along each axis... if we don't, our transform
     must be pretty severely distorting the original square (e.g., with
     coordinates so large as to cause overflow).*)
  nusize:=nsums[0]+nsums[1];
  if (nusize<=0) then Exit(-1);
  (*The module size is 1/3 the average edge extent.*)
  nusize:=nusize*3;
  usize:=sums[1]-sums[0];
  usize:=((usize shl 1)+nusize) div (nusize shl 1);
  if (usize<=0) then Exit(-1);
  (*Now estimate the version directly from the module size and the distance
     between the finder patterns.
    This is done independently using the extents along each axis.
    If either falls significantly outside the valid range (1 to 40), reject the
     configuration.*)
  uversion:=(_width-8*usize) div (usize shl 2);
  if(uversion<1) or (uversion>40+QR_LARGE_VERSION_SLACK) then Exit(-1);

  (*Now do the same for the other axis.*)
  nvsize:=nsums[2]+nsums[3];
  if (nvsize<=0) then Exit(-1);
  nvsize:=nvsize*3;
  vsize:=sums[3]-sums[2];
  vsize:=((vsize shl 1)+nvsize) div (nvsize shl 1);
  if(vsize<=0) then Exit(-1);
  vversion:=(_height-8*vsize) div (vsize shl 2);
  if(vversion<1) or (vversion>40+QR_LARGE_VERSION_SLACK) then Exit(-1);
  (*If the estimated version using extents along one axis is significantly
     different than the estimated version along the other axis, then the axes
     have significantly different scalings (relative to the grid).
    This can happen, e.g., when we have multiple adjacent QR codes, and we've
     picked two finder patterns from one and the third finder pattern from
     another, e.g.:
      X---DL UL---X
      |....   |....
      X....  UR....
    Such a configuration might even pass any other geometric checks if we
     didn't reject it here.*)
  if (abs(uversion-vversion)>QR_LARGE_VERSION_SLACK) then Exit(-1);
  _f^.size[0] := usize;
  _f^.size[1] := vsize;
  (*We intentionally do not compute an average version from the sizes along
     both axes.
    In the presence of projective distortion, one of them will be much more
     accurate than the other.*)
  _f^.eversion[0]:=uversion;
  _f^.eversion[1]:=vversion;
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_finder_estimate_module_size_and_version: version=%d,%d; size=%d,%d',[uversion,vversion, usize,vsize]);{$ENDIF}
  Exit(0);
end;

(*Eliminate outliers from the classified edge points with RANSAC.*)
procedure qr_finder_ransac(_f: pqr_finder; _hom: pqr_aff; _isaac: pisaac_ctx; _e: integer);
var
  edge_pts: pqr_finder_edge_pt;
  best_ninliers, n: integer;
  max_iters,i,j: integer;
  q0,q1: qr_point;
  ninliers, thresh, p0i, p1i: integer;
  p0, p1: PInteger;
  tmp: qr_finder_edge_pt;
begin
  edge_pts:=_f^.edge_pts[_e];
  n:=_f^.nedge_pts[_e];
  best_ninliers:=0;
  if (n>1) then begin
    (*17 iterations is enough to guarantee an outlier-free sample with more
       than 99% probability given as many as 50% outliers.*)
    max_iters:=17;
    for i:=0 to max_iters-1 do begin
      (*Pick two random points on this edge.*)
      p0i:=isaac_next_uint(_isaac^, n);
      p1i:=isaac_next_uint(_isaac^, n-1);
      if (p1i>=p0i) then Inc(p1i);
      p0:=edge_pts[p0i].pos;
      p1:=edge_pts[p1i].pos;
      (*If the corresponding line is not within 45 degrees of the proper
         orientation in the square domain, reject it outright.
        This can happen, e.g., when highly skewed orientations cause points to
         be misclassified into the wrong edge.
        The irony is that using such points might produce a line which _does_
         pass the corresponding validity checks.*)
      qr_aff_unproject(q0,_hom,p0[0],p0[1]);
      qr_aff_unproject(q1,_hom,p1[0],p1[1]);
      qr_point_translate(q0, -_f^.o[0], -_f^.o[1]);
      qr_point_translate(q1, -_f^.o[0], -_f^.o[1]);
      Assert(_e >= 0);
      if (abs(q0[_e shr 1]-q1[_e shr 1]) > abs(q0[1-(_e shr 1)]-q1[1-(_e shr 1)])) then continue;
      (*Identify the other edge points which are inliers.
        The squared distance should be distributed as a \Chi^2 distribution
         with one degree of freedom, which means for a 95% confidence the
         point should lie within a factor 3.8414588 ~= 4 times the expected
         variance of the point locations.
        We grossly approximate the standard deviation as 1 pixel in one
         direction, and 0.5 pixels in the other (because we average two
         coordinates).*)
      thresh := qr_isqrt(qr_point_distance2(p0,p1) shl (2*QR_FINDER_SUBPREC+1));
      ninliers:=0;
      for j:=0 to n-1 do begin
        if (abs(qr_point_ccw(p0,p1,edge_pts[j].pos))<=thresh) then begin
          edge_pts[j].extent := edge_pts[j].extent or 1;
          Inc(ninliers);
        end
        else
          edge_pts[j].extent := edge_pts[j].extent and (not 1);
      end;
      if(ninliers>best_ninliers) then begin
        for j:=0 to n-1 do edge_pts[j].extent := edge_pts[j].extent shl 1;
        best_ninliers:=ninliers;
        (*The actual number of iterations required is
            log(1-\alpha)/log(1-r*r),
           where \alpha is the required probability of taking a sample with
            no outliers (e.g., 0.99) and r is the estimated ratio of inliers
            (e.g. ninliers/n).
          This is just a rough (but conservative) approximation, but it
           should be good enough to stop the iteration early when we find
           a good set of inliers.*)
        if(ninliers > n shr 1) then max_iters:=(67*n-63*ninliers-1) div (n shl 1);
      end;
    end;
    (*Now collect all the inliers at the beginning of the list.*)
    i:=0;
    j:=0;
    while j<best_ninliers do begin
      if (edge_pts[i].extent and 2) <> 0 then begin
        if(j<i) then begin
          tmp:=edge_pts[i];
          edge_pts[j]:=edge_pts[i];
          edge_pts[i]:=tmp;
        end;
        Inc(j);
      end;
      Inc(i);
    end;
  end;
  _f^.ninliers[_e]:=best_ninliers;
end;

(*Perform a least-squares line fit to an edge of a finder pattern using the
   inliers found by RANSAC.*)
function qr_line_fit_finder_edge(var _l: qr_line; _f: pqr_finder; _e,_res: integer): integer;
var
  edge_pts: pqr_finder_edge_pt;
  pts: pqr_point;
  npts,i: integer;
begin
  npts:=_f^.ninliers[_e];
  if(npts<2) then Exit(-1);
  (*We could write a custom version of qr_line_fit_points that accesses
     edge_pts directly, but this saves on code size and doesn't measurably slow
     things down.*)
  pts:=GetMem(npts*sizeof(pts^));
  edge_pts:=_f^.edge_pts[_e];
  for i:=0 to npts-1 do begin
    pts[i][0]:=edge_pts[i].pos[0];
    pts[i][1]:=edge_pts[i].pos[1];
  end;
  qr_line_fit_points(_l,pts,npts,_res);
  (*Make sure the center of the finder pattern lies in the positive halfspace
     of the line.*)
  qr_line_orient(_l, _f^.c^.pos[0], _f^.c^.pos[1]);
  FreeMem(pts);
  Exit(0);
end;

(*Perform a least-squares line fit to a pair of common finder edges using the
   inliers found by RANSAC.
  Unlike a normal edge fit, we guarantee that this one succeeds by creating at
   least one point on each edge using the estimated module size if it has no
   inliers.*)
procedure qr_line_fit_finder_pair(out _l: qr_line; _aff: pqr_aff; _f0,_f1: pqr_finder; _e: integer);
var
  pts: pqr_point;
  npts: integer;
  edge_pts: pqr_finder_edge_pt;
  q: qr_point;
  n0,n1,i: integer;
begin
  Assert(_e>=0);
  n0:=_f0^.ninliers[_e];
  n1:=_f1^.ninliers[_e];
  (*We could write a custom version of qr_line_fit_points that accesses
     edge_pts directly, but this saves on code size and doesn't measurably slow
     things down.*)
  npts:=QR_MAXI(n0,1)+QR_MAXI(n1,1);
  pts:=GetMem(npts*sizeof(pts^));
  if(n0>0) then begin
    edge_pts:=_f0^.edge_pts[_e];
    for i:=0 to n0-1 do begin
      pts[i][0]:=edge_pts[i].pos[0];
      pts[i][1]:=edge_pts[i].pos[1];
    end;
  end
  else begin
    q[0]:=_f0^.o[0];
    q[1]:=_f0^.o[1];
    Inc(q[_e shr 1], _f0^.size[_e shr 1]*(2*(_e and 1)-1));
    qr_aff_project(pts[0],_aff,q[0],q[1]);
    Inc(n0);
  end;
  if(n1>0) then begin
    edge_pts:=_f1^.edge_pts[_e];
    for i:=0 to n1-1 do begin
      pts[n0+i][0]:=edge_pts[i].pos[0];
      pts[n0+i][1]:=edge_pts[i].pos[1];
    end
  end
  else begin
    q[0]:=_f1^.o[0];
    q[1]:=_f1^.o[1];
    Inc(q[_e shr 1], _f1^.size[_e shr 1]*(2*(_e and 1)-1));
    qr_aff_project(pts[n0],_aff,q[0],q[1]);
    Inc(n1);
  end;
  qr_line_fit_points(_l,pts,npts,_aff^.res);
  (*Make sure at least one finder center lies in the positive halfspace.*)
  qr_line_orient(_l, _f0^.c^.pos[0], _f0^.c^.pos[1]);
  FreeMem(pts);
end;

function qr_finder_quick_crossing_check(_img: PByte; _width,_height, _x0,_y0, _x1,_y1, _v: integer): integer;
begin
  Assert((_v=0) or (_v=1));
  (*The points must be inside the image, and have a !_v:_v:!_v pattern.
    We don't scan the whole line initially, but quickly reject if the endpoints
     aren't !_v, or the midpoint isn't _v.
    If either end point is out of the image, or we don't encounter a _v pixel,
     we return a negative value, indicating the region should be considered
     empty.
    Otherwise, we return a positive value to indicate it is non-empty.*)
  if (_x0<0) or (_x0>=_width) or (_y0<0) or (_y0>=_height) or
     (_x1<0) or (_x1>=_width) or (_y1<0) or (_y1>=_height) then
     Result := -1
  else if (Ord(_img[_y0*_width+_x0]=0)<>_v) or (Ord(_img[_y1*_width+_x1]=0)<>_v) then
    Result := 1
  else if(Ord(_img[((_y0+_y1) {>>1}div 2)*_width + ((_x0+_x1) {>>1}div 2)]=0)=_v) then
    Result := -1
  else
    Result := 0;
end;

(*Locate the midpoint of a _v segment along a !_v:_v:!_v line from (_x0,_y0) to
   (_x1,_y1).
  All coordinates, which are NOT in subpel resolution, must lie inside the
   image, and the endpoints are already assumed to have the value !_v.
  The returned value is in subpel resolution.
*)
function qr_finder_locate_crossing(_img: PByte; _width,_height: integer;  _x0,_y0, _x1,_y1, _v: integer; out _p: qr_point): integer;
var
  x0, x1, dx: qr_point;
  step: array[0..1] of integer;
  steep, err, derr: integer;
begin
  Assert((_v=0) or (_v=1));
  (* Use Bresenham's algorithm to trace along the line and find the exact
     transitions from !_v to _v and back. *)
  x0[0]:=_x0;
  x0[1]:=_y0;
  x1[0]:=_x1;
  x1[1]:=_y1;
  dx[0]:=abs(_x1-_x0);
  dx[1]:=abs(_y1-_y0);
  steep:=Ord(dx[1]>dx[0]);
  err:=0;
  derr:=dx[1-steep];
  step[0]:=(Ord(_x0<_x1) shl 1)-1;
  step[1]:=(Ord(_y0<_y1) shl 1)-1;
  (*Find the first crossing from !_v to _v.*)
  while True do begin
    (*If we make it all the way to the other side, there's no crossing.*)
    if (x0[steep]=x1[steep]) then Exit(-1);
    Inc(x0[steep], step[steep]);
    Inc(err, derr);
    if (err shl 1 > dx[steep]) then begin
      Inc(x0[1-steep], step[1-steep]);
      Dec(err, dx[steep]);
    end;
    if (Ord(_img[x0[1]*_width+x0[0]]=0) <> _v) then break;
  end;
  (*Find the last crossing from _v to !_v.*)
  err:=0;
  while True do begin
    if(x0[steep]=x1[steep]) then break;
    Dec(x1[steep], step[steep]);
    Inc(err, derr);
    if (err shl 1 > dx[steep]) then begin
      Dec(x1[1-steep], step[1-steep]);
      Dec(err, dx[steep]);
    end;
    if (Ord(_img[x1[1]*_width+x1[0]]=0) <> _v) then break;
  end;
  (*Return the midpoint of the _v segment.*)
  _p[0] := ((x0[0]+x1[0]+1) shl QR_FINDER_SUBPREC) shr 1;
  _p[1] := ((x0[1]+x1[1]+1) shl QR_FINDER_SUBPREC) shr 1;
  Exit(0);
end;

function qr_aff_line_step(_aff: pqr_aff; const _l: qr_line; _v,_du: integer; out _dv: integer): integer;
var
  shift, round, dv, n, d: integer;
begin
  n := _aff^.fwd[0][_v]*_l[0] + _aff^.fwd[1][_v]*_l[1];
  d := _aff^.fwd[0][1-_v]*_l[0] + _aff^.fwd[1][1-_v]*_l[1];
  if(d<0) then begin
    n:=-n;
    d:=-d;
  end;
  shift:=QR_MAXI(0,qr_ilog(_du)+qr_ilog(abs(n))+3-QR_INT_BITS);
  round:=(1 shl shift) shr 1;
  n := (n+round) div (1 shl shift);
  d := (d+round) div (1 shl shift);
  (*The line should not be outside 45 degrees of horizontal/vertical.
    TODO: We impose this restriction to help ensure the loop below terminates,
     but it should not technically be required.
    It also, however, ensures we avoid division by zero.*)
  if(abs(n)>=d) then Exit(-1);
  n:=-_du*n;
  dv:=QR_DIVROUND(n,d);
  if(abs(dv)>=_du) then Exit(-1);
  _dv:=dv;
  Exit(0);
end;

function qr_hom_fit(_hom: pqr_hom; _ul, _ur, _dl: pqr_finder; _p{[4]}: pqr_point; _aff: pqr_aff; _isaac: pisaac_ctx;
  _img: PByte; _width,_height: integer): integer;
var
  b, r: pqr_point;
  p,q: qr_point;
  nb, cb, nr, cr: integer;
  l: array[0..3] of qr_line;
  ox, oy, ru, rv, dru, drv, bu,bv, dbu, dbv, rx, ry, drxi, dryi, drxj, dryj, nrempty, rlastfit, bx, by: integer;
  dbxi, dbyi, dbxj, dbyj, nbempty, blastfit, shift, round, version4, brx, bry, i: integer;
  rdone,bdone: boolean;
  ret, x0,y0, x1,y1: integer;
  cell: qr_hom_cell;
  p3: qr_point;
  dim: integer;
  w, mask: int64;
  c21, dx21, dy21: integer;
begin
  (*We attempt to correct large-scale perspective distortion by fitting lines
     to the edge of the code area.
    We could also look for an alignment pattern now, but that wouldn't work for
     version 1 codes, which have no alignment pattern.
    Even if the code is supposed to have one, there's go guarantee we'd find it
     intact.*)
  (*Fitting lines is easy for the edges on which we have two finder patterns.
    After the fit, UL is guaranteed to be on the proper side, but if either of
     the other two finder patterns aren't, something is wrong.*)
  qr_finder_ransac(_ul,_aff,_isaac,0);
  qr_finder_ransac(_dl,_aff,_isaac,0);
  qr_line_fit_finder_pair(l[0],_aff,_ul,_dl,0);
  if (qr_line_eval(l[0], _dl^.c^.pos[0], _dl^.c^.pos[1])<0) or
     (qr_line_eval(l[0], _ur^.c^.pos[0], _ur^.c^.pos[1])<0) then Exit(-1);

  qr_finder_ransac(_ul,_aff,_isaac,2);
  qr_finder_ransac(_ur,_aff,_isaac,2);
  qr_line_fit_finder_pair(l[2],_aff,_ul,_ur,2);
  if (qr_line_eval(l[2],_dl^.c^.pos[0],_dl^.c^.pos[1])<0) or
     (qr_line_eval(l[2],_ur^.c^.pos[0],_ur^.c^.pos[1])<0) then Exit(-1);

  (*The edges which only have one finder pattern are more difficult.
    We start by fitting a line to the edge of the one finder pattern we do
     have.
    This can fail due to an insufficient number of sample points, and even if
     it succeeds can be fairly inaccurate, because all of the points are
     clustered in one corner of the QR code.
    If it fails, we just use an axis-aligned line in the affine coordinate
     system.
    Then we walk along the edge of the entire code, looking for
     light:dark:light patterns perpendicular to the edge.
    Wherever we find one, we take the center of the dark portion as an
     additional sample point.
    At the end, we re-fit the line using all such sample points found.*)
  Assert(_ur^.size[1]>=0);
  drv:=_ur^.size[1] shr 1;
  qr_finder_ransac(_ur,_aff,_isaac,1);
  if(qr_line_fit_finder_edge(l[1],_ur,1,_aff^.res)>=0) then begin
    if (qr_line_eval(l[1],_ul^.c^.pos[0],_ul^.c^.pos[1])<0) or
       (qr_line_eval(l[1],_dl^.c^.pos[0],_dl^.c^.pos[1])<0) then Exit(-1);
    (*Figure out the change in ru for a given change in rv when stepping along
       the fitted line.*)
    if(qr_aff_line_step(_aff,l[1],1,drv,dru)<0) then Exit(-1);
  end
  else
    dru:=0;
  ru:=_ur^.o[0] + 3*_ur^.size[0] - 2*dru;
  rv:=_ur^.o[1] - 2*drv;

  Assert(_dl^.size[0]>=0);
  dbu:=_dl^.size[0] shr 1;
  qr_finder_ransac(_dl,_aff,_isaac,3);
  if(qr_line_fit_finder_edge(l[3],_dl,3,_aff^.res)>=0) then begin
    if (qr_line_eval(l[3],_ul^.c^.pos[0],_ul^.c^.pos[1])<0) or
       (qr_line_eval(l[3],_ur^.c^.pos[0],_ur^.c^.pos[1])<0) then Exit(-1);
    (*Figure out the change in bv for a given change in bu when stepping along
       the fitted line.*)
    if(qr_aff_line_step(_aff,l[3],0,dbu,dbv)<0) then Exit(-1);
  end
  else
    dbv:=0;
  bu:=_dl^.o[0] - 2*dbu;
  bv:=_dl^.o[1] + 3*_dl^.size[1] - 2*dbv;

  (*Set up the initial point lists.*)
  rlastfit:=_ur^.ninliers[1];
  nr:=rlastfit;
  cr:=nr+(_dl^.o[1]-rv+drv-1) div drv;
  r:=GetMem(cr*sizeof(r^));
  for i:=0 to _ur^.ninliers[1]-1 do
    Move(_ur^.edge_pts[1][i].pos, r[i], sizeof(r[i]));

  blastfit:=_dl^.ninliers[3];
  nb:=blastfit;
  cb:=nb+(_ur^.o[0]-bu+dbu-1) div dbu;
  b:=GetMem(cb*sizeof(b^));
  for i:=0 to _dl^.ninliers[3]-1 do
    Move(_dl^.edge_pts[3][i].pos, b[i], sizeof(b[i]));

  (*Set up the step parameters for the affine projection.*)
  ox:=(_aff^.x0 shl _aff^.res)+(1 shl (_aff^.res-1));
  oy:=(_aff^.y0 shl _aff^.res)+(1 shl (_aff^.res-1));
  rx:=_aff^.fwd[0][0]*ru+_aff^.fwd[0][1]*rv+ox;
  ry:=_aff^.fwd[1][0]*ru+_aff^.fwd[1][1]*rv+oy;
  drxi:=_aff^.fwd[0][0]*dru+_aff^.fwd[0][1]*drv;
  dryi:=_aff^.fwd[1][0]*dru+_aff^.fwd[1][1]*drv;
  drxj:=_aff^.fwd[0][0]*_ur^.size[0];
  dryj:=_aff^.fwd[1][0]*_ur^.size[0];
  bx:=_aff^.fwd[0][0]*bu+_aff^.fwd[0][1]*bv+ox;
  by:=_aff^.fwd[1][0]*bu+_aff^.fwd[1][1]*bv+oy;
  dbxi:=_aff^.fwd[0][0]*dbu+_aff^.fwd[0][1]*dbv;
  dbyi:=_aff^.fwd[1][0]*dbu+_aff^.fwd[1][1]*dbv;
  dbxj:=_aff^.fwd[0][1]*_dl^.size[1];
  dbyj:=_aff^.fwd[1][1]*_dl^.size[1];

  (*Now step along the lines, looking for new sample points.*)
  nrempty:=0;
  nbempty:=0;
  while True do begin
    (*If we take too many steps without encountering a non-zero pixel, assume
       we have wandered off the edge and stop looking before we hit the other
       side of the quiet region.
      Otherwise, stop when the lines cross (if they do so inside the affine
       region) or come close to crossing (outside the affine region).
      TODO: We don't have any way of detecting when we've wandered into the
       code interior; we could stop if the outside sample ever shows up dark,
       but this could happen because of noise in the quiet region, too.*)
    rdone := (rv>=QR_MINI(bv,(_dl^.o[1]+bv) div 2)) or (nrempty>14);
    bdone := (bu>=QR_MINI(ru,(_ur^.o[0]+ru) div 2)) or (nbempty>14);
    if (not rdone and (bdone or (rv<bu))) then begin
      x0:=(rx+drxj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      y0:=(ry+dryj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      x1:=(rx-drxj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      y1:=(ry-dryj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      if(nr>=cr) then begin
        cr:= (cr shl 1) or 1;
        r:=ReallocMem(r, cr*sizeof(r^));
      end;
      ret:=qr_finder_quick_crossing_check(_img,_width,_height,x0,y0,x1,y1,1);
      if(ret=0) then
        ret:=qr_finder_locate_crossing(_img,_width,_height,x0,y0,x1,y1,1,r[nr]);

      if(ret>=0) then begin
        if (ret=0) then begin
          qr_aff_unproject(q,_aff,r[nr][0],r[nr][1]);
          (*Move the current point halfway towards the crossing.
            We don't move the whole way to give us some robustness to noise.*)
          ru:=(ru+q[0]) {>>1}div 2;
          (*But ensure that rv monotonically increases.*)
          if (q[1]+drv>rv) then rv:=(rv+q[1]) div 2;
          rx:=_aff^.fwd[0][0]*ru+_aff^.fwd[0][1]*rv+ox;
          ry:=_aff^.fwd[1][0]*ru+_aff^.fwd[1][1]*rv+oy;
          Inc(nr);
          (*Re-fit the line to update the step direction periodically.*)
          if(nr>QR_MAXI(1,rlastfit+(rlastfit {>>2}div 4))) then begin
            qr_line_fit_points(l[1],r,nr,_aff^.res);
            if(qr_aff_line_step(_aff,l[1],1,drv,dru)>=0) then begin
              drxi:=_aff^.fwd[0][0]*dru+_aff^.fwd[0][1]*drv;
              dryi:=_aff^.fwd[1][0]*dru+_aff^.fwd[1][1]*drv;
            end;
            rlastfit:=nr;
          end;
        end;
        nrempty:=0;
      end
      else
        Inc(nrempty);
      Inc(ru, dru);
      (*Our final defense: if we overflow, stop.*)
      if (rv+drv>rv) then Inc(rv, drv)
      else nrempty:=High(integer);
      Inc(rx, drxi);
      Inc(ry, dryi);
    end
    else if( not bdone) then begin
      x0:=(bx+dbxj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      y0:=(by+dbyj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      x1:=(bx-dbxj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      y1:=(by-dbyj) div (1 shl (_aff^.res+QR_FINDER_SUBPREC));
      if(nb>=cb) then begin
        cb:= (cb shl 1) or 1;
        b:=ReallocMem(b,cb*sizeof(b^));
      end;
      ret:=qr_finder_quick_crossing_check(_img,_width,_height,x0,y0,x1,y1,1);
      if(ret=0) then
        ret:=qr_finder_locate_crossing(_img,_width,_height,x0,y0,x1,y1,1,b[nb]);

      if(ret>=0) then begin
        if(ret=0) then begin
          qr_aff_unproject(q,_aff,b[nb][0],b[nb][1]);
          (*Move the current point halfway towards the crossing.
            We don't move the whole way to give us some robustness to noise.*)
          (*But ensure that bu monotonically increases.*)
          if(q[0]+dbu>bu) then bu:=(bu+q[0]) div 2;
          bv:=(bv+q[1]) div 2;
          bx:=_aff^.fwd[0][0]*bu+_aff^.fwd[0][1]*bv+ox;
          by:=_aff^.fwd[1][0]*bu+_aff^.fwd[1][1]*bv+oy;
          Inc(nb);
          (*Re-fit the line to update the step direction periodically.*)
          if(nb>QR_MAXI(1,blastfit+(blastfit {>>2}div 4))) then begin
            qr_line_fit_points(l[3],b,nb,_aff^.res);
            if(qr_aff_line_step(_aff,l[3],0,dbu,dbv)>=0) then begin
              dbxi:=_aff^.fwd[0][0]*dbu+_aff^.fwd[0][1]*dbv;
              dbyi:=_aff^.fwd[1][0]*dbu+_aff^.fwd[1][1]*dbv;
            end;
            blastfit:=nb;
          end;
        end;
        nbempty:=0;
      end
      else Inc(nbempty);
      (*Our final defense: if we overflow, stop.*)
      if(bu+dbu>bu) then Inc(bu, dbu)
      else nbempty:=High(integer);
      Inc(bv, dbv);
      Inc(bx, dbxi);
      Inc(by, dbyi);
    end
    else break;
  end;
  (*Fit the new lines.
    If we _still_ don't have enough sample points, then just use an
     axis-aligned line from the affine coordinate system (e.g., one parallel
     to the opposite edge in the image).*)
  if(nr>1) then qr_line_fit_points(l[1],r,nr,_aff^.res)
  else begin
    qr_aff_project(p,_aff,_ur^.o[0]+3*_ur^.size[0],_ur^.o[1]);
    shift:=QR_MAXI(0, qr_ilog(QR_MAXI(abs(_aff^.fwd[0][1]),abs(_aff^.fwd[1][1]))) - ((_aff^.res+1) shr 1));
    round:=(1 shl shift) shr 1;
    l[1][0]:=(_aff^.fwd[1][1]+round) div (1 shl shift);
    l[1][1]:=(-_aff^.fwd[0][1]+round) div (1 shl shift);
    l[1][2]:=-(l[1][0]*p[0]+l[1][1]*p[1]);
  end;
  FreeMem(r);
  if(nb>1) then qr_line_fit_points(l[3],b,nb,_aff^.res)
  else begin
    qr_aff_project(p,_aff,_dl^.o[0],_dl^.o[1]+3*_dl^.size[1]);
    shift:=QR_MAXI(0, qr_ilog(QR_MAXI(abs(_aff^.fwd[0][1]),abs(_aff^.fwd[1][1]))) - ((_aff^.res+1) shr 1));
    round:=(1 shl shift) shr 1;
    l[3][0]:=(_aff^.fwd[1][0]+round) div (1 shl shift);
    l[3][1]:=(-_aff^.fwd[0][0]+round) div (1 shl shift);
    l[3][2]:=-(l[1][0]*p[0]+l[1][1]*p[1]);
  end;
  FreeMem(b);
  for i:=0 to 3 do begin
    if(qr_line_isect(_p[i], l[i and 1], l[2+(i shr 1)])<0) then Exit(-1);
    (*It's plausible for points to be somewhat outside the image, but too far
       and too much of the pattern will be gone for it to be decodable.*)
    if (_p[i][0] < -_width shl QR_FINDER_SUBPREC) or
       (_p[i][0] >= _width shl (QR_FINDER_SUBPREC+1)) or
       (_p[i][1] < -_height shl QR_FINDER_SUBPREC) or
       (_p[i][1] >= _height shl (QR_FINDER_SUBPREC+1)) then Exit(-1);
  end;
  (*By default, use the edge intersection point for the bottom-right corner.*)
  brx:=_p[3][0];
  bry:=_p[3][1];
  (*However, if our average version estimate is greater than 1, NOW we try to
     search for an alignment pattern.
    We get a much better success rate by doing this after our initial attempt
     to promote the transform to a homography than before.
    You might also think it would be more reliable to use the interior finder
     pattern edges, since the outer ones may be obscured or damaged, and it
     would save us a reprojection below, since they would form a nice square
     with the location of the alignment pattern, but this turns out to be a bad
     idea.
    Non-linear distortion is usually maximal on the outside edge, and thus
     estimating the grid position from points on the interior means we might
     get mis-aligned by the time we reach the edge.*)
  version4:=_ul^.eversion[0]+_ul^.eversion[1]+_ur^.eversion[0]+_dl^.eversion[1];
  if(version4>4) then begin
    {$IFDEF DEBUG_ZBAR}DebugLn('qr_hom_fit: UL=(%d,%d), UR=(%d,%d), DL=(%d,%d)',[_ul^.o[0],_ul^.o[1], _ur^.o[0],_ur^.o[1], _dl^.o[0],_dl^.o[1]]);{$ENDIF}
    dim:=17+version4;
    // 0 - 1
    // |   |
    // 2 - 3
    qr_hom_cell_init(cell, 0,0, dim-1,0, 0,dim-1, dim-1,dim-1,
      _p[0][0],_p[0][1], _p[1][0],_p[1][1], _p[2][0],_p[2][1], _p[3][0],_p[3][1]);
    if(qr_alignment_pattern_search(p3, @cell, dim-7, dim-7, 4, _img,_width,_height)>=0) then begin
      (*There's no real need to update the bounding box corner, and in fact we
         actively perform worse if we do.
        Clearly it was good enough for us to find this alignment pattern, so
         it should be good enough to use for grid initialization.
        The point of doing the search was to get more accurate version
         estimates and a better chance of decoding the version and format info.
        This is particularly important for small versions that have no encoded
         version info, since any mismatch in version renders the code
         undecodable.*)
      (*We do, however, need four points in a square to initialize our
         homography, so project the point from the alignment center to the
         corner of the code area.*)
      c21:=_p[2][0]*_p[1][1]-_p[2][1]*_p[1][0];
      dx21:=_p[2][0]-_p[1][0];
      dy21:=_p[2][1]-_p[1][1];
      w:=QR_EXTMUL(dim-7, c21,
          QR_EXTMUL(dim-13, _p[0][0]*dy21-_p[0][1]*dx21,
           QR_EXTMUL(6, p3[0]*dy21-p3[1]*dx21, 0)));
      (*The projection failed: invalid geometry.*)
      if(w=0) then Exit(-1);
      mask := QR_SIGNMASK(w);
      w := (w + mask) xor mask;
      brx:=QR_DIVROUND((QR_EXTMUL((dim-7)*_p[0][0], p3[0]*dy21,
                         QR_EXTMUL((dim-13)*p3[0], c21-_p[0][1]*dx21,
                          QR_EXTMUL(6*_p[0][0], c21-p3[1]*dx21, 0)))+mask) xor mask, w);
      bry:=QR_DIVROUND((QR_EXTMUL((dim-7)*_p[0][1], -p3[1]*dx21,
                         QR_EXTMUL((dim-13)*p3[1], c21+_p[0][0]*dy21,
                          QR_EXTMUL(6*_p[0][1], c21+p3[0]*dy21, 0)))+mask) xor mask, w);
    end;
  end;
  (*Now we have four points that map to a square: initialize the projection.*)
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_hom_fit: (%d,%d)-(%d,%d)-(%d,%d)-(%d,%d)',[_p[0][0],_p[0][1], _p[1][0],_p[1][1], _p[2][0],_p[2][1], brx,bry]);{$ENDIF}
  qr_hom_init(_hom^, _p[0][0], _p[0][1], _p[1][0],_p[1][1], _p[2][0],_p[2][1], brx, bry, QR_HOM_BITS);
  Exit(0);
end;


(*Allocates a client reader handle.*)
function _zbar_qr_create(): pqr_reader;
begin
  Result := AllocMem(sizeof(Result^));
  qr_reader_init(Result^);
end;

procedure _zbar_qr_destroy(reader: pqr_reader);
begin
  //zprintf(1, "max finder lines = %dx%d\n", reader^.finder_lines[0].clines, reader^.finder_lines[1].clines);
  if Assigned(reader^.finder_lines[0].lines) then
     FreeMem(reader^.finder_lines[0].lines);
  if Assigned(reader^.finder_lines[1].lines) then
     FreeMem(reader^.finder_lines[1].lines);
  FreeMem(reader);
end;

procedure qr_code_data_list_init(var _qrlist: qr_code_data_list);
begin
  _qrlist.qrdata:=nil;
  _qrlist.nqrdata:=0;
  _qrlist.cqrdata:=0;
end;

procedure qr_code_data_list_clear(var _qrlist: qr_code_data_list);
var i: integer;
begin
  for i:=0 to _qrlist.nqrdata-1 do qr_code_data_clear(_qrlist.qrdata[i]);
  FreeMem(_qrlist.qrdata);
  qr_code_data_list_init(_qrlist);
end;

procedure qr_code_data_list_add(_qrlist: pqr_code_data_list; _qrdata: pqr_code_data);
begin
  if (_qrlist^.nqrdata >= _qrlist^.cqrdata) then begin
    _qrlist^.cqrdata := (_qrlist^.cqrdata shl 1) or 1;
    _qrlist^.qrdata := ReallocMem(_qrlist^.qrdata, _qrlist^.cqrdata*sizeof(_qrlist^.qrdata^));
  end;
  Move(_qrdata^, _qrlist^.qrdata[_qrlist^.nqrdata{++}], sizeof(_qrdata^));
  Inc(_qrlist^.nqrdata);
end;

(* Searches for an arrangement of these three finder centers that yields a valid
   configuration.
  _c: On input, the three finder centers to consider in any order.
  Return: The detected version number, or a negative value on error.
*)
function qr_reader_try_configuration(_reader: pqr_reader;  _qrdata: pqr_code_data;  _img: PByte;  _width, _height: integer;  _c: ppqr_finder_center): integer;
var
  ci: array[0..6] of integer;
  maxd,d: unsigned;
  ccw, i0, i: integer;
  aff: qr_aff;
  hom: qr_hom;
  ul, ur, dl: qr_finder;
  bbox: array[0..3] of qr_point;
  res, ur_version, dl_version, fmt_info: integer;
begin
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_reader_try_configuration: Finder Centers (%d,%d), (%d,%d), (%d,%d)',[_c[0]^.pos[0] shr QR_FINDER_SUBPREC,_c[0]^.pos[1] shr QR_FINDER_SUBPREC, _c[1]^.pos[0] shr QR_FINDER_SUBPREC,_c[1]^.pos[1] shr QR_FINDER_SUBPREC, _c[2]^.pos[0] shr QR_FINDER_SUBPREC,_c[2]^.pos[1] shr QR_FINDER_SUBPREC]);{$ENDIF}
  (*Sort the points in counter-clockwise order.*)
  ccw := qr_point_ccw(_c[0]^.pos, _c[1]^.pos, _c[2]^.pos);

  (*Colinear points can't be the corners of a quadrilateral.*)
  if(ccw=0) then Exit(-1);

  (*Include a few extra copies of the cyclical list to avoid mods.*)
  ci[6]:=0;
  ci[3]:=0;
  ci[0]:=0;
  ci[1]:=1+Ord(ccw<0);
  ci[4]:=ci[1];
  ci[2]:=2-Ord(ccw<0);
  ci[5]:=ci[2];

  (*Assume the points farthest from each other are the opposite corners, and
     find the top-left point.*)
  maxd:=qr_point_distance2(_c[1]^.pos, _c[2]^.pos);
  i0:=0;
  for i:=1 to 2 do begin
    d:=qr_point_distance2(_c[ci[i+1]]^.pos, _c[ci[i+2]]^.pos);
    if(d>maxd) then begin
      i0:=i;
      maxd:=d;
    end;
  end;

  (*However, try all three possible orderings, just to be sure (a severely
     skewed projection could move opposite corners closer than adjacent).*)
  for i:=i0 to i0+2 do begin
    ul.c:=_c[ci[i]];    // lavy-horny
    ur.c:=_c[ci[i+1]];  // pravy-horny
    dl.c:=_c[ci[i+2]];  // lavy-dolny
    {$IFDEF DEBUG_ZBAR}DebugLn('qr_reader_try_configuration: UL=(%d,%d), UR=(%d,%d), DL=(%d,%d)',[ul.c^.pos[0] shr QR_FINDER_SUBPREC,ul.c^.pos[1] shr QR_FINDER_SUBPREC, ur.c^.pos[0] shr QR_FINDER_SUBPREC,ur.c^.pos[1] shr QR_FINDER_SUBPREC, dl.c^.pos[0] shr QR_FINDER_SUBPREC,dl.c^.pos[1] shr QR_FINDER_SUBPREC]);{$ENDIF}
    (*Estimate the module size and version number from the two opposite corners.
      The module size is not constant in the image, so we compute an affine
       projection from the three points we have to a square domain, and
       estimate it there.
      Although it should be the same along both axes, we keep separate
       estimates to account for any remaining projective distortion.*)
    res := QR_INT_BITS - 2 - QR_FINDER_SUBPREC - qr_ilog(QR_MAXI(_width,_height)-1);
    qr_aff_init(aff, ul.c^.pos, ur.c^.pos, dl.c^.pos, res);

    qr_aff_unproject(ur.o, @aff, ur.c^.pos[0], ur.c^.pos[1]);
    qr_finder_edge_pts_aff_classify(@ur,@aff);
    if qr_finder_estimate_module_size_and_version(@ur, 1 shl res, 1 shl res) < 0 then continue;

    qr_aff_unproject(dl.o, @aff, dl.c^.pos[0], dl.c^.pos[1]);
    qr_finder_edge_pts_aff_classify(@dl,@aff);
    if (qr_finder_estimate_module_size_and_version(@dl, 1 shl res, 1 shl res) < 0) then continue;

    (*If the estimated versions are significantly different, reject the
       configuration.*)
    if(abs(ur.eversion[1]-dl.eversion[0]) > QR_LARGE_VERSION_SLACK) then continue;

    qr_aff_unproject(ul.o, @aff, ul.c^.pos[0], ul.c^.pos[1]);
    qr_finder_edge_pts_aff_classify(@ul,@aff);
    if (qr_finder_estimate_module_size_and_version(@ul, 1 shl res, 1 shl res) < 0) or
       (abs(ul.eversion[1]-ur.eversion[1]) > QR_LARGE_VERSION_SLACK) or
       (abs(ul.eversion[0]-dl.eversion[0]) > QR_LARGE_VERSION_SLACK) then begin
      continue;
    end;

    (*If we made it this far, upgrade the affine homography to a full
       homography.*)
    if (qr_hom_fit(@hom, @ul, @ur, @dl, @bbox, @aff, @_reader^.isaac, _img,_width,_height)<0) then continue;

    Move(bbox, _qrdata^.bbox, sizeof(bbox));
    qr_hom_unproject(ul.o, @hom, ul.c^.pos[0], ul.c^.pos[1]);
    qr_hom_unproject(ur.o, @hom, ur.c^.pos[0], ur.c^.pos[1]);
    qr_hom_unproject(dl.o, @hom, dl.c^.pos[0], dl.c^.pos[1]);
    qr_finder_edge_pts_hom_classify(@ur,@hom);
    if (qr_finder_estimate_module_size_and_version(@ur, ur.o[0]-ul.o[0], ur.o[0]-ul.o[0])<0) then continue;
    qr_finder_edge_pts_hom_classify(@dl,@hom);
    if (qr_finder_estimate_module_size_and_version(@dl, dl.o[1]-ul.o[1], dl.o[1]-ul.o[1])<0) then continue;

    (*If we have a small version (less than 7), there's no encoded version
       information.
      If the estimated version on the two corners matches and is sufficiently
       small, we assume this is the case.*)
    if (ur.eversion[1]=dl.eversion[0]) and (ur.eversion[1]<7) then begin
      (*We used to do a whole bunch of extra geometric checks for small
         versions, because with just an affine correction, it was fairly easy
         to estimate two consistent module sizes given a random configuration.
        However, now that we're estimating a full homography, these appear to
         be unnecessary.*)
      ur_version:=ur.eversion[1];
    end
    else begin
      (*If the estimated versions are significantly different, reject the
         configuration.*)
      if (abs(ur.eversion[1]-dl.eversion[0])>QR_LARGE_VERSION_SLACK) then continue;

      (*Otherwise we try to read the actual version data from the image.
        If the real version is not sufficiently close to our estimated version,
         then we assume there was an unrecoverable decoding error (so many bit
         errors we were within 3 errors of another valid code), and throw that
         value away.
        If no decoded version could be sufficiently close, we don't even try.*)
      if(ur.eversion[1]>=7-QR_LARGE_VERSION_SLACK) then begin
        ur_version:=qr_finder_version_decode(@ur,@hom,_img,_width,_height,0);
        if(abs(ur_version-ur.eversion[1])>QR_LARGE_VERSION_SLACK) then ur_version:=-1;
      end
      else
        ur_version:=-1;

      if(dl.eversion[0]>=7-QR_LARGE_VERSION_SLACK) then begin
        dl_version:=qr_finder_version_decode(@dl,@hom,_img,_width,_height,1);
        if(abs(dl_version-dl.eversion[0])>QR_LARGE_VERSION_SLACK) then dl_version:=-1;
      end
      else
        dl_version:=-1;

      (*If we got at least one valid version, or we got two and they match,
         then we found a valid configuration.*)
      if(ur_version>=0) then begin
        if(dl_version>=0) and (dl_version<>ur_version) then continue;
      end
      else if(dl_version<0) then continue
      else ur_version:=dl_version;
    end;

    qr_finder_edge_pts_hom_classify(@ul,@hom);
    if(qr_finder_estimate_module_size_and_version(@ul, ur.o[0]-dl.o[0], dl.o[1]-ul.o[1])<0) or
      (abs(ul.eversion[1]-ur.eversion[1])>QR_SMALL_VERSION_SLACK) or
      (abs(ul.eversion[0]-dl.eversion[0])>QR_SMALL_VERSION_SLACK) then continue;

    fmt_info:=qr_finder_fmt_info_decode(@ul,@ur,@dl,@hom,_img,_width,_height);
    if (fmt_info<0) or
       (qr_code_decode(_qrdata^, @_reader^.gf, ul.c^.pos, ur.c^.pos, dl.c^.pos, ur_version, fmt_info, _img,_width,_height)<0) then begin
      (*The code may be flipped.
        Try again, swapping the UR and DL centers.
        We should get a valid version either way, so it's relatively cheap to
         check this, as we've already filtered out a lot of invalid
         configurations.*)
      QR_SWAP2I(hom.inv[0][0], hom.inv[1][0]);
      QR_SWAP2I(hom.inv[0][1], hom.inv[1][1]);
      QR_SWAP2I(hom.fwd[0][0], hom.fwd[0][1]);
      QR_SWAP2I(hom.fwd[1][0], hom.fwd[1][1]);
      QR_SWAP2I(hom.fwd[2][0], hom.fwd[2][1]);
      QR_SWAP2I(ul.o[0], ul.o[1]);
      QR_SWAP2I(ul.size[0], ul.size[1]);
      QR_SWAP2I(ur.o[0], ur.o[1]);
      QR_SWAP2I(ur.size[0], ur.size[1]);
      QR_SWAP2I(dl.o[0], dl.o[1]);
      QR_SWAP2I(dl.size[0], dl.size[1]);

      fmt_info:=qr_finder_fmt_info_decode(@ul,@dl,@ur,@hom,_img,_width,_height);
      if (fmt_info<0) then continue;
      QR_SWAP2I(bbox[1][0],bbox[2][0]);
      QR_SWAP2I(bbox[1][1],bbox[2][1]);
      Move(bbox, _qrdata^.bbox, sizeof(bbox));
      if(qr_code_decode(_qrdata^, @_reader^.gf, ul.c^.pos, dl.c^.pos, ur.c^.pos, ur_version, fmt_info, _img,_width,_height)<0) then begin
        continue;
      end;
    end;
    Exit(ur_version);
  end;
  Exit(-1);
end;

procedure qr_reader_match_centers(_reader: pqr_reader; _qrlist: pqr_code_data_list;
 _centers: pqr_finder_center; _ncenters: integer;
 _img: PByte; _width,_height: integer);
var
  (* The number of centers should be small, so an O(n^3) exhaustive search of
     which ones go together should be reasonable.*)
  mark: PByte;
  nfailures_max, nfailures, i,j,k,l, version, ninside: integer;
  c: array[0..2] of pqr_finder_center;
  qrdata: qr_code_data;
  inside: pqr_finder_center;
begin
  mark := AllocMem(_ncenters*sizeof(mark^));
  nfailures_max := QR_MAXI(8192, (_width*_height) shr 9);
  nfailures := 0;
  i := 0;
  while i<_ncenters do begin
    (*TODO: We might be able to accelerate this step significantly by
       considering the remaining finder centers in a more intelligent order,
       based on the first finder center we just chose.*)
    j:=i+1;
    while (mark[i]=0) and (j<_ncenters) do begin
      k:=j+1;
      while (mark[j]=0) and (k<_ncenters) do begin
        if mark[k]=0 then begin
          c[0] := _centers+i;
          c[1] := _centers+j;
          c[2] := _centers+k;
          version := qr_reader_try_configuration(_reader, @qrdata, _img,_width,_height, @c);
          {$IFDEF DEBUG_ZBAR}DebugLn('qr_reader_match_centers[%d,%d,%d]: version=%d',[i,j,k,version]);{$ENDIF}
          if version>=0 then begin
            (*Add the data to the list.*)
            qr_code_data_list_add(_qrlist, @qrdata);
            (*Convert the bounding box we're returning to the user to normal
              image coordinates.*)
            for l:=0 to 3 do begin
              _qrlist^.qrdata[_qrlist^.nqrdata-1].bbox[l][0] := _qrlist^.qrdata[_qrlist^.nqrdata-1].bbox[l][0] shr QR_FINDER_SUBPREC;
              _qrlist^.qrdata[_qrlist^.nqrdata-1].bbox[l][1] := _qrlist^.qrdata[_qrlist^.nqrdata-1].bbox[l][1] shr QR_FINDER_SUBPREC;
            end;

            (*Mark these centers as used.*)
            mark[i]:=1;
            mark[j]:=1;
            mark[k]:=1;

            (*Find any other finder centers located inside this code.*)
            l:=0;
            ninside:=0;
            while l<_ncenters do begin
              if mark[l]=0 then begin
                if (qr_point_ccw(qrdata.bbox[0],qrdata.bbox[1],_centers[l].pos) >= 0) and
                   (qr_point_ccw(qrdata.bbox[1],qrdata.bbox[3],_centers[l].pos) >= 0) and
                   (qr_point_ccw(qrdata.bbox[3],qrdata.bbox[2],_centers[l].pos) >= 0) and
                   (qr_point_ccw(qrdata.bbox[2],qrdata.bbox[0],_centers[l].pos) >= 0) then begin
                  mark[l]:=2;
                  Inc(ninside);
                end;
              end;
              Inc(l);
            end;

            if ninside>=3 then begin
              (*We might have a "Double QR": a code inside a code.
                Copy the relevant centers to a new array and do a search confined
                 to that subset.*)
              inside := GetMem(ninside*sizeof(inside^));
              l:=0;
              ninside:=0;
              while l<_ncenters do begin
                if mark[l]=2 then begin {*&}inside[ninside{++}]:={*&}_centers[l]; Inc(ninside); end;
                Inc(l);
              end;
              qr_reader_match_centers(_reader, _qrlist, inside, ninside, _img,_width,_height);
              FreeMem(inside);
            end;

            (*Mark _all_ such centers used: codes cannot partially overlap.*)
            for l:=0 to _ncenters-1 do if mark[l]=2 then mark[l]:=1;
            nfailures:=0;
          end
          else begin
            Inc(nfailures);
            if ({++}nfailures>nfailures_max) then begin
              (*Give up.
                We're unlikely to find a valid code in all this clutter, and we
                 could spent quite a lot of time trying.*)
              // Added to ZBar on 30. jun. 2011
              //i:=_ncenters;
              //j:=_ncenters;
              //k:=_ncenters;
            end;
            {$IFDEF DEBUG_ZBAR}DebugLn('qr_reader_match_centers[%d,%d,%d]: failure %d',[i,j,k,nfailures]);{$ENDIF}
          end;
        end;
        Inc(k);
      end;
      Inc(j);
    end;
    Inc(i);
  end;
  FreeMem(mark);
end;

function _zbar_qr_decode(reader: pqr_reader;  iscn: pzbar_image_scanner_t;  img: pzbar_image_t): integer;
var
  edge_pts: pqr_finder_edge_pt;
  centers: pqr_finder_center;
  ncenters: integer;
  bin: pbyte;
  qrlist: qr_code_data_list;
{$IFNDEF HAVE_qrdectxt}
  i,j,l: integer;
  entry: qr_code_data_entry;
  sym: pzbar_symbol_t;
{$ENDIF}
begin
  Result := 0;
  edge_pts := nil;
  centers := nil;
  if (reader^.finder_lines[0].nlines<9) or (reader^.finder_lines[1].nlines<9) then Exit;

  //svg_group_start('finder',0,1. div (1 shl QR_FINDER_SUBPREC),0,0,0); 
  ncenters := qr_finder_centers_locate(centers, edge_pts, reader, 0, 0);
  {$IFDEF DEBUG_ZBAR}DebugLn('qr_finder_centers_locate: %dx%d finders, %d centers', [reader^.finder_lines[0].nlines, reader^.finder_lines[1].nlines, ncenters]);{$ENDIF}
  //qr_svg_centers(centers,ncenters); 
  if ncenters>=3 then begin
    bin := qr_binarize(img^.data, img^.width, img^.height);
    
    qr_code_data_list_init(qrlist);
    qr_reader_match_centers(reader, @qrlist, centers, ncenters, bin, img^.width, img^.height);
    {$IFDEF DEBUG_ZBAR}DebugLn('_zbar_qr_decode: %d entries',[qrlist.nqrdata]);{$ENDIF}
    if qrlist.nqrdata>0 then begin
{$IFDEF HAVE_qrdectxt}
      Result := qr_code_data_list_extract_text(qrlist, iscn, img);
{$ELSE}
      // qrlist.qrdata -> iscn^.syms
      for i:=0 to Pred(qrlist.nqrdata) do begin
        sym := _zbar_image_scanner_alloc_sym(iscn,ZBAR_QRCODE_,0);

        sym_add_point(sym^, qrlist.qrdata[i].bbox[0][0], qrlist.qrdata[i].bbox[0][1]);
        sym_add_point(sym^, qrlist.qrdata[i].bbox[2][0], qrlist.qrdata[i].bbox[2][1]);
        sym_add_point(sym^, qrlist.qrdata[i].bbox[3][0], qrlist.qrdata[i].bbox[3][1]);
        sym_add_point(sym^, qrlist.qrdata[i].bbox[1][0], qrlist.qrdata[i].bbox[1][1]);

        l:=0;
        for j:=0 to Pred(qrlist.qrdata[i].nentries) do begin
          entry := qrlist.qrdata[i].entries[j];
          sym^.data := ReAllocMem(sym^.data, l+entry.payload.data.len+1);
          Move(entry.payload.data.buf^, sym^.data[l], entry.payload.data.len);
          Inc(l, entry.payload.data.len);
        end;
        sym^.data[l] := 0;
        sym^.data_alloc := l+1;
        sym^.datalen := l;
        _zbar_image_scanner_add_sym(iscn, sym);
      end;
      Result := 0;
{$ENDIF}
    end;
    qr_code_data_list_clear(qrlist);
    FreeMem(bin);
  end;
  //svg_group_end(); 
  if Assigned(centers) then FreeMem(centers);
  if Assigned(edge_pts) then FreeMem(edge_pts);
end;
{$ENDIF}

end.
